@@ -16,16 +16,11 @@ public class Text implements TextWalker
 	}
 
 	private native long newNative();
-	private native long cloneNative(Text old);
 
 	private Text(long p) {
 		pointer = p;
 	}
 
-	public Text(Text old) {
-		pointer = cloneNative(old);
-	}
-
 	public Text() {
 		pointer = newNative();
 	}
