@@ -59,6 +59,7 @@ struct pdfapp_s
 	char *docpath;
 	char *doctitle;
 	fz_outline *outline;
+	int outline_deferred;
 
 	float layout_w;
 	float layout_h;
@@ -73,11 +74,14 @@ struct pdfapp_s
 	float resolution;
 	int rotate;
 	fz_pixmap *image;
+	int imgw, imgh;
 	int grayscale;
 	fz_colorspace *colorspace;
 	int invert;
 	int tint, tint_white;
+	int useicc;
 	int useseparations;
+	int aalevel;
 
 	/* presentation mode */
 	int presentation_mode;
@@ -98,6 +102,7 @@ struct pdfapp_s
 	fz_stext_page *page_text;
 	fz_link *page_links;
 	int errored;
+	int incomplete;
 
 	/* separations */
 	fz_separations *seps;
@@ -145,11 +150,15 @@ struct pdfapp_s
 	void *userdata;
 
 	fz_context *ctx;
+#ifdef HAVE_CURL
+	fz_stream *stream;
+#endif
 };
 
 void pdfapp_init(fz_context *ctx, pdfapp_t *app);
 void pdfapp_setresolution(pdfapp_t *app, int res);
 void pdfapp_open(pdfapp_t *app, char *filename, int reload);
+void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int kbps);
 void pdfapp_close(pdfapp_t *app);
 int pdfapp_preclose(pdfapp_t *app);
 void pdfapp_reloadfile(pdfapp_t *app);
