@@ -163,77 +163,28 @@ fz_drop_context(fz_context *ctx)
 	fz_drop_document_handler_context(ctx);
 	fz_drop_glyph_cache_context(ctx);
 	fz_drop_store_context(ctx);
-	fz_drop_aa_context(ctx);
 	fz_drop_style_context(ctx);
 	fz_drop_tuning_context(ctx);
 	fz_drop_colorspace_context(ctx);
-	fz_drop_cmm_context(ctx);
 	fz_drop_font_context(ctx);
-	fz_drop_output_context(ctx);
 
-	if (ctx->warn)
-	{
-		fz_flush_warnings(ctx);
-		fz_free(ctx, ctx->warn);
-	}
+	fz_flush_warnings(ctx);
 
-	if (ctx->error)
-	{
-		assert(ctx->error->top == ctx->error->stack);
-		fz_free(ctx, ctx->error);
-	}
+	assert(ctx->error.top == ctx->error.stack);
 
 	/* Free the context itself */
-	ctx->alloc->free(ctx->alloc->user, ctx);
+	ctx->alloc.free(ctx->alloc.user, ctx);
 }
 
-/* Allocate new context structure, and initialise allocator, and sections
- * that aren't shared between contexts.
- */
-static fz_context *
-new_context_phase1(const fz_alloc_context *alloc, const fz_locks_context *locks)
+static void
+fz_init_error_context(fz_context *ctx)
 {
-	fz_context *ctx;
+	ctx->error.top = ctx->error.stack;
+	ctx->error.errcode = FZ_ERROR_NONE;
+	ctx->error.message[0] = 0;
 
-	ctx = alloc->malloc(alloc->user, sizeof(fz_context));
-	if (!ctx)
-		return NULL;
-	memset(ctx, 0, sizeof *ctx);
-	ctx->user = NULL;
-	ctx->alloc = alloc;
-	ctx->locks = *locks;
-
-	ctx->glyph_cache = NULL;
-
-	ctx->error = Memento_label(fz_malloc_no_throw(ctx, sizeof(fz_error_context)), "fz_error_context");
-	if (!ctx->error)
-		goto cleanup;
-	ctx->error->top = ctx->error->stack;
-	ctx->error->errcode = FZ_ERROR_NONE;
-	ctx->error->message[0] = 0;
-
-	ctx->warn = Memento_label(fz_malloc_no_throw(ctx, sizeof(fz_warn_context)), "fz_warn_context");
-	if (!ctx->warn)
-		goto cleanup;
-	ctx->warn->message[0] = 0;
-	ctx->warn->count = 0;
-
-	/* New initialisation calls for context entries go here */
-	fz_try(ctx)
-	{
-		fz_new_aa_context(ctx);
-	}
-	fz_catch(ctx)
-	{
-		goto cleanup;
-	}
-
-	return ctx;
-
-cleanup:
-	fprintf(stderr, "cannot create context (phase 1)\n");
-	fz_drop_context(ctx);
-	return NULL;
+	ctx->warn.message[0] = 0;
+	ctx->warn.count = 0;
 }
 
 /*
@@ -281,23 +232,35 @@ fz_new_context_imp(const fz_alloc_context *alloc, const fz_locks_context *locks,
 	if (!locks)
 		locks = &fz_locks_default;
 
-	ctx = new_context_phase1(alloc, locks);
+	ctx = alloc->malloc(alloc->user, sizeof(fz_context));
 	if (!ctx)
+	{
+		fprintf(stderr, "cannot create context (phase 1)\n");
 		return NULL;
+	}
+	memset(ctx, 0, sizeof *ctx);
+
+	ctx->user = NULL;
+	ctx->alloc = *alloc;
+	ctx->locks = *locks;
+
+	ctx->error.print = fz_default_error_callback;
+	ctx->warn.print = fz_default_warning_callback;
+
+	fz_init_error_context(ctx);
+	fz_init_aa_context(ctx);
+	fz_init_random_context(ctx);
 
 	/* Now initialise sections that are shared */
 	fz_try(ctx)
 	{
-		fz_new_output_context(ctx);
 		fz_new_store_context(ctx, max_store);
 		fz_new_glyph_cache_context(ctx);
-		fz_new_cmm_context(ctx);
 		fz_new_colorspace_context(ctx);
 		fz_new_font_context(ctx);
 		fz_new_document_handler_context(ctx);
 		fz_new_style_context(ctx);
 		fz_new_tuning_context(ctx);
-		fz_init_random_context(ctx);
 	}
 	fz_catch(ctx)
 	{
@@ -326,48 +289,31 @@ fz_new_context_imp(const fz_alloc_context *alloc, const fz_locks_context *locks,
 fz_context *
 fz_clone_context(fz_context *ctx)
 {
+	fz_context *new_ctx;
+
 	/* We cannot safely clone the context without having locking/
 	 * unlocking functions. */
 	if (ctx == NULL || (ctx->locks.lock == fz_locks_default.lock && ctx->locks.unlock == fz_locks_default.unlock))
 		return NULL;
-	return fz_clone_context_internal(ctx);
-}
 
-fz_context *
-fz_clone_context_internal(fz_context *ctx)
-{
-	fz_context *new_ctx;
-
-	if (ctx == NULL || ctx->alloc == NULL)
-		return NULL;
-
-	new_ctx = new_context_phase1(ctx->alloc, &ctx->locks);
+	new_ctx = ctx->alloc.malloc(ctx->alloc.user, sizeof(fz_context));
 	if (!new_ctx)
 		return NULL;
 
-	/* Inherit AA defaults from old context. */
-	fz_copy_aa_context(new_ctx, ctx);
+	/* First copy old context, including pointers to shared contexts */
+	memcpy(new_ctx, ctx, sizeof (fz_context));
 
-	/* Keep thread lock checking happy by copying pointers first and locking under new context */
-	new_ctx->output = ctx->output;
-	new_ctx->output = fz_keep_output_context(new_ctx);
-	new_ctx->user = ctx->user;
-	new_ctx->store = ctx->store;
-	new_ctx->store = fz_keep_store_context(new_ctx);
-	new_ctx->glyph_cache = ctx->glyph_cache;
-	new_ctx->glyph_cache = fz_keep_glyph_cache(new_ctx);
-	new_ctx->colorspace = ctx->colorspace;
-	new_ctx->colorspace = fz_keep_colorspace_context(new_ctx);
-	fz_new_cmm_context(new_ctx);
-	new_ctx->font = ctx->font;
-	new_ctx->font = fz_keep_font_context(new_ctx);
-	new_ctx->style = ctx->style;
-	new_ctx->style = fz_keep_style_context(new_ctx);
-	new_ctx->tuning = ctx->tuning;
-	new_ctx->tuning = fz_keep_tuning_context(new_ctx);
-	memcpy(new_ctx->seed48, ctx->seed48, sizeof ctx->seed48);
-	new_ctx->handler = ctx->handler;
-	new_ctx->handler = fz_keep_document_handler_context(new_ctx);
+	/* Reset error context to initial state. */
+	fz_init_error_context(new_ctx);
+
+	/* Then keep lock checking happy by keeping shared contexts with new context */
+	fz_keep_document_handler_context(new_ctx);
+	fz_keep_style_context(new_ctx);
+	fz_keep_tuning_context(new_ctx);
+	fz_keep_font_context(new_ctx);
+	fz_keep_colorspace_context(new_ctx);
+	fz_keep_store_context(new_ctx);
+	fz_keep_glyph_cache(new_ctx);
 
 	return new_ctx;
 }
