@@ -1,5 +1,4 @@
 #include "fitz-imp.h"
-#include "glyph-cache-imp.h"
 #include "draw-imp.h"
 
 #include <string.h>
@@ -149,12 +148,12 @@ static void grow_stack(fz_context *ctx, fz_draw_device *dev)
 	fz_draw_state *stack;
 	if (dev->stack == &dev->init_stack[0])
 	{
-		stack = Memento_label(fz_malloc_array(ctx, max, sizeof *stack), "draw device stack");
+		stack = fz_malloc_array(ctx, max, fz_draw_state);
 		memcpy(stack, dev->stack, sizeof(*stack) * dev->stack_cap);
 	}
 	else
 	{
-		stack = fz_resize_array(ctx, dev->stack, max, sizeof(*stack));
+		stack = fz_realloc_array(ctx, dev->stack, max, fz_draw_state);
 	}
 	dev->stack = stack;
 	dev->stack_cap = max;
@@ -480,7 +479,7 @@ resolve_color(fz_context *ctx,
 	const float *color,
 	fz_colorspace *colorspace,
 	float alpha,
-	const fz_color_params *color_params,
+	fz_color_params color_params,
 	unsigned char *colorbv,
 	fz_pixmap *dest)
 {
@@ -494,15 +493,12 @@ resolve_color(fz_context *ctx,
 	if (colorspace == NULL && model != NULL)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "color destination requires source color");
 
-	if (color_params == NULL)
-		color_params = fz_default_color_params(ctx);
-
-	effective_opm = color_params->opm;
+	effective_opm = color_params.opm;
 	devn = fz_colorspace_is_device_n(ctx, colorspace);
 	devgray = fz_colorspace_is_device_gray(ctx, colorspace);
 
 	/* We can only overprint when enabled, and when we are in a subtractive colorspace */
-	if (!color_params || color_params->op == 0 || !fz_colorspace_is_subtractive(ctx, dest->colorspace))
+	if (color_params.op == 0 || !fz_colorspace_is_subtractive(ctx, dest->colorspace))
 		op = NULL;
 
 	/* Device Gray is additive, but seems to still be counted for overprint
@@ -523,7 +519,7 @@ resolve_color(fz_context *ctx,
 		i = 0;
 	else if (devn && colors_supported(ctx, colorspace, dest))
 	{
-		fz_convert_separation_colors(ctx, color_params, dest->colorspace, dest->seps, colorfv, colorspace, color);
+		fz_convert_separation_colors(ctx, colorspace, color, dest->seps, dest->colorspace, colorfv, color_params);
 		for (i = 0; i < n; i++)
 			colorbv[i] = colorfv[i] * 255;
 		op = set_op_from_spaces(ctx, op, dest, colorspace, effective_opm);
@@ -531,7 +527,7 @@ resolve_color(fz_context *ctx,
 	else
 	{
 		int c = n - dest->s;
-		fz_convert_color(ctx, color_params, NULL, dest->colorspace, colorfv, colorspace, color);
+		fz_convert_color(ctx, colorspace, color, dest->colorspace, colorfv, NULL, color_params);
 		for (i = 0; i < c; i++)
 			colorbv[i] = colorfv[i] * 255;
 		for (; i < n; i++)
@@ -560,7 +556,7 @@ resolve_color(fz_context *ctx,
 }
 
 static fz_draw_state *
-push_group_for_separations(fz_context *ctx, fz_draw_device *dev, const fz_color_params *color_params, fz_default_colorspaces *default_cs)
+push_group_for_separations(fz_context *ctx, fz_draw_device *dev, fz_color_params color_params, fz_default_colorspaces *default_cs)
 {
 	fz_separations *clone = fz_clone_separations_for_overprint(ctx, dev->stack[0].dest->seps);
 	fz_colorspace *oi = fz_default_output_intent(ctx, default_cs);
@@ -602,7 +598,7 @@ push_group_for_separations(fz_context *ctx, fz_draw_device *dev, const fz_color_
 
 static void
 fz_draw_fill_path(fz_context *ctx, fz_device *devp, const fz_path *path, int even_odd, fz_matrix in_ctm,
-	fz_colorspace *colorspace_in, const float *color, float alpha, const fz_color_params *color_params)
+	fz_colorspace *colorspace_in, const float *color, float alpha, fz_color_params color_params)
 {
 	fz_draw_device *dev = (fz_draw_device*)devp;
 	fz_matrix ctm = fz_concat(in_ctm, dev->transform);
@@ -658,7 +654,7 @@ fz_draw_fill_path(fz_context *ctx, fz_device *devp, const fz_path *path, int eve
 
 static void
 fz_draw_stroke_path(fz_context *ctx, fz_device *devp, const fz_path *path, const fz_stroke_state *stroke, fz_matrix in_ctm,
-	fz_colorspace *colorspace_in, const float *color, float alpha, const fz_color_params *color_params)
+	fz_colorspace *colorspace_in, const float *color, float alpha, fz_color_params color_params)
 {
 	fz_draw_device *dev = (fz_draw_device*)devp;
 	fz_matrix ctm = fz_concat(in_ctm, dev->transform);
@@ -752,7 +748,7 @@ fz_draw_clip_path(fz_context *ctx, fz_device *devp, const fz_path *path, int eve
 	fz_colorspace *model;
 
 	if (dev->top == 0 && dev->resolve_spots)
-		state = push_group_for_separations(ctx, dev, fz_default_color_params(ctx)/* FIXME */, dev->default_cs);
+		state = push_group_for_separations(ctx, dev, fz_default_color_params /* FIXME */, dev->default_cs);
 
 	if (expansion < FLT_EPSILON)
 		expansion = 1;
@@ -826,7 +822,7 @@ fz_draw_clip_stroke_path(fz_context *ctx, fz_device *devp, const fz_path *path,
 	float mlw = fz_rasterizer_graphics_min_line_width(rast);
 
 	if (dev->top == 0 && dev->resolve_spots)
-		state = push_group_for_separations(ctx, dev, fz_default_color_params(ctx) /* FIXME */, dev->default_cs);
+		state = push_group_for_separations(ctx, dev, fz_default_color_params /* FIXME */, dev->default_cs);
 
 	if (mlw > aa_level)
 		aa_level = mlw;
@@ -967,7 +963,7 @@ draw_glyph(unsigned char *colorbv, fz_pixmap *dst, fz_glyph *glyph,
 
 static void
 fz_draw_fill_text(fz_context *ctx, fz_device *devp, const fz_text *text, fz_matrix in_ctm,
-	fz_colorspace *colorspace_in, const float *color, float alpha, const fz_color_params *color_params)
+	fz_colorspace *colorspace_in, const float *color, float alpha, fz_color_params color_params)
 {
 	fz_draw_device *dev = (fz_draw_device*)devp;
 	fz_matrix ctm = fz_concat(in_ctm, dev->transform);
@@ -991,9 +987,6 @@ fz_draw_fill_text(fz_context *ctx, fz_device *devp, const fz_text *text, fz_matr
 	if (colorspace == NULL && model != NULL)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "color destination requires source color");
 
-	if (color_params == NULL)
-		color_params = fz_default_color_params(ctx);
-
 	if (state->blendmode & FZ_BLEND_KNOCKOUT)
 		state = fz_knockout_begin(ctx, dev);
 
@@ -1064,7 +1057,7 @@ fz_draw_fill_text(fz_context *ctx, fz_device *devp, const fz_text *text, fz_matr
 
 static void
 fz_draw_stroke_text(fz_context *ctx, fz_device *devp, const fz_text *text, const fz_stroke_state *stroke,
-	fz_matrix in_ctm, fz_colorspace *colorspace_in, const float *color, float alpha, const fz_color_params *color_params)
+	fz_matrix in_ctm, fz_colorspace *colorspace_in, const float *color, float alpha, fz_color_params color_params)
 {
 	fz_draw_device *dev = (fz_draw_device*)devp;
 	fz_matrix ctm = fz_concat(in_ctm, dev->transform);
@@ -1155,7 +1148,7 @@ fz_draw_clip_text(fz_context *ctx, fz_device *devp, const fz_text *text, fz_matr
 	fz_rasterizer *rast = dev->rast;
 
 	if (dev->top == 0 && dev->resolve_spots)
-		state = push_group_for_separations(ctx, dev, fz_default_color_params(ctx)/* FIXME */, dev->default_cs);
+		state = push_group_for_separations(ctx, dev, fz_default_color_params /* FIXME */, dev->default_cs);
 
 	state = push_stack(ctx, dev, "clip text");
 
@@ -1244,7 +1237,7 @@ fz_draw_clip_text(fz_context *ctx, fz_device *devp, const fz_text *text, fz_matr
 						state[1].mask = NULL;
 						fz_try(ctx)
 						{
-							fz_draw_fill_path(ctx, devp, path, 0, in_ctm, fz_device_gray(ctx), &white, 1, NULL);
+							fz_draw_fill_path(ctx, devp, path, 0, in_ctm, fz_device_gray(ctx), &white, 1, fz_default_color_params);
 						}
 						fz_always(ctx)
 						{
@@ -1283,7 +1276,7 @@ fz_draw_clip_stroke_text(fz_context *ctx, fz_device *devp, const fz_text *text,
 	int aa = fz_rasterizer_text_aa_level(dev->rast);
 
 	if (dev->top == 0 && dev->resolve_spots)
-		state = push_group_for_separations(ctx, dev, fz_default_color_params(ctx)/* FIXME */, dev->default_cs);
+		state = push_group_for_separations(ctx, dev, fz_default_color_params /* FIXME */, dev->default_cs);
 
 	/* make the mask the exact size needed */
 	bbox = fz_irect_from_rect(fz_bound_text(ctx, text, stroke, ctm));
@@ -1369,7 +1362,7 @@ fz_draw_clip_stroke_text(fz_context *ctx, fz_device *devp, const fz_text *text,
 						state[0].mask = NULL;
 						fz_try(ctx)
 						{
-							fz_draw_stroke_path(ctx, devp, path, stroke, in_ctm, fz_device_gray(ctx), &white, 1, NULL);
+							fz_draw_stroke_path(ctx, devp, path, stroke, in_ctm, fz_device_gray(ctx), &white, 1, fz_default_color_params);
 						}
 						fz_always(ctx)
 						{
@@ -1398,7 +1391,7 @@ fz_draw_ignore_text(fz_context *ctx, fz_device *dev, const fz_text *text, fz_mat
 }
 
 static void
-fz_draw_fill_shade(fz_context *ctx, fz_device *devp, fz_shade *shade, fz_matrix in_ctm, float alpha, const fz_color_params *color_params)
+fz_draw_fill_shade(fz_context *ctx, fz_device *devp, fz_shade *shade, fz_matrix in_ctm, float alpha, fz_color_params color_params)
 {
 	fz_draw_device *dev = (fz_draw_device*)devp;
 	fz_matrix ctm = fz_concat(in_ctm, dev->transform);
@@ -1421,9 +1414,6 @@ fz_draw_fill_shade(fz_context *ctx, fz_device *devp, fz_shade *shade, fz_matrix
 	if (fz_is_empty_irect(bbox))
 		return;
 
-	if (color_params == NULL)
-		color_params = fz_default_color_params(ctx);
-
 	if (state->blendmode & FZ_BLEND_KNOCKOUT)
 		state = fz_knockout_begin(ctx, dev);
 
@@ -1460,18 +1450,11 @@ fz_draw_fill_shade(fz_context *ctx, fz_device *devp, fz_shade *shade, fz_matrix
 		{
 			unsigned char *s;
 			int x, y, n, i;
-			fz_color_params  local_cp;
-			fz_color_params *cp = NULL;
 
 			/* Disable OPM */
-			if (color_params)
-			{
-				local_cp = *color_params;
-				local_cp.opm = 0;
-				cp = &local_cp;
-			}
+			color_params.opm = 0;
 
-			eop = resolve_color(ctx, &op, shade->background, colorspace, alpha, cp, colorbv, state->dest);
+			eop = resolve_color(ctx, &op, shade->background, colorspace, alpha, color_params, colorbv, state->dest);
 
 			n = dest->n;
 			if (fz_overprint_required(eop))
@@ -1523,10 +1506,8 @@ fz_draw_fill_shade(fz_context *ctx, fz_device *devp, fz_shade *shade, fz_matrix
 			}
 		}
 
-		if (color_params->op)
-		{
+		if (color_params.op)
 			eop = set_op_from_spaces(ctx, &op, dest, colorspace, 0);
-		}
 		else
 			eop = NULL;
 
@@ -1645,7 +1626,7 @@ fz_default_image_scale(void *arg, int dst_w, int dst_h, int src_w, int src_h)
 }
 
 static fz_pixmap *
-convert_pixmap_for_painting(fz_context *ctx, fz_pixmap *pixmap, fz_colorspace *model, fz_colorspace *src_cs, fz_pixmap *dest, const fz_color_params *color_params, fz_draw_device *dev, fz_overprint **eop)
+convert_pixmap_for_painting(fz_context *ctx, fz_pixmap *pixmap, fz_colorspace *model, fz_colorspace *src_cs, fz_pixmap *dest, fz_color_params color_params, fz_draw_device *dev, fz_overprint **eop)
 {
 	fz_pixmap *converted;
 
@@ -1683,7 +1664,7 @@ convert_pixmap_for_painting(fz_context *ctx, fz_pixmap *pixmap, fz_colorspace *m
 }
 
 static void
-fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix in_ctm, float alpha, const fz_color_params *color_params)
+fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix in_ctm, float alpha, fz_color_params color_params)
 {
 	fz_draw_device *dev = (fz_draw_device*)devp;
 	fz_matrix local_ctm = fz_concat(in_ctm, dev->transform);
@@ -1711,7 +1692,7 @@ fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix
 	if (image->w == 0 || image->h == 0)
 		return;
 
-	if (!color_params || color_params->op == 0)
+	if (color_params.op == 0)
 		eop = NULL;
 
 	/* ctm maps the image (expressed as the unit square) onto the
@@ -1830,7 +1811,7 @@ fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix
 
 static void
 fz_draw_fill_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix in_ctm,
-	fz_colorspace *colorspace_in, const float *color, float alpha, const fz_color_params *color_params)
+	fz_colorspace *colorspace_in, const float *color, float alpha, fz_color_params color_params)
 {
 	fz_draw_device *dev = (fz_draw_device*)devp;
 	fz_matrix local_ctm = fz_concat(in_ctm, dev->transform);
@@ -1953,7 +1934,7 @@ fz_draw_clip_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 	fz_var(pixmap);
 
 	if (dev->top == 0 && dev->resolve_spots)
-		state = push_group_for_separations(ctx, dev, fz_default_color_params(ctx)/* FIXME */, dev->default_cs);
+		state = push_group_for_separations(ctx, dev, fz_default_color_params /* FIXME */, dev->default_cs);
 
 	clip = fz_pixmap_bbox(ctx, state->dest);
 	clip = fz_intersect_irect(clip, state->scissor);
@@ -2116,7 +2097,7 @@ fz_draw_pop_clip(fz_context *ctx, fz_device *devp)
 }
 
 static void
-fz_draw_begin_mask(fz_context *ctx, fz_device *devp, fz_rect area, int luminosity, fz_colorspace *colorspace_in, const float *colorfv, const fz_color_params *color_params)
+fz_draw_begin_mask(fz_context *ctx, fz_device *devp, fz_rect area, int luminosity, fz_colorspace *colorspace_in, const float *colorfv, fz_color_params color_params)
 {
 	fz_draw_device *dev = (fz_draw_device*)devp;
 	fz_pixmap *dest;
@@ -2133,9 +2114,6 @@ fz_draw_begin_mask(fz_context *ctx, fz_device *devp, fz_rect area, int luminosit
 	if (colorspace_in)
 		colorspace = fz_default_colorspace(ctx, dev->default_cs, colorspace_in);
 
-	if (color_params == NULL)
-		color_params = fz_default_color_params(ctx);
-
 	trect = fz_transform_rect(area, dev->transform);
 	bbox = fz_intersect_irect(fz_irect_from_rect(trect), state->scissor);
 
@@ -2170,7 +2148,7 @@ fz_draw_begin_mask(fz_context *ctx, fz_device *devp, fz_rect area, int luminosit
 		float bc;
 		if (!colorspace)
 			colorspace = fz_device_gray(ctx);
-		fz_convert_color(ctx, color_params, NULL, fz_device_gray(ctx), &bc, colorspace, colorfv);
+		fz_convert_color(ctx, colorspace, colorfv, fz_device_gray(ctx), &bc, NULL, color_params);
 		fz_clear_pixmap_with_value(ctx, dest, bc * 255);
 		if (shape)
 			fz_clear_pixmap_with_value(ctx, shape, 255);
@@ -2269,7 +2247,7 @@ fz_draw_begin_group(fz_context *ctx, fz_device *devp, fz_rect area, fz_colorspac
 	fz_rect trect;
 
 	if (dev->top == 0 && dev->resolve_spots)
-		state = push_group_for_separations(ctx, dev, fz_default_color_params(ctx)/* FIXME */, dev->default_cs);
+		state = push_group_for_separations(ctx, dev, fz_default_color_params /* FIXME */, dev->default_cs);
 
 	if (cs != NULL)
 		model = fz_default_colorspace(ctx, dev->default_cs, cs);
@@ -2374,7 +2352,7 @@ fz_draw_end_group(fz_context *ctx, fz_device *devp)
 
 	if (state[0].dest->colorspace != state[1].dest->colorspace)
 	{
-		fz_pixmap *converted = fz_convert_pixmap(ctx, state[1].dest, state[0].dest->colorspace, NULL, dev->default_cs, fz_default_color_params(ctx), 1);
+		fz_pixmap *converted = fz_convert_pixmap(ctx, state[1].dest, state[0].dest->colorspace, NULL, dev->default_cs, fz_default_color_params, 1);
 		fz_drop_pixmap(ctx, state[1].dest);
 		state[1].dest = converted;
 	}
@@ -2439,6 +2417,8 @@ typedef struct
 	int refs;
 	float ctm[4];
 	int id;
+	char has_shape;
+	char has_group_alpha;
 	fz_colorspace *cs;
 } tile_key;
 
@@ -2456,6 +2436,8 @@ fz_make_hash_tile_key(fz_context *ctx, fz_store_hash *hash, void *key_)
 	tile_key *key = key_;
 
 	hash->u.im.id = key->id;
+	hash->u.im.has_shape = key->has_shape;
+	hash->u.im.has_group_alpha = key->has_group_alpha;
 	hash->u.im.m[0] = key->ctm[0];
 	hash->u.im.m[1] = key->ctm[1];
 	hash->u.im.m[2] = key->ctm[2];
@@ -2488,6 +2470,8 @@ fz_cmp_tile_key(fz_context *ctx, void *k0_, void *k1_)
 	tile_key *k0 = k0_;
 	tile_key *k1 = k1_;
 	return k0->id == k1->id &&
+		k0->has_shape == k1->has_shape &&
+		k0->has_group_alpha == k1->has_group_alpha &&
 		k0->ctm[0] == k1->ctm[0] &&
 		k0->ctm[1] == k1->ctm[1] &&
 		k0->ctm[2] == k1->ctm[2] &&
@@ -2499,8 +2483,9 @@ static void
 fz_format_tile_key(fz_context *ctx, char *s, int n, void *key_)
 {
 	tile_key *key = (tile_key *)key_;
-	fz_snprintf(s, n, "(tile id=%x, ctm=%g %g %g %g, cs=%x)",
-			key->id, key->ctm[0], key->ctm[1], key->ctm[2], key->ctm[3], key->cs);
+	fz_snprintf(s, n, "(tile id=%x, ctm=%g %g %g %g, cs=%x, shape=%d, ga=%d)",
+			key->id, key->ctm[0], key->ctm[1], key->ctm[2], key->ctm[3], key->cs,
+			key->has_shape, key->has_group_alpha);
 }
 
 static const fz_store_type fz_tile_store_type =
@@ -2561,7 +2546,7 @@ fz_draw_begin_tile(fz_context *ctx, fz_device *devp, fz_rect area, fz_rect view,
 	fz_rect local_view;
 
 	if (dev->top == 0 && dev->resolve_spots)
-		state = push_group_for_separations(ctx, dev, fz_default_color_params(ctx)/* FIXME */, dev->default_cs);
+		state = push_group_for_separations(ctx, dev, fz_default_color_params /* FIXME */, dev->default_cs);
 
 	/* area, view, xstep, ystep are in pattern space */
 	/* ctm maps from pattern space to device space */
@@ -2592,6 +2577,8 @@ fz_draw_begin_tile(fz_context *ctx, fz_device *devp, fz_rect area, fz_rect view,
 		tk.ctm[3] = ctm.d;
 		tk.id = id;
 		tk.cs = state[1].dest->colorspace;
+		tk.has_shape = (state[1].shape != NULL);
+		tk.has_group_alpha = (state[1].group_alpha != NULL);
 
 		tile = fz_find_item(ctx, fz_drop_tile_record_imp, &tk, &fz_tile_store_type);
 		if (tile)
@@ -2805,6 +2792,8 @@ fz_draw_end_tile(fz_context *ctx, fz_device *devp)
 			key->ctm[2] = ctm.c;
 			key->ctm[3] = ctm.d;
 			key->cs = fz_keep_colorspace_store_key(ctx, state[1].dest->colorspace);
+			key->has_shape = (state[1].shape != NULL);
+			key->has_group_alpha = (state[1].group_alpha != NULL);
 			existing_tile = fz_store_item(ctx, key, tile, fz_tile_size(ctx, tile), &fz_tile_store_type);
 			if (existing_tile)
 			{
@@ -2875,7 +2864,7 @@ fz_draw_close_device(fz_context *ctx, fz_device *devp)
 		fz_draw_state *state = &dev->stack[--dev->top];
 		fz_try(ctx)
 		{
-			fz_copy_pixmap_area_converting_seps(ctx, state[0].dest, state[1].dest, fz_default_color_params(ctx)/* FIXME */, dev->proof_cs, dev->default_cs);
+			fz_copy_pixmap_area_converting_seps(ctx, state[1].dest, state[0].dest, dev->proof_cs, fz_default_color_params, dev->default_cs);
 			assert(state[1].mask == NULL);
 			assert(state[1].shape == NULL);
 			assert(state[1].group_alpha == NULL);
@@ -3235,7 +3224,7 @@ fz_parse_draw_options(fz_context *ctx, fz_draw_options *opts, const char *args)
 fz_device *
 fz_new_draw_device_with_options(fz_context *ctx, const fz_draw_options *opts, fz_rect mediabox, fz_pixmap **pixmap)
 {
-	fz_aa_context aa = *ctx->aa;
+	fz_aa_context aa = ctx->aa;
 	float x_zoom = opts->x_resolution / 72.0f;
 	float y_zoom = opts->y_resolution / 72.0f;
 	float page_w = mediabox.x1 - mediabox.x0;
