@@ -41,29 +41,28 @@ static void
 close_leech(fz_context *ctx, void *state_)
 {
 	fz_leech *state = (fz_leech *)state_;
-
 	fz_drop_stream(ctx, state->chain);
+	fz_drop_buffer(ctx, state->buffer);
 	fz_free(ctx, state);
 }
 
+/*
+	Attach a filter to a stream that will store any
+	characters read from the stream into the supplied buffer.
+
+	chain: The underlying stream to leech from.
+
+	buf: The buffer into which the read data should be appended.
+	The buffer will be resized as required.
+
+	Returns pointer to newly created stream. May throw exceptions on
+	failure to allocate.
+*/
 fz_stream *
 fz_open_leecher(fz_context *ctx, fz_stream *chain, fz_buffer *buffer)
 {
-	fz_leech *state = NULL;
-
-	fz_var(state);
-
-	fz_try(ctx)
-	{
-		state = fz_malloc_struct(ctx, fz_leech);
-		state->chain = chain;
-		state->buffer = buffer;
-	}
-	fz_catch(ctx)
-	{
-		fz_free(ctx, state);
-		fz_drop_stream(ctx, chain);
-		fz_rethrow(ctx);
-	}
+	fz_leech *state = fz_malloc_struct(ctx, fz_leech);
+	state->chain = fz_keep_stream(ctx, chain);
+	state->buffer = fz_keep_buffer(ctx, buffer);
 	return fz_new_stream(ctx, state, next_leech, close_leech);
 }
