@@ -402,7 +402,8 @@ pnm_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 		h = img->h;
 		n = img->n;
 
-		if (pnm->maxval == 255) {
+		if (pnm->maxval == 255)
+		{
 			memcpy(dp, p, w * h * n);
 			p += n * w * h;
 		}
@@ -468,8 +469,7 @@ pam_binary_read_header(fz_context *ctx, struct info *pnm, const unsigned char *p
 		case TOKEN_MAXVAL: p = pnm_read_number(ctx, p, e, &pnm->maxval); break;
 		case TOKEN_TUPLTYPE: p = pnm_read_tupletype(ctx, p, e, &pnm->tupletype); break;
 		case TOKEN_ENDHDR: break;
-		default:
-			   fz_throw(ctx, FZ_ERROR_GENERIC, "unknown header token in pnm image");
+		default: fz_throw(ctx, FZ_ERROR_GENERIC, "unknown header token in pnm image");
 		}
 
 		if (token != TOKEN_ENDHDR)
@@ -570,7 +570,8 @@ pam_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 
 		/* some encoders incorrectly pack bits into bytes and invert the image */
 		packed = 0;
-		if (pnm->maxval == 1) {
+		if (pnm->maxval == 1)
+		{
 			const unsigned char *e_packed = p + w * h * n / 8;
 			if (e_packed < e - 1 && e_packed[0] == 'P' && e_packed[1] >= '0' && e_packed[1] <= '7')
 				e = e_packed;
@@ -611,7 +612,8 @@ pam_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 
 			/* some encoders incorrectly pack bits into bytes and invert the image */
 			packed = 0;
-			if (pnm->maxval == 1) {
+			if (pnm->maxval == 1)
+			{
 				const unsigned char *e_packed = p + w * h * n / 8;
 				if (e_packed < e - 1 && e_packed[0] == 'P' && e_packed[1] >= '0' && e_packed[1] <= '7')
 					e = e_packed;
