@@ -13,16 +13,6 @@ typedef struct ps_band_writer_s
 	unsigned char *output;
 } ps_band_writer;
 
-static void *zalloc_ps(void *opaque, unsigned int items, unsigned int size)
-{
-	return fz_malloc_array_no_throw(opaque, items, size);
-}
-
-static void zfree_ps(void *opaque, void *address)
-{
-	fz_free(opaque, address);
-}
-
 void
 fz_write_ps_file_header(fz_context *ctx, fz_output *out)
 {
@@ -53,7 +43,7 @@ fz_write_ps_file_trailer(fz_context *ctx, fz_output *out, int pages)
 }
 
 static void
-ps_write_header(fz_context *ctx, fz_band_writer *writer_, const fz_colorspace *cs)
+ps_write_header(fz_context *ctx, fz_band_writer *writer_, fz_colorspace *cs)
 {
 	ps_band_writer *writer = (ps_band_writer *)writer_;
 	fz_output *out = writer->super.out;
@@ -80,8 +70,8 @@ ps_write_header(fz_context *ctx, fz_band_writer *writer_, const fz_colorspace *c
 	writer->super.h = h;
 	writer->super.n = n;
 
-	writer->stream.zalloc = zalloc_ps;
-	writer->stream.zfree = zfree_ps;
+	writer->stream.zalloc = fz_zlib_alloc;
+	writer->stream.zfree = fz_zlib_free;
 	writer->stream.opaque = ctx;
 
 	err = deflateInit(&writer->stream, Z_DEFAULT_COMPRESSION);
@@ -304,26 +294,22 @@ ps_end_page(fz_context *ctx, fz_document_writer *wri_, fz_device *dev)
 	fz_pixmap *pix = wri->pixmap;
 	fz_band_writer *bw;
 
-	fz_try(ctx)
-		fz_close_device(ctx, dev);
-	fz_always(ctx)
-		fz_drop_device(ctx, dev);
-	fz_catch(ctx)
-		fz_rethrow(ctx);
-
-	bw = fz_new_ps_band_writer(ctx, wri->out);
 	fz_try(ctx)
 	{
+		fz_close_device(ctx, dev);
+		bw = fz_new_ps_band_writer(ctx, wri->out);
 		fz_write_header(ctx, bw, pix->w, pix->h, pix->n, pix->alpha, pix->xres, pix->yres, 0, pix->colorspace, pix->seps);
 		fz_write_band(ctx, bw, pix->stride, pix->h, pix->samples);
 	}
 	fz_always(ctx)
+	{
+		fz_drop_device(ctx, dev);
 		fz_drop_band_writer(ctx, bw);
+		fz_drop_pixmap(ctx, wri->pixmap);
+		wri->pixmap = NULL;
+	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
-
-	fz_drop_pixmap(ctx, wri->pixmap);
-	wri->pixmap = NULL;
 }
 
 static void
