@@ -30,6 +30,29 @@ fz_drop_pixmap_imp(fz_context *ctx, fz_storable *pix_)
 	fz_free(ctx, pix);
 }
 
+/*
+	Create a new pixmap, with its origin at
+	(0,0) using the supplied data block.
+
+	cs: The colorspace to use for the pixmap, or NULL for an alpha
+	plane/mask.
+
+	w: The width of the pixmap (in pixels)
+
+	h: The height of the pixmap (in pixels)
+
+	seps: Details of separations.
+
+	alpha: 0 for no alpha, 1 for alpha.
+
+	stride: The byte offset from the pixel data in a row to the pixel
+	data in the next row.
+
+	samples: The data block to keep the samples in.
+
+	Returns a pointer to the new pixmap. Throws exception on failure to
+	allocate.
+*/
 fz_pixmap *
 fz_new_pixmap_with_data(fz_context *ctx, fz_colorspace *colorspace, int w, int h, fz_separations *seps, int alpha, int stride, unsigned char *samples)
 {
@@ -45,6 +68,8 @@ fz_new_pixmap_with_data(fz_context *ctx, fz_colorspace *colorspace, int w, int h
 		fz_throw(ctx, FZ_ERROR_GENERIC, "Illegal stride for pixmap (n=%d w=%d, stride=%d)", n, w, stride);
 	if (samples == NULL && stride < n*w)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "Illegal -ve stride for pixmap without data");
+	if (n > FZ_MAX_COLORS)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Illegal number of colorants");
 
 	pix = fz_malloc_struct(ctx, fz_pixmap);
 	FZ_INIT_STORABLE(pix, 1, fz_drop_pixmap_imp);
@@ -78,11 +103,11 @@ fz_new_pixmap_with_data(fz_context *ctx, fz_colorspace *colorspace, int w, int h
 		{
 			if (pix->stride - 1 > INT_MAX / pix->n)
 				fz_throw(ctx, FZ_ERROR_GENERIC, "overly wide image");
-			pix->samples = fz_malloc_array(ctx, pix->h, pix->stride);
+			pix->samples = fz_malloc(ctx, pix->h * pix->stride);
 		}
 		fz_catch(ctx)
 		{
-			fz_drop_colorspace(ctx, colorspace);
+			fz_drop_colorspace(ctx, pix->colorspace);
 			fz_free(ctx, pix);
 			fz_rethrow(ctx);
 		}
@@ -92,6 +117,23 @@ fz_new_pixmap_with_data(fz_context *ctx, fz_colorspace *colorspace, int w, int h
 	return pix;
 }
 
+/*
+	Create a new pixmap, with its origin at (0,0)
+
+	cs: The colorspace to use for the pixmap, or NULL for an alpha
+	plane/mask.
+
+	w: The width of the pixmap (in pixels)
+
+	h: The height of the pixmap (in pixels)
+
+	seps: Details of separations.
+
+	alpha: 0 for no alpha, 1 for alpha.
+
+	Returns a pointer to the new pixmap. Throws exception on failure to
+	allocate.
+*/
 fz_pixmap *
 fz_new_pixmap(fz_context *ctx, fz_colorspace *colorspace, int w, int h, fz_separations *seps, int alpha)
 {
@@ -102,31 +144,86 @@ fz_new_pixmap(fz_context *ctx, fz_colorspace *colorspace, int w, int h, fz_separ
 	return fz_new_pixmap_with_data(ctx, colorspace, w, h, seps, alpha, stride, NULL);
 }
 
+/*
+	Create a pixmap of a given size,
+	location and pixel format.
+
+	The bounding box specifies the size of the created pixmap and
+	where it will be located. The colorspace determines the number
+	of components per pixel. Alpha is always present. Pixmaps are
+	reference counted, so drop references using fz_drop_pixmap.
+
+	colorspace: Colorspace format used for the created pixmap. The
+	pixmap will keep a reference to the colorspace.
+
+	bbox: Bounding box specifying location/size of created pixmap.
+
+	seps: Details of separations.
+
+	alpha: 0 for no alpha, 1 for alpha.
+
+	Returns a pointer to the new pixmap. Throws exception on failure to
+	allocate.
+*/
 fz_pixmap *
-fz_new_pixmap_with_bbox(fz_context *ctx, fz_colorspace *colorspace, const fz_irect *r, fz_separations *seps, int alpha)
+fz_new_pixmap_with_bbox(fz_context *ctx, fz_colorspace *colorspace, fz_irect bbox, fz_separations *seps, int alpha)
 {
 	fz_pixmap *pixmap;
-	pixmap = fz_new_pixmap(ctx, colorspace, r->x1 - r->x0, r->y1 - r->y0, seps, alpha);
-	pixmap->x = r->x0;
-	pixmap->y = r->y0;
+	pixmap = fz_new_pixmap(ctx, colorspace, bbox.x1 - bbox.x0, bbox.y1 - bbox.y0, seps, alpha);
+	pixmap->x = bbox.x0;
+	pixmap->y = bbox.y0;
 	return pixmap;
 }
 
+/*
+	Create a pixmap of a given size,
+	location and pixel format, using the supplied data block.
+
+	The bounding box specifies the size of the created pixmap and
+	where it will be located. The colorspace determines the number
+	of components per pixel. Alpha is always present. Pixmaps are
+	reference counted, so drop references using fz_drop_pixmap.
+
+	colorspace: Colorspace format used for the created pixmap. The
+	pixmap will keep a reference to the colorspace.
+
+	rect: Bounding box specifying location/size of created pixmap.
+
+	seps: Details of separations.
+
+	alpha: Number of alpha planes (0 or 1).
+
+	samples: The data block to keep the samples in.
+
+	Returns a pointer to the new pixmap. Throws exception on failure to
+	allocate.
+*/
 fz_pixmap *
-fz_new_pixmap_with_bbox_and_data(fz_context *ctx, fz_colorspace *colorspace, const fz_irect *r, fz_separations *seps, int alpha, unsigned char *samples)
+fz_new_pixmap_with_bbox_and_data(fz_context *ctx, fz_colorspace *colorspace, fz_irect bbox, fz_separations *seps, int alpha, unsigned char *samples)
 {
-	int w = r->x1 - r->x0;
+	int w = bbox.x1 - bbox.x0;
 	int stride;
 	int s = fz_count_active_separations(ctx, seps);
 	fz_pixmap *pixmap;
 	if (!colorspace && s == 0) alpha = 1;
 	stride = (fz_colorspace_n(ctx, colorspace) + s + alpha) * w;
-	pixmap = fz_new_pixmap_with_data(ctx, colorspace, w, r->y1 - r->y0, seps, alpha, stride, samples);
-	pixmap->x = r->x0;
-	pixmap->y = r->y0;
+	pixmap = fz_new_pixmap_with_data(ctx, colorspace, w, bbox.y1 - bbox.y0, seps, alpha, stride, samples);
+	pixmap->x = bbox.x0;
+	pixmap->y = bbox.y0;
 	return pixmap;
 }
 
+/*
+	Create a new pixmap that represents
+	a subarea of the specified pixmap. A reference is taken to his
+	pixmap that will be dropped on destruction.
+
+	The supplied rectangle must be wholly contained within the original
+	pixmap.
+
+	Returns a pointer to the new pixmap. Throws exception on failure to
+	allocate.
+*/
 fz_pixmap *fz_new_pixmap_from_pixmap(fz_context *ctx, fz_pixmap *pixmap, const fz_irect *rect)
 {
 	fz_irect local_rect;
@@ -162,26 +259,43 @@ fz_pixmap *fz_new_pixmap_from_pixmap(fz_context *ctx, fz_pixmap *pixmap, const f
 	return subpix;
 }
 
-fz_irect *
-fz_pixmap_bbox(fz_context *ctx, const fz_pixmap *pix, fz_irect *bbox)
+fz_pixmap *fz_clone_pixmap(fz_context *ctx, fz_pixmap *old)
 {
-	bbox->x0 = pix->x;
-	bbox->y0 = pix->y;
-	bbox->x1 = pix->x + pix->w;
-	bbox->y1 = pix->y + pix->h;
+	fz_pixmap *pix = fz_new_pixmap_with_bbox(ctx, old->colorspace, fz_make_irect(old->x, old->y, old->w, old->h), old->seps, old->alpha);
+	memcpy(pix->samples, old->samples, pix->stride * pix->h);
+	return pix;
+}
+
+/*
+	Return the bounding box for a pixmap.
+*/
+fz_irect
+fz_pixmap_bbox(fz_context *ctx, const fz_pixmap *pix)
+{
+	fz_irect bbox;
+	bbox.x0 = pix->x;
+	bbox.y0 = pix->y;
+	bbox.x1 = pix->x + pix->w;
+	bbox.y1 = pix->y + pix->h;
 	return bbox;
 }
 
-fz_irect *
-fz_pixmap_bbox_no_ctx(const fz_pixmap *pix, fz_irect *bbox)
+fz_irect
+fz_pixmap_bbox_no_ctx(const fz_pixmap *pix)
 {
-	bbox->x0 = pix->x;
-	bbox->y0 = pix->y;
-	bbox->x1 = pix->x + pix->w;
-	bbox->y1 = pix->y + pix->h;
+	fz_irect bbox;
+	bbox.x0 = pix->x;
+	bbox.y0 = pix->y;
+	bbox.x1 = pix->x + pix->w;
+	bbox.y1 = pix->y + pix->h;
 	return bbox;
 }
 
+/*
+	Return the colorspace of a pixmap
+
+	Returns colorspace.
+*/
 fz_colorspace *
 fz_pixmap_colorspace(fz_context *ctx, fz_pixmap *pix)
 {
@@ -190,60 +304,100 @@ fz_pixmap_colorspace(fz_context *ctx, fz_pixmap *pix)
 	return pix->colorspace;
 }
 
+/*
+	Return the x value of the pixmap in pixels.
+*/
 int
 fz_pixmap_x(fz_context *ctx, fz_pixmap *pix)
 {
 	return pix->x;
 }
 
+/*
+	Return the y value of the pixmap in pixels.
+*/
 int
 fz_pixmap_y(fz_context *ctx, fz_pixmap *pix)
 {
 	return pix->y;
 }
 
+/*
+	Return the width of the pixmap in pixels.
+*/
 int
 fz_pixmap_width(fz_context *ctx, fz_pixmap *pix)
 {
 	return pix->w;
 }
 
+/*
+	Return the height of the pixmap in pixels.
+*/
 int
 fz_pixmap_height(fz_context *ctx, fz_pixmap *pix)
 {
 	return pix->h;
 }
 
+/*
+	Return the number of components in a pixmap.
+
+	Returns the number of components (including spots and alpha).
+*/
 int
 fz_pixmap_components(fz_context *ctx, fz_pixmap *pix)
 {
 	return pix->n;
 }
 
+/*
+	Return the number of colorants in a pixmap.
+
+	Returns the number of colorants (components, less any spots and alpha).
+*/
 int
 fz_pixmap_colorants(fz_context *ctx, fz_pixmap *pix)
 {
 	return pix->n - pix->alpha - pix->s;
 }
 
+/*
+	Return the number of spots in a pixmap.
+
+	Returns the number of spots (components, less colorants and alpha). Does not throw exceptions.
+*/
 int
 fz_pixmap_spots(fz_context *ctx, fz_pixmap *pix)
 {
 	return pix->s;
 }
 
+/*
+	Return the number of alpha planes in a pixmap.
+
+	Returns the number of alphas. Does not throw exceptions.
+*/
 int
 fz_pixmap_alpha(fz_context *ctx, fz_pixmap *pix)
 {
 	return pix->alpha;
 }
 
+/*
+	Return the number of bytes in a row in the pixmap.
+*/
 int
 fz_pixmap_stride(fz_context *ctx, fz_pixmap *pix)
 {
 	return pix->stride;
 }
 
+/*
+	Returns a pointer to the pixel data of a pixmap.
+
+	Returns the pointer.
+*/
 unsigned char *
 fz_pixmap_samples(fz_context *ctx, fz_pixmap *pix)
 {
@@ -475,6 +629,12 @@ clear_cmyk_bitmap(unsigned char *samples, int w, int h, int spots, int stride, i
 	}
 }
 
+/*
+	Sets all components (including alpha) of
+	all pixels in a pixmap to 0.
+
+	pix: The pixmap to clear.
+*/
 void
 fz_clear_pixmap(fz_context *ctx, fz_pixmap *pix)
 {
@@ -494,7 +654,7 @@ fz_clear_pixmap(fz_context *ctx, fz_pixmap *pix)
 			s += pix->stride;
 		}
 	}
-	else
+	else if (pix->s == 0)
 	{
 		while (h--)
 		{
@@ -502,8 +662,46 @@ fz_clear_pixmap(fz_context *ctx, fz_pixmap *pix)
 			s += pix->stride;
 		}
 	}
+	else
+	{
+		/* Horrible, slow case: additive with spots */
+		int w = stride/pix->n;
+		int spots = pix->s;
+		int colorants = pix->n - spots; /* We know there is no alpha */
+		while (h--)
+		{
+			int w2 = w;
+			while (w2--)
+			{
+				int i = colorants;
+				do
+				{
+					*s++ = 0xff;
+					i--;
+				}
+				while (i != 0);
+
+				i = spots;
+				do
+				{
+					*s++ = 0;
+					i--;
+				}
+				while (i != 0);
+			}
+		}
+	}
 }
 
+/*
+	Clears a pixmap with the given value.
+
+	pix: The pixmap to clear.
+
+	value: Values in the range 0 to 255 are valid. Each component
+	sample for each pixel in the pixmap will be set to this value,
+	while alpha will always be set to 255 (non-transparent).
+*/
 /* This function is horrible, and should be removed from the
  * API and replaced with a less magic one. */
 void
@@ -561,8 +759,11 @@ fz_clear_pixmap_with_value(fz_context *ctx, fz_pixmap *pix, int value)
 	}
 }
 
+/*
+	Fill pixmap with solid color.
+*/
 void
-fz_fill_pixmap_with_color(fz_context *ctx, fz_pixmap *pix, fz_colorspace *colorspace, float *color, const fz_color_params *color_params)
+fz_fill_pixmap_with_color(fz_context *ctx, fz_pixmap *pix, fz_colorspace *colorspace, float *color, fz_color_params color_params)
 {
 	float colorfv[FZ_MAX_COLORS];
 	unsigned char colorbv[FZ_MAX_COLORS];
@@ -571,7 +772,7 @@ fz_fill_pixmap_with_color(fz_context *ctx, fz_pixmap *pix, fz_colorspace *colors
 	n = fz_colorspace_n(ctx, pix->colorspace);
 	a = pix->alpha;
 	s = pix->s;
-	fz_convert_color(ctx, color_params, NULL, pix->colorspace, colorfv, colorspace, color);
+	fz_convert_color(ctx, colorspace, color, pix->colorspace, colorfv, NULL, color_params);
 	for (i = 0; i < n; ++i)
 		colorbv[i] = colorfv[i] * 255;
 
@@ -593,25 +794,23 @@ fz_fill_pixmap_with_color(fz_context *ctx, fz_pixmap *pix, fz_colorspace *colors
 }
 
 void
-fz_copy_pixmap_rect(fz_context *ctx, fz_pixmap *dest, fz_pixmap *src, const fz_irect *b, const fz_default_colorspaces *default_cs)
+fz_copy_pixmap_rect(fz_context *ctx, fz_pixmap *dest, fz_pixmap *src, fz_irect b, const fz_default_colorspaces *default_cs)
 {
 	unsigned char *srcp;
 	unsigned char *destp;
 	int y, w, destspan, srcspan;
-	fz_irect local_b, bb;
 
-	local_b = *b;
-	fz_intersect_irect(&local_b, fz_pixmap_bbox(ctx, dest, &bb));
-	fz_intersect_irect(&local_b, fz_pixmap_bbox(ctx, src, &bb));
-	w = local_b.x1 - local_b.x0;
-	y = local_b.y1 - local_b.y0;
+	b = fz_intersect_irect(b, fz_pixmap_bbox(ctx, dest));
+	b = fz_intersect_irect(b, fz_pixmap_bbox(ctx, src));
+	w = b.x1 - b.x0;
+	y = b.y1 - b.y0;
 	if (w <= 0 || y <= 0)
 		return;
 
 	srcspan = src->stride;
-	srcp = src->samples + (unsigned int)(srcspan * (local_b.y0 - src->y) + src->n * (local_b.x0 - src->x));
+	srcp = src->samples + (unsigned int)(srcspan * (b.y0 - src->y) + src->n * (b.x0 - src->x));
 	destspan = dest->stride;
-	destp = dest->samples + (unsigned int)(destspan * (local_b.y0 - dest->y) + dest->n * (local_b.x0 - dest->x));
+	destp = dest->samples + (unsigned int)(destspan * (b.y0 - dest->y) + dest->n * (b.x0 - dest->x));
 
 	if (src->n == dest->n)
 	{
@@ -626,35 +825,41 @@ fz_copy_pixmap_rect(fz_context *ctx, fz_pixmap *dest, fz_pixmap *src, const fz_i
 	}
 	else
 	{
-		fz_pixmap_converter *pc = fz_lookup_pixmap_converter(ctx, dest->colorspace, src->colorspace);
 		fz_pixmap fake_src = *src;
-
-		fake_src.x = local_b.x0;
-		fake_src.y = local_b.y0;
+		fake_src.x = b.x0;
+		fake_src.y = b.y0;
 		fake_src.w = w;
 		fake_src.h = y;
 		fake_src.samples = srcp;
-
-		pc(ctx, dest, &fake_src, NULL, default_cs, fz_default_color_params(ctx), 0);
+		fz_convert_pixmap_samples(ctx, dest, &fake_src, NULL, default_cs, fz_default_color_params, 0);
 	}
 }
 
+/*
+	Clears a subrect of a pixmap with the given value.
+
+	pix: The pixmap to clear.
+
+	value: Values in the range 0 to 255 are valid. Each component
+	sample for each pixel in the pixmap will be set to this value,
+	while alpha will always be set to 255 (non-transparent).
+
+	r: the rectangle.
+*/
 void
-fz_clear_pixmap_rect_with_value(fz_context *ctx, fz_pixmap *dest, int value, const fz_irect *b)
+fz_clear_pixmap_rect_with_value(fz_context *ctx, fz_pixmap *dest, int value, fz_irect b)
 {
 	unsigned char *destp;
 	int x, y, w, k, destspan;
-	fz_irect bb;
-	fz_irect local_b = *b;
 
-	fz_intersect_irect(&local_b, fz_pixmap_bbox(ctx, dest, &bb));
-	w = local_b.x1 - local_b.x0;
-	y = local_b.y1 - local_b.y0;
+	b = fz_intersect_irect(b, fz_pixmap_bbox(ctx, dest));
+	w = b.x1 - b.x0;
+	y = b.y1 - b.y0;
 	if (w <= 0 || y <= 0)
 		return;
 
 	destspan = dest->stride;
-	destp = dest->samples + (unsigned int)(destspan * (local_b.y0 - dest->y) + dest->n * (local_b.x0 - dest->x));
+	destp = dest->samples + (unsigned int)(destspan * (b.y0 - dest->y) + dest->n * (b.x0 - dest->x));
 
 	/* CMYK needs special handling (and potentially any other subtractive colorspaces) */
 	if (fz_colorspace_n(ctx, dest->colorspace) == 4)
@@ -727,42 +932,16 @@ fz_premultiply_pixmap(fz_context *ctx, fz_pixmap *pix)
 	}
 }
 
-void
-fz_unmultiply_pixmap(fz_context *ctx, fz_pixmap *pix)
-{
-	unsigned char *s = pix->samples;
-	int a, inva;
-	int k, x, y;
-	int stride = pix->stride - pix->w * pix->n;
-
-	if (!pix->alpha)
-		return;
-
-	for (y = 0; y < pix->h; y++)
-	{
-		for (x = 0; x < pix->w; x++)
-		{
-			a = s[pix->n - 1];
-			inva = a ? 255 * 256 / a : 0;
-			for (k = 0; k < pix->n - 1; k++)
-				s[k] = (s[k] * inva) >> 8;
-			s += pix->n;
-		}
-		s += stride;
-	}
-}
-
 fz_pixmap *
 fz_alpha_from_gray(fz_context *ctx, fz_pixmap *gray)
 {
 	fz_pixmap *alpha;
 	unsigned char *sp, *dp;
 	int w, h, sstride, dstride;
-	fz_irect bbox;
 
 	assert(gray->n == 1);
 
-	alpha = fz_new_pixmap_with_bbox(ctx, NULL, fz_pixmap_bbox(ctx, gray, &bbox), 0, 1);
+	alpha = fz_new_pixmap_with_bbox(ctx, NULL, fz_pixmap_bbox(ctx, gray), 0, 1);
 	dp = alpha->samples;
 	dstride = alpha->stride;
 	sp = gray->samples;
@@ -780,22 +959,39 @@ fz_alpha_from_gray(fz_context *ctx, fz_pixmap *gray)
 	return alpha;
 }
 
+/*
+	Tint all the pixels in an RGB, BGR, or Gray pixmap.
+
+	black: Map black to this hexadecimal RGB color.
+	white: Map white to this hexadecimal RGB color.
+*/
 void
-fz_tint_pixmap(fz_context *ctx, fz_pixmap *pix, int r, int g, int b)
+fz_tint_pixmap(fz_context *ctx, fz_pixmap *pix, int black, int white)
 {
 	unsigned char *s = pix->samples;
 	int n = pix->n;
 	int x, y, save;
+	int rb = (black>>16)&255;
+	int gb = (black>>8)&255;
+	int bb = (black)&255;
+	int rw = (white>>16)&255;
+	int gw = (white>>8)&255;
+	int bw = (white)&255;
+	int rm = (rw - rb);
+	int gm = (gw - gb);
+	int bm = (bw - bb);
 
 	switch (fz_colorspace_type(ctx, pix->colorspace))
 	{
 	case FZ_COLORSPACE_GRAY:
-		g = (r + g + b) / 3;
+		gw = (rw + gw + bw) / 3;
+		gb = (rb + gb + bb) / 3;
+		gm = gw - gb;
 		for (y = 0; y < pix->h; y++)
 		{
 			for (x = 0; x < pix->w; x++)
 			{
-				*s = fz_mul255(*s, g);
+				*s = gb + fz_mul255(*s, gm);
 				s += n;
 			}
 			s += pix->stride - pix->w * n;
@@ -803,18 +999,17 @@ fz_tint_pixmap(fz_context *ctx, fz_pixmap *pix, int r, int g, int b)
 		break;
 
 	case FZ_COLORSPACE_BGR:
-		save = r;
-		r = b;
-		b = save;
+		save = rm; rm = bm; bm = save;
+		save = rb; rb = bb; bb = save;
 		/* fall through */
 	case FZ_COLORSPACE_RGB:
 		for (y = 0; y < pix->h; y++)
 		{
 			for (x = 0; x < pix->w; x++)
 			{
-				s[0] = fz_mul255(s[0], r);
-				s[1] = fz_mul255(s[1], g);
-				s[2] = fz_mul255(s[2], b);
+				s[0] = rb + fz_mul255(s[0], rm);
+				s[1] = gb + fz_mul255(s[1], gm);
+				s[2] = bb + fz_mul255(s[2], bm);
 				s += n;
 			}
 			s += pix->stride - pix->w * n;
@@ -827,6 +1022,59 @@ fz_tint_pixmap(fz_context *ctx, fz_pixmap *pix, int r, int g, int b)
 	}
 }
 
+/* Invert luminance in RGB pixmap, but keep the colors as is. */
+static void invert_luminance_rgb(unsigned char *rgb)
+{
+	int r, g, b, y, u, v, c, d, e;
+
+	/* Convert to YUV */
+	r = rgb[0];
+	g = rgb[1];
+	b = rgb[2];
+	y = ((66 * r + 129 * g + 25 * b + 128) >> 8) + 16;
+	u = ((-38 * r - 74 * g + 112 * b + 128) >> 8) + 128;
+	v = ((112 * r - 94 * g - 18 * b + 128) >> 8) + 128;
+
+	/* Invert luminance */
+	y = 255 - y;
+
+	/* Convert to RGB */
+	c = y - 16;
+	d = u - 128;
+	e = v - 128;
+	r = (298 * c + 409 * e + 128) >> 8;
+	g = (298 * c - 100 * d - 208 * e + 128) >> 8;
+	b = (298 * c + 516 * d + 128) >> 8;
+
+	rgb[0] = r > 255 ? 255 : r < 0 ? 0 : r;
+	rgb[1] = g > 255 ? 255 : g < 0 ? 0 : g;
+	rgb[2] = b > 255 ? 255 : b < 0 ? 0 : b;
+}
+
+void
+fz_invert_pixmap_luminance(fz_context *ctx, fz_pixmap *pix)
+{
+	unsigned char *s = pix->samples;
+	int x, y, n = pix->n;
+
+	if (pix->colorspace->type != FZ_COLORSPACE_RGB)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "can only invert luminance of RGB pixmaps");
+
+	for (y = 0; y < pix->h; y++)
+	{
+		for (x = 0; x < pix->w; x++)
+		{
+			invert_luminance_rgb(s);
+			s += n;
+		}
+		s += pix->stride - pix->w * n;
+	}
+}
+
+/*
+	Invert all the pixels in a pixmap. All components
+	of all pixels are inverted (except alpha, which is unchanged).
+*/
 void
 fz_invert_pixmap(fz_context *ctx, fz_pixmap *pix)
 {
@@ -847,15 +1095,20 @@ fz_invert_pixmap(fz_context *ctx, fz_pixmap *pix)
 	}
 }
 
-void fz_invert_pixmap_rect(fz_context *ctx, fz_pixmap *image, const fz_irect *rect)
+/*
+	Invert all the pixels in a given rectangle of a
+	pixmap. All components of all pixels in the rectangle are inverted
+	(except alpha, which is unchanged).
+*/
+void fz_invert_pixmap_rect(fz_context *ctx, fz_pixmap *image, fz_irect rect)
 {
 	unsigned char *p;
 	int x, y, n;
 
-	int x0 = fz_clampi(rect->x0 - image->x, 0, image->w);
-	int x1 = fz_clampi(rect->x1 - image->x, 0, image->w);
-	int y0 = fz_clampi(rect->y0 - image->y, 0, image->h);
-	int y1 = fz_clampi(rect->y1 - image->y, 0, image->h);
+	int x0 = fz_clampi(rect.x0 - image->x, 0, image->w);
+	int x1 = fz_clampi(rect.x1 - image->x, 0, image->w);
+	int y0 = fz_clampi(rect.y0 - image->y, 0, image->h);
+	int y1 = fz_clampi(rect.y1 - image->y, 0, image->h);
 
 	for (y = y0; y < y1; y++)
 	{
@@ -869,6 +1122,12 @@ void fz_invert_pixmap_rect(fz_context *ctx, fz_pixmap *image, const fz_irect *re
 	}
 }
 
+/*
+	Apply gamma correction to a pixmap. All components
+	of all pixels are modified (except alpha, which is unchanged).
+
+	gamma: The gamma value to apply; 1.0 for no change.
+*/
 void
 fz_gamma_pixmap(fz_context *ctx, fz_pixmap *pix, float gamma)
 {
@@ -901,17 +1160,33 @@ fz_pixmap_size(fz_context *ctx, fz_pixmap * pix)
 	return sizeof(*pix) + pix->n * pix->w * pix->h;
 }
 
+/*
+	Convert an existing pixmap to a desired
+	colorspace. Other properties of the pixmap, such as resolution
+	and position are copied to the converted pixmap.
+
+	pix: The pixmap to convert.
+
+	default_cs: If NULL pix->colorspace is used. It is possible that the data
+	may need to be interpreted as one of the color spaces in default_cs.
+
+	cs_des: Desired colorspace, may be NULL to denote alpha-only.
+
+	prf: Proofing color space through which we need to convert.
+
+	color_params: Parameters that may be used in conversion (e.g. ri).
+
+	keep_alpha: If 0 any alpha component is removed, otherwise
+	alpha is kept if present in the pixmap.
+*/
 fz_pixmap *
-fz_convert_pixmap(fz_context *ctx, fz_pixmap *pix, fz_colorspace *ds, fz_colorspace *prf, fz_default_colorspaces *default_cs, const fz_color_params *color_params, int keep_alpha)
+fz_convert_pixmap(fz_context *ctx, fz_pixmap *pix, fz_colorspace *ds, fz_colorspace *prf, fz_default_colorspaces *default_cs, fz_color_params color_params, int keep_alpha)
 {
 	fz_pixmap *cvt;
 
 	if (!ds && !keep_alpha)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot both throw away and keep alpha");
 
-	if (color_params == NULL)
-		color_params = fz_default_color_params(ctx);
-
 	cvt = fz_new_pixmap(ctx, ds, pix->w, pix->h, pix->seps, keep_alpha && pix->alpha);
 
 	cvt->xres = pix->xres;
@@ -925,8 +1200,7 @@ fz_convert_pixmap(fz_context *ctx, fz_pixmap *pix, fz_colorspace *ds, fz_colorsp
 
 	fz_try(ctx)
 	{
-		fz_pixmap_converter *pc = fz_lookup_pixmap_converter(ctx, ds, pix->colorspace);
-		pc(ctx, cvt, pix, prf, default_cs, color_params, 1);
+		fz_convert_pixmap_samples(ctx, pix, cvt, prf, default_cs, color_params, 1);
 	}
 	fz_catch(ctx)
 	{
@@ -1305,9 +1579,14 @@ fz_subsample_pixmap(fz_context *ctx, fz_pixmap *tile, int factor)
 	tile->w = dst_w;
 	tile->h = dst_h;
 	tile->stride = dst_w * n;
-	tile->samples = fz_resize_array(ctx, tile->samples, dst_w * n, dst_h);
+	if (dst_h > INT_MAX / (dst_w * n))
+		fz_throw(ctx, FZ_ERROR_MEMORY, "pixmap too large");
+	tile->samples = fz_realloc(ctx, tile->samples, dst_h * dst_w * n);
 }
 
+/*
+	Set the pixels per inch resolution of the pixmap.
+*/
 void
 fz_set_pixmap_resolution(fz_context *ctx, fz_pixmap *pix, int xres, int yres)
 {
@@ -1315,6 +1594,9 @@ fz_set_pixmap_resolution(fz_context *ctx, fz_pixmap *pix, int xres, int yres)
 	pix->yres = yres;
 }
 
+/*
+	Return the md5 digest for a pixmap
+*/
 void
 fz_md5_pixmap(fz_context *ctx, fz_pixmap *pix, unsigned char digest[16])
 {
@@ -1362,3 +1644,158 @@ int fz_valgrind_pixmap(const fz_pixmap *pix)
 	return total;
 }
 #endif /* HAVE_VALGRIND */
+
+/*
+ * Convert pixmap from indexed to base colorspace.
+ */
+fz_pixmap *
+fz_convert_indexed_pixmap_to_base(fz_context *ctx, const fz_pixmap *src)
+{
+	fz_pixmap *dst;
+	fz_colorspace *base;
+	const unsigned char *s;
+	unsigned char *d;
+	int y, x, k, n, high;
+	unsigned char *lookup;
+	int s_line_inc, d_line_inc;
+
+	if (src->colorspace->type != FZ_COLORSPACE_INDEXED)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot convert non-indexed pixmap");
+	if (src->n != 1 + src->alpha)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot convert indexed pixmap mis-matching components");
+
+	base = src->colorspace->u.indexed.base;
+	high = src->colorspace->u.indexed.high;
+	lookup = src->colorspace->u.indexed.lookup;
+	n = base->n;
+
+	dst = fz_new_pixmap_with_bbox(ctx, base, fz_pixmap_bbox(ctx, src), src->seps, src->alpha);
+	s = src->samples;
+	d = dst->samples;
+	s_line_inc = src->stride - src->w * src->n;
+	d_line_inc = dst->stride - dst->w * dst->n;
+
+	if (src->alpha)
+	{
+		for (y = 0; y < src->h; y++)
+		{
+			for (x = 0; x < src->w; x++)
+			{
+				int v = *s++;
+				int a = *s++;
+				int aa = a + (a>>7);
+				v = fz_mini(v, high);
+				for (k = 0; k < n; k++)
+					*d++ = (aa * lookup[v * n + k] + 128)>>8;
+				*d++ = a;
+			}
+			s += s_line_inc;
+			d += d_line_inc;
+		}
+	}
+	else
+	{
+		for (y = 0; y < src->h; y++)
+		{
+			for (x = 0; x < src->w; x++)
+			{
+				int v = *s++;
+				v = fz_mini(v, high);
+				for (k = 0; k < n; k++)
+					*d++ = lookup[v * n + k];
+			}
+			s += s_line_inc;
+			d += d_line_inc;
+		}
+	}
+
+	if (src->flags & FZ_PIXMAP_FLAG_INTERPOLATE)
+		dst->flags |= FZ_PIXMAP_FLAG_INTERPOLATE;
+	else
+		dst->flags &= ~FZ_PIXMAP_FLAG_INTERPOLATE;
+
+	return dst;
+}
+
+/*
+ * Convert pixmap from DeviceN/Separation to base colorspace.
+ */
+fz_pixmap *
+fz_convert_separation_pixmap_to_base(fz_context *ctx, const fz_pixmap *src)
+{
+	fz_pixmap *dst;
+	fz_colorspace *ss, *base;
+	const unsigned char *s;
+	unsigned char *d;
+	int y, x, k, sn, bn, a;
+	float src_v[FZ_MAX_COLORS];
+	float base_v[FZ_MAX_COLORS];
+	int s_line_inc, d_line_inc;
+
+	ss = src->colorspace;
+
+	if (ss->type != FZ_COLORSPACE_SEPARATION)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot expand non-separation pixmap");
+	if (src->n != ss->n + src->alpha)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot expand separation pixmap mis-matching alpha channel");
+
+	base = ss->u.separation.base;
+	dst = fz_new_pixmap_with_bbox(ctx, base, fz_pixmap_bbox(ctx, src), src->seps, src->alpha);
+	fz_clear_pixmap(ctx, dst);
+	fz_try(ctx)
+	{
+		s = src->samples;
+		d = dst->samples;
+		s_line_inc = src->stride - src->w * src->n;
+		d_line_inc = dst->stride - dst->w * dst->n;
+		sn = ss->n;
+		bn = base->n;
+
+		if (src->alpha)
+		{
+			for (y = 0; y < src->h; y++)
+			{
+				for (x = 0; x < src->w; x++)
+				{
+					for (k = 0; k < sn; ++k)
+						src_v[k] = *s++ / 255.0f;
+					a = *s++;
+					ss->u.separation.eval(ctx, ss->u.separation.tint, src_v, sn, base_v, bn);
+					for (k = 0; k < bn; ++k)
+						*d++ = base_v[k] * 255.0f;
+					*d++ = a;
+				}
+				s += s_line_inc;
+				d += d_line_inc;
+			}
+		}
+		else
+		{
+			for (y = 0; y < src->h; y++)
+			{
+				for (x = 0; x < src->w; x++)
+				{
+					for (k = 0; k < sn; ++k)
+						src_v[k] = *s++ / 255.0f;
+					ss->u.separation.eval(ctx, ss->u.separation.tint, src_v, sn, base_v, bn);
+					for (k = 0; k < bn; ++k)
+						*d++ = base_v[k] * 255.0f;
+				}
+				s += s_line_inc;
+				d += d_line_inc;
+			}
+		}
+
+		if (src->flags & FZ_PIXMAP_FLAG_INTERPOLATE)
+			dst->flags |= FZ_PIXMAP_FLAG_INTERPOLATE;
+		else
+			dst->flags &= ~FZ_PIXMAP_FLAG_INTERPOLATE;
+	}
+	fz_catch(ctx)
+	{
+		fz_drop_pixmap(ctx, dst);
+		fz_rethrow(ctx);
+	}
+
+	return dst;
+}
