@@ -241,7 +241,7 @@ fz_separations *fz_clone_separations_for_overprint(fz_context *ctx, fz_separatio
 	different separation results.
 */
 fz_pixmap *
-fz_clone_pixmap_area_with_different_seps(fz_context *ctx, fz_pixmap *src, const fz_irect *bbox, fz_colorspace *dcs, fz_separations *dseps, const fz_color_params *color_params, fz_default_colorspaces *default_cs)
+fz_clone_pixmap_area_with_different_seps(fz_context *ctx, fz_pixmap *src, const fz_irect *bbox, fz_colorspace *dcs, fz_separations *dseps, fz_color_params color_params, fz_default_colorspaces *default_cs)
 {
 	fz_irect local_bbox;
 	fz_pixmap *dst, *pix;
@@ -262,7 +262,7 @@ fz_clone_pixmap_area_with_different_seps(fz_context *ctx, fz_pixmap *src, const
 		dst->flags &= ~FZ_PIXMAP_FLAG_INTERPOLATE;
 
 	fz_try(ctx)
-		pix = fz_copy_pixmap_area_converting_seps(ctx, dst, src, color_params, NULL, default_cs);
+		pix = fz_copy_pixmap_area_converting_seps(ctx, src, dst, NULL, color_params, default_cs);
 	fz_catch(ctx)
 	{
 		fz_drop_pixmap(ctx, dst);
@@ -276,7 +276,7 @@ fz_clone_pixmap_area_with_different_seps(fz_context *ctx, fz_pixmap *src, const
 	We assume that we never map from a DeviceN space to another DeviceN space here.
  */
 fz_pixmap *
-fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *src, const fz_color_params *color_params, fz_colorspace *prf, fz_default_colorspaces *default_cs)
+fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *src, fz_pixmap *dst, fz_colorspace *prf, fz_color_params color_params, fz_default_colorspaces *default_cs)
 {
 	int dw = dst->w;
 	int dh = dst->h;
@@ -391,7 +391,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 		/* Now map the colorants down. */
 		n = fz_colorspace_n(ctx, src->colorspace);
 
-		fz_find_color_converter(ctx, &cc, proof_cs, dst->colorspace, src->colorspace, color_params);
+		fz_find_color_converter(ctx, &cc, src->colorspace, dst->colorspace, proof_cs, color_params);
 
 		fz_try(ctx)
 		{
@@ -496,7 +496,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 					/* Src component i is not mapped. We need to convert that down. */
 					memset(colors, 0, sizeof(float) * n);
 					colors[i] = 1;
-					cc.convert(ctx, &cc, convert, colors);
+					cc.convert(ctx, &cc, colors, convert);
 
 					if (fz_colorspace_is_subtractive(ctx, dst->colorspace))
 					{
@@ -603,7 +603,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 
 						memset(colors, 0, sizeof(float) * n);
 						colors[i] = 1;
-						cc.convert(ctx, &cc, convert, colors);
+						cc.convert(ctx, &cc, colors, convert);
 
 						if (sa)
 						{
@@ -656,7 +656,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 							{
 								for (j = 0; j < n; j++)
 									colors[j] = mapped[j] ? 0 : sd[j] / 255.0f;
-								cc.convert(ctx, &cc, convert, colors);
+								cc.convert(ctx, &cc, colors, convert);
 
 								for (j = 0; j < dc; j++)
 									dd[j] = fz_clampi(255 * convert[j], 0, 255);
@@ -677,7 +677,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 								float inva = 1.0f/a;
 								for (j = 0; j < n; j++)
 									colors[j] = mapped[j] ? 0 : sd[j] * inva;
-								cc.convert(ctx, &cc, convert, colors);
+								cc.convert(ctx, &cc, colors, convert);
 
 								for (j = 0; j < dc; j++)
 									dd[j] = fz_clampi(a * convert[j], 0, a);
@@ -700,7 +700,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 	else
 	{
 		/* Use a standard pixmap converter to convert the process + alpha. */
-		fz_convert_pixmap_samples(ctx, dst, src, proof_cs, default_cs, NULL, 0);
+		fz_convert_pixmap_samples(ctx, src, dst, proof_cs, default_cs, fz_default_color_params, 0);
 
 		/* And handle the spots ourselves. First make a map of what spots go where. */
 		/* We want to set it up so that:
@@ -778,7 +778,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 					continue;
 				i++;
 				/* Src spot m (the i'th one) is not mapped. We need to convert that down. */
-				fz_separation_equivalent(ctx, sseps, m, color_params, dst->colorspace, proof_cs, convert);
+				fz_separation_equivalent(ctx, sseps, m, dst->colorspace, convert, proof_cs, color_params);
 
 				if (fz_colorspace_is_subtractive(ctx, dst->colorspace))
 				{
@@ -795,7 +795,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 									unsigned char v = sd[i];
 									if (v != 0)
 									{
-										unsigned char a = sd[sc];
+										unsigned char a = sd[ss];
 										for (k = 0; k < dc; k++)
 											dd[k] = fz_clampi(dd[k] + v * convert[k], 0, a);
 									}
@@ -837,7 +837,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 									unsigned char v = 0xff - sd[i];
 									if (v != 0)
 									{
-										unsigned char a = sd[sc];
+										unsigned char a = sd[ss];
 										for (k = 0; k < dc; k++)
 											dd[k] = fz_clampi(dd[k] + v * convert[k], 0, a);
 									}
@@ -884,7 +884,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 									unsigned char v = sd[i];
 									if (v != 0)
 									{
-										unsigned char a = sd[sc];
+										unsigned char a = sd[ss];
 										for (k = 0; k < dc; k++)
 											dd[k] = fz_clampi(dd[k] - v * convert[k], 0, a);
 									}
@@ -926,7 +926,7 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 									unsigned char v = 0xff - sd[i];
 									if (v != 0)
 									{
-										unsigned char a = sd[sc];
+										unsigned char a = sd[ss];
 										for (k = 0; k < dc; k++)
 											dd[k] = fz_clampi(dd[k] - v * convert[k], 0, a);
 									}
@@ -965,7 +965,11 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *dst, fz_pixmap *
 
 /* Convert a color given in terms of one colorspace,
  * to a color in terms of another colorspace/separations. */
-void fz_convert_separation_colors(fz_context *ctx, const fz_color_params *color_params, fz_colorspace *dst_cs, const fz_separations *dst_seps, float *dst_color, fz_colorspace *src_cs, const float *src_color)
+void
+fz_convert_separation_colors(fz_context *ctx,
+	fz_colorspace *src_cs, const float *src_color,
+	fz_separations *dst_seps, fz_colorspace *dst_cs, float *dst_color,
+	fz_color_params color_params)
 {
 	int i, j, n, dc, ds, dn, pred;
 	float remainders[FZ_MAX_COLORS];
@@ -1050,15 +1054,20 @@ found_process:
 	{
 		/* There were some spots that didn't copy over */
 		float converted[FZ_MAX_COLORS];
-		fz_convert_color(ctx, color_params, NULL, dst_cs, converted, src_cs, remainders);
-
+		fz_convert_color(ctx, src_cs, remainders, dst_cs, converted, NULL, color_params);
 		for (i = 0; i < dc; i++)
 			dst_color[i] += converted[i];
 	}
 }
 
 /* Get the equivalent separation color in a given colorspace. */
-void fz_separation_equivalent(fz_context *ctx, const fz_separations *seps, int i, const fz_color_params *color_params, fz_colorspace *dst_cs, fz_colorspace *prf, float *convert)
+void
+fz_separation_equivalent(fz_context *ctx,
+	const fz_separations *seps,
+	int i,
+	fz_colorspace *dst_cs, float *convert,
+	fz_colorspace *prf,
+	fz_color_params color_params)
 {
 	float colors[FZ_MAX_COLORS];
 
@@ -1085,5 +1094,5 @@ void fz_separation_equivalent(fz_context *ctx, const fz_separations *seps, int i
 
 	memset(colors, 0, sizeof(float) * fz_colorspace_n(ctx, seps->cs[i]));
 	colors[seps->cs_pos[i]] = 1;
-	fz_convert_color(ctx, color_params, prf, dst_cs, convert, seps->cs[i], colors);
+	fz_convert_color(ctx, seps->cs[i], colors, dst_cs, convert, prf, color_params);
 }
