@@ -4,6 +4,7 @@
 #include <assert.h>
 #include <limits.h>
 #include <stdio.h>
+#include <string.h>
 
 typedef struct fz_item_s fz_item;
 
@@ -463,19 +464,14 @@ fz_store_item(fz_context *ctx, void *key, void *val_, size_t itemsize, const fz_
 	if (!store)
 		return NULL;
 
-	fz_var(item);
-
 	/* If we fail for any reason, we swallow the exception and continue.
 	 * All that the above program will see is that we failed to store
 	 * the item. */
-	fz_try(ctx)
-	{
-		item = fz_malloc_struct(ctx, fz_item);
-	}
-	fz_catch(ctx)
-	{
+
+	item = fz_malloc_no_throw(ctx, sizeof (fz_item));
+	if (!item)
 		return NULL;
-	}
+	memset(item, 0, sizeof (fz_item));
 
 	if (type->make_hash_key)
 	{
@@ -777,7 +773,7 @@ fz_debug_store_item(fz_context *ctx, void *state, void *key_, int keylen, void *
 	printf("hash[");
 	for (i=0; i < keylen; ++i)
 		printf("%02x", key[i]);
-	printf("][refs=%d][size=%d] key=%s val=%p\n", item->val->refs, (int)item->size, buf, item->val);
+	printf("][refs=%d][size=%d] key=%s val=%p\n", item->val->refs, (int)item->size, buf, (void *)item->val);
 }
 
 static void
@@ -801,7 +797,7 @@ fz_debug_store_locked(fz_context *ctx)
 		item->type->format_key(ctx, buf, sizeof buf, item->key);
 		fz_lock(ctx, FZ_LOCK_ALLOC);
 		printf("store[*][refs=%d][size=%d] key=%s val=%p\n",
-				item->val->refs, (int)item->size, buf, item->val);
+				item->val->refs, (int)item->size, buf, (void *)item->val);
 		if (next)
 		{
 			(void)Memento_dropRef(next->val);
