@@ -50,6 +50,38 @@ fz_option_eq(const char *a, const char *b)
 	return !strncmp(a, b, n) && (a[n] == ',' || a[n] == 0);
 }
 
+int
+fz_copy_option(fz_context *ctx, const char *val, char *dest, size_t maxlen)
+{
+	const char *e = val;
+	size_t len, len2;
+
+	if (val == NULL) {
+		if (maxlen)
+			*dest = 0;
+		return 0;
+	}
+
+	while (*e != ',' && *e != 0)
+		e++;
+
+	len = e-val;
+	len2 = len+1; /* Allow for terminator */
+	if (len > maxlen)
+		len = maxlen;
+	memcpy(dest, val, len);
+	if (len < maxlen)
+		memset(dest+len, 0, maxlen-len);
+
+	return len2 >= maxlen ? len2 - maxlen : 0;
+}
+
+/*
+	Internal function to allocate a
+	block for a derived document_writer structure, with the base
+	structure's function pointers populated correctly, and the extra
+	space zero initialised.
+*/
 fz_document_writer *fz_new_document_writer_of_size(fz_context *ctx, size_t size, fz_document_writer_begin_page_fn *begin_page,
 	fz_document_writer_end_page_fn *end_page, fz_document_writer_close_writer_fn *close, fz_document_writer_drop_writer_fn *drop)
 {
@@ -68,11 +100,6 @@ fz_document_writer *fz_new_png_pixmap_writer(fz_context *ctx, const char *path,
 	return fz_new_pixmap_writer(ctx, path, options, "out-%04d.png", 0, fz_save_pixmap_as_png);
 }
 
-fz_document_writer *fz_new_tga_pixmap_writer(fz_context *ctx, const char *path, const char *options)
-{
-	return fz_new_pixmap_writer(ctx, path, options, "out-%04d.tga", 0, fz_save_pixmap_as_tga);
-}
-
 fz_document_writer *fz_new_pam_pixmap_writer(fz_context *ctx, const char *path, const char *options)
 {
 	return fz_new_pixmap_writer(ctx, path, options, "out-%04d.pam", 0, fz_save_pixmap_as_pam);
@@ -103,6 +130,18 @@ fz_document_writer *fz_new_pkm_pixmap_writer(fz_context *ctx, const char *path,
 	return fz_new_pixmap_writer(ctx, path, options, "out-%04d.pkm", 4, fz_save_pixmap_as_pkm);
 }
 
+/*
+	Create a new fz_document_writer, for a
+	file of the given type.
+
+	path: The document name to write (or NULL for default)
+
+	format: Which format to write (currently cbz, html, pdf, pam, pbm,
+	pgm, pkm, png, ppm, pnm, svg, text, xhtml)
+
+	options: NULL, or pointer to comma separated string to control
+	file generation.
+*/
 fz_document_writer *
 fz_new_document_writer(fz_context *ctx, const char *path, const char *format, const char *options)
 {
@@ -125,8 +164,6 @@ fz_new_document_writer(fz_context *ctx, const char *path, const char *format, co
 
 	if (!fz_strcasecmp(format, "png"))
 		return fz_new_png_pixmap_writer(ctx, path, options);
-	if (!fz_strcasecmp(format, "tga"))
-		return fz_new_tga_pixmap_writer(ctx, path, options);
 	if (!fz_strcasecmp(format, "pam"))
 		return fz_new_pam_pixmap_writer(ctx, path, options);
 	if (!fz_strcasecmp(format, "pnm"))
@@ -161,6 +198,13 @@ fz_new_document_writer(fz_context *ctx, const char *path, const char *format, co
 	fz_throw(ctx, FZ_ERROR_GENERIC, "unknown output document format: %s", format);
 }
 
+/*
+	Called to end the process of writing
+	pages to a document.
+
+	This writes any file level trailers required. After this
+	completes successfully the file is up to date and complete.
+*/
 void
 fz_close_document_writer(fz_context *ctx, fz_document_writer *wri)
 {
@@ -169,6 +213,14 @@ fz_close_document_writer(fz_context *ctx, fz_document_writer *wri)
 	wri->close_writer = NULL;
 }
 
+/*
+	Called to discard a fz_document_writer.
+	This may be called at any time during the process to release all
+	the resources owned by the writer.
+
+	Calling drop without having previously called close may leave
+	the file in an inconsistent state.
+*/
 void
 fz_drop_document_writer(fz_context *ctx, fz_document_writer *wri)
 {
@@ -184,15 +236,29 @@ fz_drop_document_writer(fz_context *ctx, fz_document_writer *wri)
 	fz_free(ctx, wri);
 }
 
+/*
+	Called to start the process of writing a page to
+	a document.
+
+	mediabox: page size rectangle in points.
+
+	Returns a fz_device to write page contents to.
+*/
 fz_device *
 fz_begin_page(fz_context *ctx, fz_document_writer *wri, fz_rect mediabox)
 {
 	if (!wri)
 		return NULL;
+	if (wri->dev)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "called begin page without ending the previous page");
 	wri->dev = wri->begin_page(ctx, wri, mediabox);
 	return wri->dev;
 }
 
+/*
+	Called to end the process of writing a page to a
+	document.
+*/
 void
 fz_end_page(fz_context *ctx, fz_document_writer *wri)
 {
