@@ -2,8 +2,9 @@
 
 #include <string.h>
 #include <stdlib.h>
+#include <stdio.h>
 
-static const struct { const char *ent; int ucs; } html_entities[] = {
+static const struct { const char *name; int c; } html_entities[] = {
 	{"nbsp",160}, {"iexcl",161}, {"cent",162}, {"pound",163},
 	{"curren",164}, {"yen",165}, {"brvbar",166}, {"sect",167},
 	{"uml",168}, {"copy",169}, {"ordf",170}, {"laquo",171},
@@ -68,6 +69,7 @@ static const struct { const char *ent; int ucs; } html_entities[] = {
 
 struct parser
 {
+	fz_pool *pool;
 	fz_xml *head;
 	int preserve_white;
 	int depth;
@@ -80,6 +82,12 @@ struct attribute
 	struct attribute *next;
 };
 
+struct fz_xml_doc_s
+{
+	fz_pool *pool;
+	fz_xml *root;
+};
+
 struct fz_xml_s
 {
 	char name[40];
@@ -88,7 +96,6 @@ struct fz_xml_s
 	fz_xml *up, *down, *tail, *prev, *next;
 };
 
-#if 0
 static void xml_indent(int n)
 {
 	while (n--) {
@@ -97,6 +104,9 @@ static void xml_indent(int n)
 	}
 }
 
+/*
+	Pretty-print an XML tree to stdout.
+*/
 void fz_debug_xml(fz_xml *item, int level)
 {
 	if (item->text)
@@ -110,9 +120,9 @@ void fz_debug_xml(fz_xml *item, int level)
 			default:
 				if (c < 32 || c > 127) {
 					putchar('\\');
-					putchar('0' + ((c >> 6) & 7));
-					putchar('0' + ((c >> 3) & 7));
-					putchar('0' + ((c) & 7));
+					putchar('x');
+					putchar("0123456789ABCDEF"[(c>>4) & 15]);
+					putchar("0123456789ABCDEF"[(c) & 15]);
 				} else {
 					putchar(c);
 				}
@@ -145,38 +155,59 @@ void fz_debug_xml(fz_xml *item, int level)
 		printf(")%s\n", item->name);
 	}
 }
-#endif
 
+/*
+	Return previous sibling of XML node.
+*/
 fz_xml *fz_xml_prev(fz_xml *item)
 {
 	return item ? item->prev : NULL;
 }
 
+/*
+	Return next sibling of XML node.
+*/
 fz_xml *fz_xml_next(fz_xml *item)
 {
 	return item ? item->next : NULL;
 }
 
+/*
+	Return parent of XML node.
+*/
 fz_xml *fz_xml_up(fz_xml *item)
 {
 	return item ? item->up : NULL;
 }
 
+/*
+	Return first child of XML node.
+*/
 fz_xml *fz_xml_down(fz_xml *item)
 {
 	return item ? item->down : NULL;
 }
 
+/*
+	Return the text content of an XML node.
+	Return NULL if the node is a tag.
+*/
 char *fz_xml_text(fz_xml *item)
 {
 	return item ? item->text : NULL;
 }
 
+/*
+	Return tag of XML node. Return NULL for text nodes.
+*/
 char *fz_xml_tag(fz_xml *item)
 {
 	return item && item->name[0] ? item->name : NULL;
 }
 
+/*
+	Return true if the tag name matches.
+*/
 int fz_xml_is_tag(fz_xml *item, const char *name)
 {
 	if (!item)
@@ -184,6 +215,10 @@ int fz_xml_is_tag(fz_xml *item, const char *name)
 	return !strcmp(item->name, name);
 }
 
+/*
+	Return the value of an attribute of an XML node.
+	NULL if the attribute doesn't exist.
+*/
 char *fz_xml_att(fz_xml *item, const char *name)
 {
 	struct attribute *att;
@@ -220,37 +255,29 @@ fz_xml *fz_xml_find_down(fz_xml *item, const char *tag)
 	return fz_xml_find(item, tag);
 }
 
-static void xml_free_attribute(fz_context *ctx, struct attribute *att)
+fz_xml *fz_xml_root(fz_xml_doc *xml)
 {
-	while (att) {
-		struct attribute *next = att->next;
-		if (att->value)
-			fz_free(ctx, att->value);
-		fz_free(ctx, att);
-		att = next;
-	}
+	return xml ? xml->root : NULL;
 }
 
-void fz_drop_xml(fz_context *ctx, fz_xml *item)
+/*
+	Free the XML node and all its children and siblings.
+*/
+void fz_drop_xml(fz_context *ctx, fz_xml_doc *xml)
 {
-	while (item)
-	{
-		fz_xml *next = item->next;
-		if (item->text)
-			fz_free(ctx, item->text);
-		if (item->atts)
-			xml_free_attribute(ctx, item->atts);
-		if (item->down)
-			fz_drop_xml(ctx, item->down);
-		fz_free(ctx, item);
-		item = next;
-	}
+	if (xml)
+		fz_drop_pool(ctx, xml->pool);
 }
 
-void fz_detach_xml(fz_xml *node)
+/*
+	Detach a node from the tree, unlinking it from its parent,
+	and setting the document root to the node.
+*/
+void fz_detach_xml(fz_context *ctx, fz_xml_doc *xml, fz_xml *node)
 {
 	if (node->up)
 		node->up->down = NULL;
+	xml->root = node;
 }
 
 static size_t xml_parse_entity(int *c, char *a)
@@ -289,9 +316,9 @@ static size_t xml_parse_entity(int *c, char *a)
 
 	/* We should only be doing this for XHTML, but it shouldn't be a problem. */
 	for (i = 0; i < nelem(html_entities); ++i) {
-		size_t n = strlen(html_entities[i].ent);
-		if (!memcmp(a+1, html_entities[i].ent, n) && a[1+n] == ';') {
-			*c = html_entities[i].ucs;
+		size_t n = strlen(html_entities[i].name);
+		if (!strncmp(a+1, html_entities[i].name, n) && a[n+1] == ';') {
+			*c = html_entities[i].c;
 			return n + 2;
 		}
 	}
@@ -323,7 +350,7 @@ static void xml_emit_open_tag(fz_context *ctx, struct parser *parser, char *a, c
 		if (*ns == ':')
 			a = ns + 1;
 
-	head = fz_malloc_struct(ctx, fz_xml);
+	head = fz_pool_alloc(ctx, parser->pool, sizeof *head);
 	if (b - a > sizeof(head->name) - 1)
 		b = a + sizeof(head->name) - 1;
 	memcpy(head->name, a, b - a);
@@ -356,7 +383,7 @@ static void xml_emit_att_name(fz_context *ctx, struct parser *parser, char *a, c
 	fz_xml *head = parser->head;
 	struct attribute *att;
 
-	att = fz_malloc_struct(ctx, struct attribute);
+	att = fz_pool_alloc(ctx, parser->pool, sizeof *att);
 	if (b - a > sizeof(att->name) - 1)
 		b = a + sizeof(att->name) - 1;
 	memcpy(att->name, a, b - a);
@@ -374,7 +401,7 @@ static void xml_emit_att_value(fz_context *ctx, struct parser *parser, char *a,
 	int c;
 
 	/* entities are all longer than UTFmax so runetochar is safe */
-	s = att->value = fz_malloc(ctx, b - a + 1);
+	s = att->value = fz_pool_alloc(ctx, parser->pool, b - a + 1);
 	while (a < b) {
 		if (*a == '&') {
 			a += xml_parse_entity(&c, a);
@@ -419,7 +446,7 @@ static void xml_emit_text(fz_context *ctx, struct parser *parser, char *a, char
 	head = parser->head;
 
 	/* entities are all longer than UTFmax so runetochar is safe */
-	s = head->text = fz_malloc(ctx, b - a + 1);
+	s = head->text = fz_pool_alloc(ctx, parser->pool, b - a + 1);
 	while (a < b) {
 		if (*a == '&') {
 			a += xml_parse_entity(&c, a);
@@ -443,7 +470,7 @@ static void xml_emit_cdata(fz_context *ctx, struct parser *parser, char *a, char
 	xml_emit_open_tag(ctx, parser, empty, empty);
 	head = parser->head;
 
-	s = head->text = fz_malloc(ctx, b - a + 1);
+	s = head->text = fz_pool_alloc(ctx, parser->pool, b - a + 1);
 	while (a < b)
 		*s++ = *a++;
 	*s = 0;
@@ -481,9 +508,11 @@ parse_element:
 	return "syntax error in element";
 
 parse_comment:
+	if (p[0]=='D' && p[1]=='O' && p[2]=='C' && p[3]=='T' && p[4]=='Y' && p[5]=='P' && p[6]=='E')
+		goto parse_declaration;
+	if (p[0]=='E' && p[1]=='N' && p[2]=='T' && p[3]=='I' && p[4]=='T' && p[5]=='Y')
+		goto parse_declaration;
 	if (*p == '[') goto parse_cdata;
-	if (*p == 'D' && !memcmp(p, "DOCTYPE", 7)) goto parse_declaration;
-	if (*p == 'E' && !memcmp(p, "ENTITY", 6)) goto parse_declaration;
 	if (*p++ != '-') return "syntax error in comment (<! not followed by --)";
 	if (*p++ != '-') return "syntax error in comment (<!- not followed by -)";
 	while (*p) {
@@ -590,15 +619,57 @@ parse_attribute_value:
 	return "end of data in attribute value";
 }
 
-static char *convert_to_utf8(fz_context *doc, const unsigned char *s, size_t n, int *dofree)
+static int startswith(const char *a, const char *b)
 {
+	return !fz_strncasecmp(a, b, strlen(b));
+}
+
+static const unsigned short *find_xml_encoding(char *s)
+{
+	const unsigned short *table = NULL;
+	char *end, *xml, *enc;
+
+	end = strchr(s, '>');
+	if (end)
+	{
+		*end = 0;
+		xml = strstr(s, "<?xml");
+		if (xml)
+		{
+			enc = strstr(xml, "encoding=");
+			if (enc)
+			{
+				enc += 10;
+				if (startswith(enc, "iso-8859-1") || startswith(enc, "latin1"))
+					table = fz_unicode_from_iso8859_1;
+				else if (startswith(enc, "iso-8859-7") || startswith(enc, "greek"))
+					table = fz_unicode_from_iso8859_7;
+				else if (startswith(enc, "koi8"))
+					table = fz_unicode_from_koi8u;
+				else if (startswith(enc, "windows-1250"))
+					table = fz_unicode_from_windows_1250;
+				else if (startswith(enc, "windows-1251"))
+					table = fz_unicode_from_windows_1251;
+				else if (startswith(enc, "windows-1252"))
+					table = fz_unicode_from_windows_1252;
+			}
+		}
+		*end = '>';
+	}
+
+	return table;
+}
+
+static char *convert_to_utf8(fz_context *ctx, unsigned char *s, size_t n, int *dofree)
+{
+	const unsigned short *table;
 	const unsigned char *e = s + n;
 	char *dst, *d;
 	int c;
 
 	if (s[0] == 0xFE && s[1] == 0xFF) {
 		s += 2;
-		dst = d = fz_malloc(doc, n * FZ_UTFMAX);
+		dst = d = fz_malloc(ctx, n * FZ_UTFMAX);
 		while (s + 1 < e) {
 			c = s[0] << 8 | s[1];
 			d += fz_runetochar(d, c);
@@ -611,7 +682,7 @@ static char *convert_to_utf8(fz_context *doc, const unsigned char *s, size_t n,
 
 	if (s[0] == 0xFF && s[1] == 0xFE) {
 		s += 2;
-		dst = d = fz_malloc(doc, n * FZ_UTFMAX);
+		dst = d = fz_malloc(ctx, n * FZ_UTFMAX);
 		while (s + 1 < e) {
 			c = s[0] | s[1] << 8;
 			d += fz_runetochar(d, c);
@@ -622,6 +693,18 @@ static char *convert_to_utf8(fz_context *doc, const unsigned char *s, size_t n,
 		return dst;
 	}
 
+	table = find_xml_encoding((char*)s);
+	if (table) {
+		dst = d = fz_malloc(ctx, n * FZ_UTFMAX);
+		while (*s) {
+			c = table[*s++];
+			d += fz_runetochar(d, c);
+		}
+		*d = 0;
+		*dofree = 1;
+		return dst;
+	}
+
 	*dofree = 0;
 
 	if (s[0] == 0xEF && s[1] == 0xBB && s[2] == 0xBF)
@@ -630,32 +713,49 @@ static char *convert_to_utf8(fz_context *doc, const unsigned char *s, size_t n,
 	return (char*)s;
 }
 
-fz_xml *
+/*
+	Parse the contents of buffer into a tree of xml nodes.
+
+	preserve_white: whether to keep or delete all-whitespace nodes.
+*/
+fz_xml_doc *
 fz_parse_xml(fz_context *ctx, fz_buffer *buf, int preserve_white)
 {
 	struct parser parser;
+	fz_xml_doc *xml = NULL;
 	fz_xml root, *node;
-	char *p, *error;
+	char *p = NULL;
+	char *error;
 	int dofree;
 	unsigned char *s;
 	size_t n;
 
+	fz_var(p);
+
 	/* ensure we are zero-terminated */
 	fz_terminate_buffer(ctx, buf);
 	n = fz_buffer_storage(ctx, buf, &s);
 
 	memset(&root, 0, sizeof(root));
+	parser.pool = fz_new_pool(ctx);
 	parser.head = &root;
 	parser.preserve_white = preserve_white;
 	parser.depth = 0;
 
-	p = convert_to_utf8(ctx, s, n, &dofree);
-
 	fz_try(ctx)
 	{
+		p = convert_to_utf8(ctx, s, n, &dofree);
+
 		error = xml_parse_document_imp(ctx, &parser, p);
 		if (error)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "%s", error);
+
+		for (node = root.down; node; node = node->next)
+			node->up = NULL;
+
+		xml = fz_pool_alloc(ctx, parser.pool, sizeof *xml);
+		xml->pool = parser.pool;
+		xml->root = root.down;
 	}
 	fz_always(ctx)
 	{
@@ -664,11 +764,9 @@ fz_parse_xml(fz_context *ctx, fz_buffer *buf, int preserve_white)
 	}
 	fz_catch(ctx)
 	{
-		fz_drop_xml(ctx, root.down);
+		fz_drop_pool(ctx, parser.pool);
 		fz_rethrow(ctx);
 	}
 
-	for (node = root.down; node; node = node->next)
-		node->up = NULL;
-	return root.down;
+	return xml;
 }
