@@ -3,69 +3,6 @@ Error.prototype.toString = function() {
 	return this.name + ': ' + this.message;
 };
 
-var MuPDF = {
-	monthPattern: /Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec/,
-	monthName: [
-		'January',
-		'February',
-		'March',
-		'April',
-		'May',
-		'June',
-		'July',
-		'August',
-		'September',
-		'October',
-		'November',
-		'December'
-	],
-	shortMonthName: [
-		'Jan',
-		'Feb',
-		'Mar',
-		'Apr',
-		'May',
-		'Jun',
-		'Jul',
-		'Aug',
-		'Sep',
-		'Oct',
-		'Nov',
-		'Dec'
-	],
-	dayName: [
-		'Sunday',
-		'Monday',
-		'Tuesday',
-		'Wednesday',
-		'Thursday',
-		'Friday',
-		'Saturday'
-	],
-	dateFormats: [
-		'm/d',
-		'm/d/yy',
-		'mm/dd/yy',
-		'mm/yy',
-		'd-mmm',
-		'd-mmm-yy',
-		'dd-mm-yy',
-		'yy-mm-dd',
-		'mmm-yy',
-		'mmmm-yy',
-		'mmm d, yyyy',
-		'mmmm d, yyyy',
-		'm/d/yy h:MM tt',
-		'm/d/yy HH:MM'
-	],
-	timeFormats: [
-		'HH:MM',
-		'h:MM tt',
-		'HH:MM:ss',
-		'h:MM:ss tt'
-	],
-};
-
 // display must be kept in sync with an enum in pdf_form.c
 var display = {
 	visible: 0,
@@ -209,77 +146,101 @@ var zoomtype = {
 	refW: 'ReflowWidth',
 };
 
-var util = {};
+util.scand = function (fmt, input) {
+	// This seems to match Acrobat's parsing behavior
+	return AFParseDateEx(input, fmt);
+}
 
-util.printd = function (fmt, d) {
-	function padZeros(num, places) {
-		var s = String(num)
-		while (s.length < places)
-			s = '0' + s;
-		return s;
-	}
-	if (!d) return null;
-	var res = '';
-	var tokens = fmt.match(/(m+|d+|y+|H+|h+|M+|s+|t+|[^mdyHhMst]+)/g);
-	var length = tokens ? tokens.length : 0;
-	var i;
-	for (i = 0; i < length; ++i) {
-		switch (tokens[i]) {
-		case 'mmmm': res += MuPDF.monthName[d.getMonth()]; break;
-		case 'mmm': res += MuPDF.monthName[d.getMonth()].substring(0, 3); break;
-		case 'mm': res += padZeros(d.getMonth()+1, 2); break;
-		case 'm': res += d.getMonth()+1; break;
-		case 'dddd': res += MuPDF.dayName[d.getDay()]; break;
-		case 'ddd': res += MuPDF.dayName[d.getDay()].substring(0, 3); break;
-		case 'dd': res += padZeros(d.getDate(), 2); break;
-		case 'd': res += d.getDate(); break;
-		case 'yyyy': res += d.getFullYear(); break;
-		case 'yy': res += d.getFullYear()%100; break;
-		case 'HH': res += padZeros(d.getHours(), 2); break;
-		case 'H': res += d.getHours(); break;
-		case 'hh': res += padZeros((d.getHours()+11)%12+1, 2); break;
-		case 'h': res += (d.getHours()+11)%12+1; break;
-		case 'MM': res += padZeros(d.getMinutes(), 2); break;
-		case 'M': res += d.getMinutes(); break;
-		case 'ss': res += padZeros(d.getSeconds(), 2); break;
-		case 's': res += d.getSeconds(); break;
-		case 'tt': res += d.getHours() < 12 ? 'am' : 'pm'; break;
-		case 't': res += d.getHours() < 12 ? 'a' : 'p'; break;
-		default: res += tokens[i];
+util.printd = function (fmt, date) {
+	var monthName = [
+		'January',
+		'February',
+		'March',
+		'April',
+		'May',
+		'June',
+		'July',
+		'August',
+		'September',
+		'October',
+		'November',
+		'December'
+	];
+	var dayName = [
+		'Sunday',
+		'Monday',
+		'Tuesday',
+		'Wednesday',
+		'Thursday',
+		'Friday',
+		'Saturday'
+	];
+	if (fmt === 0)
+		fmt = 'D:yyyymmddHHMMss';
+	else if (fmt === 1)
+		fmt = 'yyyy.mm.dd HH:MM:ss';
+	else if (fmt === 2)
+		fmt = 'm/d/yy h:MM:ss tt';
+	if (!date)
+		date = new Date();
+	else if (!(date instanceof Date))
+		date = new Date(date);
+	var tokens = fmt.match(/(\\.|m+|d+|y+|H+|h+|M+|s+|t+|[^\\mdyHhMst]*)/g);
+	var out = '';
+	for (var i = 0; i < tokens.length; ++i) {
+		var token = tokens[i];
+		switch (token) {
+		case 'mmmm': out += monthName[date.getMonth()]; break;
+		case 'mmm': out += monthName[date.getMonth()].substring(0, 3); break;
+		case 'mm': out += util.printf('%02d', date.getMonth()+1); break;
+		case 'm': out += date.getMonth()+1; break;
+		case 'dddd': out += dayName[date.getDay()]; break;
+		case 'ddd': out += dayName[date.getDay()].substring(0, 3); break;
+		case 'dd': out += util.printf('%02d', date.getDate()); break;
+		case 'd': out += date.getDate(); break;
+		case 'yyyy': out += date.getFullYear(); break;
+		case 'yy': out += date.getFullYear() % 100; break;
+		case 'HH': out += util.printf('%02d', date.getHours()); break;
+		case 'H': out += date.getHours(); break;
+		case 'hh': out += util.printf('%02d', (date.getHours()+11)%12+1); break;
+		case 'h': out += (date.getHours() + 11) % 12 + 1; break;
+		case 'MM': out += util.printf('%02d', date.getMinutes()); break;
+		case 'M': out += date.getMinutes(); break;
+		case 'ss': out += util.printf('%02d', date.getSeconds()); break;
+		case 's': out += date.getSeconds(); break;
+		case 'tt': out += date.getHours() < 12 ? 'am' : 'pm'; break;
+		case 't': out += date.getHours() < 12 ? 'a' : 'p'; break;
+		default: out += (token[0] == '\\') ? token[1] : token; break;
 		}
 	}
-	return res;
+	return out;
 }
 
 util.printx = function (fmt, val) {
-	function convertCase(str, cmd) {
-		switch (cmd) {
-		case '>': return str.toUpperCase();
-		case '<': return str.toLowerCase();
-		default: return str;
-		}
-	}
-	var cs = '=';
+	function toUpper(str) { return str.toUpperCase(); }
+	function toLower(str) { return str.toLowerCase(); }
+	function toSame(str) { return str; }
+	var convertCase = toSame;
 	var res = '';
 	var i, m;
-	var length = fmt ? fmt.length : 0;
-	for (i = 0; i < length; ++i) {
+	var n = fmt ? fmt.length : 0;
+	for (i = 0; i < n; ++i) {
 		switch (fmt.charAt(i)) {
 		case '\\':
-			if (++i < length)
+			if (++i < n)
 				res += fmt.charAt(i);
 			break;
 		case 'X':
 			m = val.match(/\w/);
 			if (m) {
-				res += convertCase(m[0], cs);
+				res += convertCase(m[0]);
 				val = val.replace(/^\W*\w/, '');
 			}
 			break;
 		case 'A':
 			m = val.match(/[A-Za-z]/);
 			if (m) {
-				res += convertCase(m[0], cs);
+				res += convertCase(m[0]);
 				val = val.replace(/^[^A-Za-z]*[A-Za-z]/, '');
 			}
 			break;
@@ -291,309 +252,250 @@ util.printx = function (fmt, val) {
 			}
 			break;
 		case '*':
-			res += val;
+			res += convertCase(val);
 			val = '';
 			break;
 		case '?':
-			if (val) {
-				res += convertCase(val.charAt(0), cs);
+			if (val !== '') {
+				res += convertCase(val.charAt(0));
 				val = val.substring(1);
 			}
 			break;
 		case '=':
+			convertCase = toSame;
+			break;
 		case '>':
+			convertCase = toUpper;
+			break;
 		case '<':
-			cs = fmt.charAt(i);
+			convertCase = toLower;
 			break;
 		default:
-			res += convertCase(fmt.charAt(i), cs);
+			res += convertCase(fmt.charAt(i));
 			break;
 		}
 	}
 	return res;
 }
 
-util.printf = function () {
-	var i;
-
-	if (arguments.length < 1)
-		return '';
-
-	var res = '';
-	var arg_i = 1;
-	var regexp = /%[^dfsx]*[dfsx]|[^%]*/g;
-	var tokens = arguments[0].match(regexp);
-	var length = tokens ? tokens.length : 0;
-
-	for (i = 0; i < length; i++) {
-		var tok = tokens[i];
-		if (tok.match(/^%/)) {
-			if (arg_i < arguments.length) {
-				var val = arguments[arg_i++];
-				var fval = '';
-				var neg = false;
-				var decsep_re = /^,[0123]/;
-				var flags_re = /^[+ 0#]+/;
-				var width_re = /^\d+/;
-				var prec_re = /^\.\d+/;
-				var conv_re = /^[dfsx]/;
-
-				tok = tok.replace(/^%/, '');
-				var decsep = tok.match(decsep_re);
-				if (decsep)
-					decsep = decsep[0];
-				tok = tok.replace(decsep_re, '');
-				var flags = tok.match(flags_re);
-				if (flags)
-					flags = flags[0];
-				tok = tok.replace(flags_re, '');
-				var width = tok.match(width_re);
-				if (width)
-					width = width[0];
-				tok = tok.replace(width_re, '');
-				var prec = tok.match(prec_re);
-				if (prec)
-					prec = prec[0];
-				tok = tok.replace(prec_re, '');
-				var conv = tok.match(conv_re);
-				if (conv)
-					conv = conv[0];
-
-				prec = prec ? Number(prec.replace(/^\./, '')) : 0;
-				var poschar = (flags && flags.match(/[+ ]/)) ? flags.match(/[+ ]/)[0] : '';
-				var pad = (flags && flags.match(/0/)) ? '0' : ' ';
-
-				var point = '.';
-				var thou = '';
-
-				if (decsep) {
-					switch (decsep) {
-					case ',0': thou = ','; break;
-					case ',1': break;
-					case ',2': thou = '.'; point = ','; break;
-					case ',3': point = ','; break;
-					}
-				}
-
-				switch (conv) {
-				case 'x':
-					val = Math.floor(val);
-					neg = (val < 0);
-					if (neg)
-						val = -val;
-
-					// Convert to hex
-					while (val) {
-						fval = '0123456789ABCDEF'.charAt(val%16) + fval;
-						val = Math.floor(val/16);
-					}
-
-					if (neg)
-						fval = '-' + fval;
-					else
-						fval = poschar + fval;
-					break;
-
-				case 'd':
-					fval = String(Math.floor(val));
-					break;
-
-				case 's':
-					// Always pad strings with space
-					pad = ' ';
-					fval = String(val);
-					break;
-
-				case 'f':
-					fval = String(val);
-
-					if (prec) {
-						var frac = fval.match(/\.\d+/);
-						if (frac) {
-							frac = frac[0];
-							// Matched string includes the dot, so make it
-							// prec+1 in length
-							if (frac.length > prec+1)
-								frac = frac.substring(0, prec+1);
-							else if (frac.length < prec+1)
-								frac += new Array(prec+1-frac.length+1).join('0');
-
-							fval = fval.replace(/\.\d+/, frac);
-						}
-					}
-					break;
-				}
-
-				if (conv.match(/[fd]/)) {
-					if (fval >= 0)
-						fval = poschar + fval;
-
-					if (point !== '.')
-						fval.replace(/\./, point);
-
-					if (thou) {
-						var intpart = fval.match(/\d+/)[0];
-						intpart = new Array(2-(intpart.length+2)%3+1).join('0') + intpart;
-						intpart = intpart.match(/.../g).join(thou).replace(/^0*[,.]?/, '');
-						fval = fval.replace(/\d+/, intpart);
-					}
-				}
-
-				if (width && fval.length < width)
-					fval = new Array(width - fval.length + 1).join(pad) + fval;
-
-				res += fval;
-			}
-		} else {
-			res += tok;
-		}
-	}
-
-	return res;
-}
-
 function AFMergeChange(event) {
-	return event.value;
+	var prefix, postfix;
+	var value = event.value;
+	if (event.willCommit)
+		return value;
+	if (event.selStart >= 0)
+		prefix = value.substring(0, event.selStart);
+	else
+		prefix = '';
+	if (event.selEnd >= 0 && event.selEnd <= value.length)
+		postfix = value.substring(event.selEnd, value.length);
+	else
+		postfix = '';
+	return prefix + event.change + postfix;
 }
 
-function AFMakeNumber(str) {
-	var nums = str.match(/\d+/g);
+function AFExtractNums(string) {
+	if (string.charAt(0) == '.' || string.charAt(0) == ',')
+		string = '0' + string;
+	return string.match(/\d+/g);
+}
+
+function AFMakeNumber(string) {
+	if (typeof string == 'number')
+		return string;
+	if (typeof string != 'string')
+		return null;
+	var nums = AFExtractNums(string);
 	if (!nums)
 		return null;
-	var res = nums.join('.');
-	if (str.match(/^[^0-9]*\./))
-		res = '0.'+res;
-	return res * (str.match(/-/) ? -1.0 : 1.0);
+	var result = nums.join('.');
+	if (string.indexOf('-.') >= 0)
+		result = '0.' + result;
+	if (string.indexOf('-') >= 0)
+		return -result;
+	return +result;
 }
 
-function AFExtractTime(dt) {
-	var ampm = dt.match(/(am|pm)/);
-	dt = dt.replace(/(am|pm)/, '');
-	var t = dt.match(/\d{1,2}:\d{1,2}:\d{1,2}/);
-	dt = dt.replace(/\d{1,2}:\d{1,2}:\d{1,2}/, '');
-	if (!t) {
-		t = dt.match(/\d{1,2}:\d{1,2}/);
-		dt = dt.replace(/\d{1,2}:\d{1,2}/, '');
+function AFExtractTime(string) {
+	var pattern = /\d\d?:\d\d?(:\d\d?)?\s*(am|pm)?/i;
+	var match = pattern.exec(string);
+	if (match) {
+		var prefix = string.substring(0, match.index);
+		var suffix = string.substring(match.index + match[0].length);
+		return [ prefix + suffix, match[0] ];
 	}
-	return [dt, t?t[0]+(ampm?ampm[0]:''):''];
+	return null;
 }
 
 function AFParseDateOrder(fmt) {
-	var i;
 	var order = '';
-
-	// Ensure all present with those not added in default order
-	fmt += 'mdy';
-
-	for (i = 0; i < fmt.length; i++) {
+	fmt += 'mdy'; // Default order if any parts are missing.
+	for (var i = 0; i < fmt.length; i++) {
 		var c = fmt.charAt(i);
-		if ('ymd'.indexOf(c) !== -1 && order.indexOf(c) === -1)
+		if ((c == 'y' || c == 'm' || c == 'd') && order.indexOf(c) < 0)
 			order += c;
 	}
-
 	return order;
 }
 
-function AFMatchMonth(d) {
-	var m = d.match(MuPDF.monthPattern);
-	return m ? MuPDF.shortMonthName.indexOf(m[0]) : null;
+function AFMatchMonth(date) {
+	var names = ['jan','feb','mar','apr','may','jun','jul','aug','sep','oct','nov','dec'];
+	var month = date.match(/Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec/i);
+	if (month)
+		return names.indexOf(month[0].toLowerCase());
+	return null;
 }
 
-function AFParseTime(str, d) {
-	if (!str)
-		return d;
+function AFParseTime(string, date) {
+	if (!date)
+		date = new Date();
+	if (!string)
+		return date;
+	var nums = AFExtractNums(string);
+	if (!nums || nums.length < 2 || nums.length > 3)
+		return null;
+	var hour = nums[0];
+	var min = nums[1];
+	var sec = (nums.length == 3) ? nums[2] : 0;
+	if (hour < 12 && (/pm/i).test(string))
+		hour += 12;
+	if (hour >= 12 && (/am/i).test(string))
+		hour -= 12;
+	date.setHours(hour, min, sec);
+	if (date.getHours() != hour || date.getMinutes() != min || date.getSeconds() != sec)
+		return null;
+	return date;
+}
 
-	if (!d)
-		d = new Date();
+function AFMakeDate(out, year, month, date, time)
+{
+	if (year < 50)
+		year += 2000;
+	if (year < 100)
+		year += 1900;
+	out.setFullYear(year, month, date);
+	if (out.getFullYear() != year || out.getMonth() != month || out.getDate() != date)
+		return null;
+	if (time)
+		out = AFParseTime(time, out);
+	else
+		out.setHours(0, 0, 0);
+	return out;
+}
 
-	var ampm = str.match(/(am|pm)/);
-	var nums = str.match(/\d+/g);
-	var hour, min, sec;
+function AFParseDateEx(string, fmt) {
+	var out = new Date();
+	var year = out.getFullYear();
+	var month;
+	var date;
+	var i;
 
+	out.setHours(12, 0, 0);
+
+	var order = AFParseDateOrder(fmt);
+
+	var time = AFExtractTime(string);
+	if (time) {
+		string = time[0];
+		time = time[1];
+	}
+
+	var nums = AFExtractNums(string);
 	if (!nums)
 		return null;
 
-	sec = 0;
-
-	switch (nums.length) {
-	case 3:
-		sec = parseInt(nums[2]);
-	case 2:
-		hour = parseInt(nums[0]);
-		min = parseInt(nums[1]);
-		break;
-	default:
-		return null;
+	if (nums.length == 3) {
+		year = nums[order.indexOf('y')];
+		month = nums[order.indexOf('m')];
+		date = nums[order.indexOf('d')];
+		return AFMakeDate(out, year, month-1, date, time);
 	}
 
-	ampm = ampm && ampm[0]
+	month = AFMatchMonth(string);
 
-	if (ampm === 'am' && hour < 12)
-		hour = 12 + hour;
-	if (ampm === 'pm' && hour >= 12)
-		hour = 0 + hour - 12;
+	if (nums.length == 2) {
+		// We have a textual month.
+		if (month) {
+			if (order.indexOf('y') < order.indexOf('d')) {
+				year = nums[0];
+				date = nums[1];
+			} else {
+				year = nums[1];
+				date = nums[0];
+			}
+		}
 
-	d.setHours(hour, min, sec);
+		// Year before date: set year and month.
+		else if (order.indexOf('y') < order.indexOf('d')) {
+			if (order.indexOf('y') < order.indexOf('m')) {
+				year = nums[0];
+				month = nums[1];
+			} else {
+				year = nums[1];
+				month = nums[0];
+			}
+		}
 
-	if (d.getHours() !== hour || d.getMinutes() !== min || d.getSeconds() !== sec)
-		return null;
+		// Date before year: set date and month.
+		else {
+			if (order.indexOf('d') < order.indexOf('m')) {
+				date = nums[0];
+				month = nums[1];
+			} else {
+				date = nums[1];
+				month = nums[0];
+			}
+		}
 
-	return d;
-}
-
-function AFParseDateEx(d, fmt) {
-	var i;
-	var dt = AFExtractTime(d);
-	var nums = dt[0].match(/\d+/g);
-	var order = AFParseDateOrder(fmt);
-	var text_month = AFMatchMonth(dt[0]);
-	var dout = new Date();
-	var year = dout.getFullYear();
-	var month = dout.getMonth();
-	var date = dout.getDate();
-
-	dout.setHours(12, 0, 0);
-
-	if (!nums || nums.length < 1 || nums.length > 3)
-		return null;
-
-	if (nums.length < 3 && text_month) {
-		// Use the text month rather than one of the numbers
-		month = text_month;
-		order = order.replace('m', '');
+		return AFMakeDate(out, year, month-1, date, time);
 	}
 
-	order = order.substring(0, nums.length);
+	if (nums.length == 1) {
+		if (month) {
+			if (order.indexOf('y') < order.indexOf('d')) {
+				year = nums[0];
+				date = 1;
+			} else {
+				date = nums[0];
+			}
+			return AFMakeDate(out, year, month-1, date, time);
+		}
 
-	// If year and month specified but not date then use the 1st
-	if (order === 'ym' || (order === 'y' && text_month))
-		date = 1;
-
-	for (i = 0; i < nums.length; i++) {
-		switch (order.charAt(i)) {
-		case 'y': year = parseInt(nums[i]); break;
-		case 'm': month = parseInt(nums[i]) - 1; break;
-		case 'd': date = parseInt(nums[i]); break;
+		// Only one number: must match format exactly!
+		if (string.length == fmt.length) {
+			year = month = date = '';
+			for (i = 0; i < fmt.length; ++i) {
+				switch (fmt.charAt(i)) {
+				case '\\': ++i; break;
+				case 'y': year += string.charAt(i); break;
+				case 'm': month += string.charAt(i); break;
+				case 'd': date += string.charAt(i); break;
+				}
+			}
+			return AFMakeDate(out, year, month-1, date, time);
 		}
 	}
 
-	if (year < 100) {
-		if (fmt.search('yyyy') !== -1)
-			return null;
-		if (year >= 50)
-			year = 1900 + year;
-		else if (year >= 0)
-			year = 2000 + year;
-	}
-
-	dout.setFullYear(year, month, date);
-
-	if (dout.getFullYear() !== year || dout.getMonth() !== month || dout.getDate() !== date)
-		return null;
-
-	return AFParseTime(dt[1], dout);
+	return null;
 }
 
+var AFDate_oldFormats = [
+	'm/d',
+	'm/d/yy',
+	'mm/dd/yy',
+	'mm/yy',
+	'd-mmm',
+	'd-mmm-yy',
+	'dd-mm-yy',
+	'yy-mm-dd',
+	'mmm-yy',
+	'mmmm-yy',
+	'mmm d, yyyy',
+	'mmmm d, yyyy',
+	'm/d/yy h:MM tt',
+	'm/d/yy HH:MM'
+];
+
 function AFDate_KeystrokeEx(fmt) {
 	if (event.willCommit && !AFParseDateEx(event.value, fmt)) {
 		app.alert('The date/time entered ('+event.value+') does not match the format ('+fmt+') of the field [ '+event.target.name+' ]');
@@ -602,7 +504,7 @@ function AFDate_KeystrokeEx(fmt) {
 }
 
 function AFDate_Keystroke(index) {
-	AFDate_KeystrokeEx(MuPDF.dateFormats[index]);
+	AFDate_KeystrokeEx(AFDate_oldFormats[index]);
 }
 
 function AFDate_FormatEx(fmt) {
@@ -611,7 +513,7 @@ function AFDate_FormatEx(fmt) {
 }
 
 function AFDate_Format(index) {
-	AFDate_FormatEx(MuPDF.dateFormats[index]);
+	AFDate_FormatEx(AFDate_oldFormats[index]);
 }
 
 function AFTime_Keystroke(index) {
@@ -627,25 +529,20 @@ function AFTime_FormatEx(fmt) {
 }
 
 function AFTime_Format(index) {
-	AFTime_FormatEx(MuPDF.timeFormats[index]);
+	var formats = [ 'HH:MM', 'h:MM tt', 'HH:MM:ss', 'h:MM:ss tt' ];
+	AFTime_FormatEx(formats[index]);
 }
 
 function AFSpecial_KeystrokeEx(fmt) {
-	var cs = '=';
+	function toUpper(str) { return str.toUpperCase(); }
+	function toLower(str) { return str.toLowerCase(); }
+	function toSame(str) { return str; }
+	var convertCase = toSame;
 	var val = event.value;
 	var res = '';
 	var i = 0;
 	var m;
 	var length = fmt ? fmt.length : 0;
-
-	function convertCase(str, cmd) {
-		switch (cmd) {
-		case '>': return str.toUpperCase();
-		case '<': return str.toLowerCase();
-		default: return str;
-		}
-	}
-
 	while (i < length) {
 		switch (fmt.charAt(i)) {
 		case '\\':
@@ -663,7 +560,7 @@ function AFSpecial_KeystrokeEx(fmt) {
 				event.rc = false;
 				break;
 			}
-			res += convertCase(m[0], cs);
+			res += convertCase(m[0]);
 			val = val.substring(1);
 			break;
 
@@ -673,7 +570,7 @@ function AFSpecial_KeystrokeEx(fmt) {
 				event.rc = false;
 				break;
 			}
-			res += convertCase(m[0], cs);
+			res += convertCase(m[0]);
 			val = val.substring(1);
 			break;
 
@@ -688,23 +585,27 @@ function AFSpecial_KeystrokeEx(fmt) {
 			break;
 
 		case '*':
-			res += val;
+			res += convertCase(val);
 			val = '';
 			break;
 
 		case '?':
-			if (!val) {
+			if (val === '') {
 				event.rc = false;
 				break;
 			}
-			res += convertCase(val.charAt(0), cs);
+			res += convertCase(val.charAt(0));
 			val = val.substring(1);
 			break;
 
 		case '=':
+			convertCase = toSame;
+			break;
 		case '>':
+			convertCase = toUpper;
+			break;
 		case '<':
-			cs = fmt.charAt(i);
+			convertCase = toLower;
 			break;
 
 		default:
@@ -789,78 +690,31 @@ function AFNumber_Keystroke(nDec, sepStyle, negStyle, currStyle, strCurrency, bC
 }
 
 function AFNumber_Format(nDec, sepStyle, negStyle, currStyle, strCurrency, bCurrencyPrepend) {
-	var val = event.value;
-	var fracpart;
-	var intpart;
-	var point = sepStyle&2 ? ',' : '.';
-	var separator = sepStyle&2 ? '.' : ',';
-
-	if (/^\D*\./.test(val))
-		val = '0'+val;
-
-	var groups = val.match(/\d+/g);
-
-	if (!groups)
+	var value = AFMakeNumber(event.value);
+	var fmt = '%,' + sepStyle + '.' + nDec + 'f';
+	if (value == null) {
+		event.value = '';
 		return;
-
-	switch (groups.length) {
-	case 0:
-		return;
-	case 1:
-		fracpart = '';
-		intpart = groups[0];
-		break;
-	default:
-		fracpart = groups.pop();
-		intpart = groups.join('');
-		break;
 	}
-
-	// Remove leading zeros
-	intpart = intpart.replace(/^0*/, '');
-	if (!intpart)
-		intpart = '0';
-
-	if ((sepStyle & 1) === 0) {
-		// Add the thousands sepearators: pad to length multiple of 3 with zeros,
-		// split into 3s, join with separator, and remove the leading zeros
-		intpart = new Array(2-(intpart.length+2)%3+1).join('0') + intpart;
-		intpart = intpart.match(/.../g).join(separator).replace(/^0*/, '');
-	}
-
-	if (!intpart)
-		intpart = '0';
-
-	// Adjust fractional part to correct number of decimal places
-	fracpart += new Array(nDec+1).join('0');
-	fracpart = fracpart.substring(0, nDec);
-
-	if (fracpart)
-		intpart += point+fracpart;
-
 	if (bCurrencyPrepend)
-		intpart = strCurrency+intpart;
+		fmt = strCurrency + fmt;
 	else
-		intpart += strCurrency;
-
-	if (/-/.test(val)) {
-		switch (negStyle) {
-		case 0:
-			intpart = '-'+intpart;
-			break;
-		case 1:
-			break;
-		case 2:
-		case 3:
-			intpart = '('+intpart+')';
-			break;
-		}
+		fmt = fmt + strCurrency;
+	if (value < 0) {
+		/* negStyle: 0=MinusBlack, 1=Red, 2=ParensBlack, 3=ParensRed */
+		value = Math.abs(value);
+		if (negStyle == 2 || negStyle == 3)
+			fmt = '(' + fmt + ')';
+		else if (negStyle == 0)
+			fmt = '-' + fmt;
+		if (negStyle == 1 || negStyle == 3)
+			event.target.textColor = color.red;
+		else
+			event.target.textColor = color.black;
+	} else {
+		event.target.textColor = color.black;
 	}
-
-	if (negStyle&1)
-		event.target.textColor = /-/.test(val) ? color.red : color.black;
-
-	event.value = intpart;
+	event.value = util.printf(fmt, value);
 }
 
 function AFPercent_Keystroke(nDec, sepStyle) {
@@ -890,7 +744,7 @@ function AFSimple_Calculate(op, list) {
 		list = list.split(/ *, */);
 
 	for (i = 0; i < list.length; i++) {
-		var field = MuPDF_Doc.getField(list[i]);
+		var field = this.getField(list[i]);
 		var value = Number(field.value);
 		switch (op) {
 		case 'SUM': res += value; break;
@@ -922,7 +776,24 @@ function AFRange_Validate(lowerCheck, lowerLimit, upperCheck, upperLimit) {
 	}
 }
 
+/* Compatibility ECMAScript functions */
+String.prototype.substr = function (start, length) {
+	if (start < 0)
+		start = this.length + start;
+	if (length === undefined)
+		return this.substring(start, this.length);
+	return this.substring(start, start + length);
+}
+Date.prototype.getYear = Date.prototype.getFullYear;
+Date.prototype.setYear = Date.prototype.setFullYear;
+Date.prototype.toGMTString = Date.prototype.toUTCString;
+
+console.clear = function() { console.println('--- clear console ---\n'); };
+console.show = function(){};
+console.hide = function(){};
+
 app.plugIns = [];
 app.viewerType = 'Reader';
 app.language = 'ENU';
 app.viewerVersion = NaN;
+app.execDialog = function () { return 'cancel'; }
