@@ -3,69 +3,6 @@
 "return this.name + ': ' + this.message;\n"
 "};\n"
 "\n"
-"var MuPDF = {\n"
-"monthPattern: /Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec/,\n"
-"monthName: [\n"
-"'January',\n"
-"'February',\n"
-"'March',\n"
-"'April',\n"
-"'May',\n"
-"'June',\n"
-"'July',\n"
-"'August',\n"
-"'September',\n"
-"'October',\n"
-"'November',\n"
-"'December'\n"
-"],\n"
-"shortMonthName: [\n"
-"'Jan',\n"
-"'Feb',\n"
-"'Mar',\n"
-"'Apr',\n"
-"'May',\n"
-"'Jun',\n"
-"'Jul',\n"
-"'Aug',\n"
-"'Sep',\n"
-"'Oct',\n"
-"'Nov',\n"
-"'Dec'\n"
-"],\n"
-"dayName: [\n"
-"'Sunday',\n"
-"'Monday',\n"
-"'Tuesday',\n"
-"'Wednesday',\n"
-"'Thursday',\n"
-"'Friday',\n"
-"'Saturday'\n"
-"],\n"
-"dateFormats: [\n"
-"'m/d',\n"
-"'m/d/yy',\n"
-"'mm/dd/yy',\n"
-"'mm/yy',\n"
-"'d-mmm',\n"
-"'d-mmm-yy',\n"
-"'dd-mm-yy',\n"
-"'yy-mm-dd',\n"
-"'mmm-yy',\n"
-"'mmmm-yy',\n"
-"'mmm d, yyyy',\n"
-"'mmmm d, yyyy',\n"
-"'m/d/yy h:MM tt',\n"
-"'m/d/yy HH:MM'\n"
-"],\n"
-"timeFormats: [\n"
-"'HH:MM',\n"
-"'h:MM tt',\n"
-"'HH:MM:ss',\n"
-"'h:MM:ss tt'\n"
-"],\n"
-"};\n"
-"\n"
 "\n"
 "var display = {\n"
 "visible: 0,\n"
@@ -209,77 +146,101 @@
 "refW: 'ReflowWidth',\n"
 "};\n"
 "\n"
-"var util = {};\n"
+"util.scand = function (fmt, input) {\n"
 "\n"
-"util.printd = function (fmt, d) {\n"
-"function padZeros(num, places) {\n"
-"var s = String(num)\n"
-"while (s.length < places)\n"
-"s = '0' + s;\n"
-"return s;\n"
+"return AFParseDateEx(input, fmt);\n"
 "}\n"
-"if (!d) return null;\n"
-"var res = '';\n"
-"var tokens = fmt.match(/(m+|d+|y+|H+|h+|M+|s+|t+|[^mdyHhMst]+)/g);\n"
-"var length = tokens ? tokens.length : 0;\n"
-"var i;\n"
-"for (i = 0; i < length; ++i) {\n"
-"switch (tokens[i]) {\n"
-"case 'mmmm': res += MuPDF.monthName[d.getMonth()]; break;\n"
-"case 'mmm': res += MuPDF.monthName[d.getMonth()].substring(0, 3); break;\n"
-"case 'mm': res += padZeros(d.getMonth()+1, 2); break;\n"
-"case 'm': res += d.getMonth()+1; break;\n"
-"case 'dddd': res += MuPDF.dayName[d.getDay()]; break;\n"
-"case 'ddd': res += MuPDF.dayName[d.getDay()].substring(0, 3); break;\n"
-"case 'dd': res += padZeros(d.getDate(), 2); break;\n"
-"case 'd': res += d.getDate(); break;\n"
-"case 'yyyy': res += d.getFullYear(); break;\n"
-"case 'yy': res += d.getFullYear()%100; break;\n"
-"case 'HH': res += padZeros(d.getHours(), 2); break;\n"
-"case 'H': res += d.getHours(); break;\n"
-"case 'hh': res += padZeros((d.getHours()+11)%12+1, 2); break;\n"
-"case 'h': res += (d.getHours()+11)%12+1; break;\n"
-"case 'MM': res += padZeros(d.getMinutes(), 2); break;\n"
-"case 'M': res += d.getMinutes(); break;\n"
-"case 'ss': res += padZeros(d.getSeconds(), 2); break;\n"
-"case 's': res += d.getSeconds(); break;\n"
-"case 'tt': res += d.getHours() < 12 ? 'am' : 'pm'; break;\n"
-"case 't': res += d.getHours() < 12 ? 'a' : 'p'; break;\n"
-"default: res += tokens[i];\n"
+"\n"
+"util.printd = function (fmt, date) {\n"
+"var monthName = [\n"
+"'January',\n"
+"'February',\n"
+"'March',\n"
+"'April',\n"
+"'May',\n"
+"'June',\n"
+"'July',\n"
+"'August',\n"
+"'September',\n"
+"'October',\n"
+"'November',\n"
+"'December'\n"
+"];\n"
+"var dayName = [\n"
+"'Sunday',\n"
+"'Monday',\n"
+"'Tuesday',\n"
+"'Wednesday',\n"
+"'Thursday',\n"
+"'Friday',\n"
+"'Saturday'\n"
+"];\n"
+"if (fmt === 0)\n"
+"fmt = 'D:yyyymmddHHMMss';\n"
+"else if (fmt === 1)\n"
+"fmt = 'yyyy.mm.dd HH:MM:ss';\n"
+"else if (fmt === 2)\n"
+"fmt = 'm/d/yy h:MM:ss tt';\n"
+"if (!date)\n"
+"date = new Date();\n"
+"else if (!(date instanceof Date))\n"
+"date = new Date(date);\n"
+"var tokens = fmt.match(/(\\\\.|m+|d+|y+|H+|h+|M+|s+|t+|[^\\\\mdyHhMst]*)/g);\n"
+"var out = '';\n"
+"for (var i = 0; i < tokens.length; ++i) {\n"
+"var token = tokens[i];\n"
+"switch (token) {\n"
+"case 'mmmm': out += monthName[date.getMonth()]; break;\n"
+"case 'mmm': out += monthName[date.getMonth()].substring(0, 3); break;\n"
+"case 'mm': out += util.printf('%02d', date.getMonth()+1); break;\n"
+"case 'm': out += date.getMonth()+1; break;\n"
+"case 'dddd': out += dayName[date.getDay()]; break;\n"
+"case 'ddd': out += dayName[date.getDay()].substring(0, 3); break;\n"
+"case 'dd': out += util.printf('%02d', date.getDate()); break;\n"
+"case 'd': out += date.getDate(); break;\n"
+"case 'yyyy': out += date.getFullYear(); break;\n"
+"case 'yy': out += date.getFullYear() % 100; break;\n"
+"case 'HH': out += util.printf('%02d', date.getHours()); break;\n"
+"case 'H': out += date.getHours(); break;\n"
+"case 'hh': out += util.printf('%02d', (date.getHours()+11)%12+1); break;\n"
+"case 'h': out += (date.getHours() + 11) % 12 + 1; break;\n"
+"case 'MM': out += util.printf('%02d', date.getMinutes()); break;\n"
+"case 'M': out += date.getMinutes(); break;\n"
+"case 'ss': out += util.printf('%02d', date.getSeconds()); break;\n"
+"case 's': out += date.getSeconds(); break;\n"
+"case 'tt': out += date.getHours() < 12 ? 'am' : 'pm'; break;\n"
+"case 't': out += date.getHours() < 12 ? 'a' : 'p'; break;\n"
+"default: out += (token[0] == '\\\\') ? token[1] : token; break;\n"
 "}\n"
 "}\n"
-"return res;\n"
+"return out;\n"
 "}\n"
 "\n"
 "util.printx = function (fmt, val) {\n"
-"function convertCase(str, cmd) {\n"
-"switch (cmd) {\n"
-"case '>': return str.toUpperCase();\n"
-"case '<': return str.toLowerCase();\n"
-"default: return str;\n"
-"}\n"
-"}\n"
-"var cs = '=';\n"
+"function toUpper(str) { return str.toUpperCase(); }\n"
+"function toLower(str) { return str.toLowerCase(); }\n"
+"function toSame(str) { return str; }\n"
+"var convertCase = toSame;\n"
 "var res = '';\n"
 "var i, m;\n"
-"var length = fmt ? fmt.length : 0;\n"
-"for (i = 0; i < length; ++i) {\n"
+"var n = fmt ? fmt.length : 0;\n"
+"for (i = 0; i < n; ++i) {\n"
 "switch (fmt.charAt(i)) {\n"
 "case '\\\\':\n"
-"if (++i < length)\n"
+"if (++i < n)\n"
 "res += fmt.charAt(i);\n"
 "break;\n"
 "case 'X':\n"
 "m = val.match(/\\w/);\n"
 "if (m) {\n"
-"res += convertCase(m[0], cs);\n"
+"res += convertCase(m[0]);\n"
 "val = val.replace(/^\\W*\\w/, '');\n"
 "}\n"
 "break;\n"
 "case 'A':\n"
 "m = val.match(/[A-Za-z]/);\n"
 "if (m) {\n"
-"res += convertCase(m[0], cs);\n"
+"res += convertCase(m[0]);\n"
 "val = val.replace(/^[^A-Za-z]*[A-Za-z]/, '');\n"
 "}\n"
 "break;\n"
@@ -291,308 +252,249 @@
 "}\n"
 "break;\n"
 "case '*':\n"
-"res += val;\n"
+"res += convertCase(val);\n"
 "val = '';\n"
 "break;\n"
 "case '?':\n"
-"if (val) {\n"
-"res += convertCase(val.charAt(0), cs);\n"
+"if (val !== '') {\n"
+"res += convertCase(val.charAt(0));\n"
 "val = val.substring(1);\n"
 "}\n"
 "break;\n"
 "case '=':\n"
+"convertCase = toSame;\n"
+"break;\n"
 "case '>':\n"
+"convertCase = toUpper;\n"
+"break;\n"
 "case '<':\n"
-"cs = fmt.charAt(i);\n"
+"convertCase = toLower;\n"
 "break;\n"
 "default:\n"
-"res += convertCase(fmt.charAt(i), cs);\n"
+"res += convertCase(fmt.charAt(i));\n"
 "break;\n"
 "}\n"
 "}\n"
 "return res;\n"
 "}\n"
 "\n"
-"util.printf = function () {\n"
-"var i;\n"
-"\n"
-"if (arguments.length < 1)\n"
-"return '';\n"
-"\n"
-"var res = '';\n"
-"var arg_i = 1;\n"
-"var regexp = /%[^dfsx]*[dfsx]|[^%]*/g;\n"
-"var tokens = arguments[0].match(regexp);\n"
-"var length = tokens ? tokens.length : 0;\n"
-"\n"
-"for (i = 0; i < length; i++) {\n"
-"var tok = tokens[i];\n"
-"if (tok.match(/^%/)) {\n"
-"if (arg_i < arguments.length) {\n"
-"var val = arguments[arg_i++];\n"
-"var fval = '';\n"
-"var neg = false;\n"
-"var decsep_re = /^,[0123]/;\n"
-"var flags_re = /^[+ 0#]+/;\n"
-"var width_re = /^\\d+/;\n"
-"var prec_re = /^\\.\\d+/;\n"
-"var conv_re = /^[dfsx]/;\n"
-"\n"
-"tok = tok.replace(/^%/, '');\n"
-"var decsep = tok.match(decsep_re);\n"
-"if (decsep)\n"
-"decsep = decsep[0];\n"
-"tok = tok.replace(decsep_re, '');\n"
-"var flags = tok.match(flags_re);\n"
-"if (flags)\n"
-"flags = flags[0];\n"
-"tok = tok.replace(flags_re, '');\n"
-"var width = tok.match(width_re);\n"
-"if (width)\n"
-"width = width[0];\n"
-"tok = tok.replace(width_re, '');\n"
-"var prec = tok.match(prec_re);\n"
-"if (prec)\n"
-"prec = prec[0];\n"
-"tok = tok.replace(prec_re, '');\n"
-"var conv = tok.match(conv_re);\n"
-"if (conv)\n"
-"conv = conv[0];\n"
-"\n"
-"prec = prec ? Number(prec.replace(/^\\./, '')) : 0;\n"
-"var poschar = (flags && flags.match(/[+ ]/)) ? flags.match(/[+ ]/)[0] : '';\n"
-"var pad = (flags && flags.match(/0/)) ? '0' : ' ';\n"
-"\n"
-"var point = '.';\n"
-"var thou = '';\n"
-"\n"
-"if (decsep) {\n"
-"switch (decsep) {\n"
-"case ',0': thou = ','; break;\n"
-"case ',1': break;\n"
-"case ',2': thou = '.'; point = ','; break;\n"
-"case ',3': point = ','; break;\n"
-"}\n"
-"}\n"
-"\n"
-"switch (conv) {\n"
-"case 'x':\n"
-"val = Math.floor(val);\n"
-"neg = (val < 0);\n"
-"if (neg)\n"
-"val = -val;\n"
-"\n"
-"\n"
-"while (val) {\n"
-"fval = '0123456789ABCDEF'.charAt(val%16) + fval;\n"
-"val = Math.floor(val/16);\n"
-"}\n"
-"\n"
-"if (neg)\n"
-"fval = '-' + fval;\n"
-"else\n"
-"fval = poschar + fval;\n"
-"break;\n"
-"\n"
-"case 'd':\n"
-"fval = String(Math.floor(val));\n"
-"break;\n"
-"\n"
-"case 's':\n"
-"\n"
-"pad = ' ';\n"
-"fval = String(val);\n"
-"break;\n"
-"\n"
-"case 'f':\n"
-"fval = String(val);\n"
-"\n"
-"if (prec) {\n"
-"var frac = fval.match(/\\.\\d+/);\n"
-"if (frac) {\n"
-"frac = frac[0];\n"
-"\n"
-"\n"
-"if (frac.length > prec+1)\n"
-"frac = frac.substring(0, prec+1);\n"
-"else if (frac.length < prec+1)\n"
-"frac += new Array(prec+1-frac.length+1).join('0');\n"
-"\n"
-"fval = fval.replace(/\\.\\d+/, frac);\n"
-"}\n"
-"}\n"
-"break;\n"
-"}\n"
-"\n"
-"if (conv.match(/[fd]/)) {\n"
-"if (fval >= 0)\n"
-"fval = poschar + fval;\n"
-"\n"
-"if (point !== '.')\n"
-"fval.replace(/\\./, point);\n"
-"\n"
-"if (thou) {\n"
-"var intpart = fval.match(/\\d+/)[0];\n"
-"intpart = new Array(2-(intpart.length+2)%3+1).join('0') + intpart;\n"
-"intpart = intpart.match(/.../g).join(thou).replace(/^0*[,.]?/, '');\n"
-"fval = fval.replace(/\\d+/, intpart);\n"
-"}\n"
-"}\n"
-"\n"
-"if (width && fval.length < width)\n"
-"fval = new Array(width - fval.length + 1).join(pad) + fval;\n"
-"\n"
-"res += fval;\n"
-"}\n"
-"} else {\n"
-"res += tok;\n"
-"}\n"
-"}\n"
-"\n"
-"return res;\n"
-"}\n"
-"\n"
 "function AFMergeChange(event) {\n"
-"return event.value;\n"
+"var prefix, postfix;\n"
+"var value = event.value;\n"
+"if (event.willCommit)\n"
+"return value;\n"
+"if (event.selStart >= 0)\n"
+"prefix = value.substring(0, event.selStart);\n"
+"else\n"
+"prefix = '';\n"
+"if (event.selEnd >= 0 && event.selEnd <= value.length)\n"
+"postfix = value.substring(event.selEnd, value.length);\n"
+"else\n"
+"postfix = '';\n"
+"return prefix + event.change + postfix;\n"
 "}\n"
 "\n"
-"function AFMakeNumber(str) {\n"
-"var nums = str.match(/\\d+/g);\n"
+"function AFExtractNums(string) {\n"
+"if (string.charAt(0) == '.' || string.charAt(0) == ',')\n"
+"string = '0' + string;\n"
+"return string.match(/\\d+/g);\n"
+"}\n"
+"\n"
+"function AFMakeNumber(string) {\n"
+"if (typeof string == 'number')\n"
+"return string;\n"
+"if (typeof string != 'string')\n"
+"return null;\n"
+"var nums = AFExtractNums(string);\n"
 "if (!nums)\n"
 "return null;\n"
-"var res = nums.join('.');\n"
-"if (str.match(/^[^0-9]*\\./))\n"
-"res = '0.'+res;\n"
-"return res * (str.match(/-/) ? -1.0 : 1.0);\n"
+"var result = nums.join('.');\n"
+"if (string.indexOf('-.') >= 0)\n"
+"result = '0.' + result;\n"
+"if (string.indexOf('-') >= 0)\n"
+"return -result;\n"
+"return +result;\n"
 "}\n"
 "\n"
-"function AFExtractTime(dt) {\n"
-"var ampm = dt.match(/(am|pm)/);\n"
-"dt = dt.replace(/(am|pm)/, '');\n"
-"var t = dt.match(/\\d{1,2}:\\d{1,2}:\\d{1,2}/);\n"
-"dt = dt.replace(/\\d{1,2}:\\d{1,2}:\\d{1,2}/, '');\n"
-"if (!t) {\n"
-"t = dt.match(/\\d{1,2}:\\d{1,2}/);\n"
-"dt = dt.replace(/\\d{1,2}:\\d{1,2}/, '');\n"
+"function AFExtractTime(string) {\n"
+"var pattern = /\\d\\d?:\\d\\d?(:\\d\\d?)?\\s*(am|pm)?/i;\n"
+"var match = pattern.exec(string);\n"
+"if (match) {\n"
+"var prefix = string.substring(0, match.index);\n"
+"var suffix = string.substring(match.index + match[0].length);\n"
+"return [ prefix + suffix, match[0] ]\n"
 "}\n"
-"return [dt, t?t[0]+(ampm?ampm[0]:''):''];\n"
+"return null;\n"
 "}\n"
 "\n"
 "function AFParseDateOrder(fmt) {\n"
-"var i;\n"
 "var order = '';\n"
-"\n"
-"\n"
-"fmt += 'mdy';\n"
-"\n"
-"for (i = 0; i < fmt.length; i++) {\n"
+"fmt += 'mdy'; \n"
+"for (var i = 0; i < fmt.length; i++) {\n"
 "var c = fmt.charAt(i);\n"
-"if ('ymd'.indexOf(c) !== -1 && order.indexOf(c) === -1)\n"
+"if ((c == 'y' || c == 'm' || c == 'd') && order.indexOf(c) < 0)\n"
 "order += c;\n"
 "}\n"
-"\n"
 "return order;\n"
 "}\n"
 "\n"
-"function AFMatchMonth(d) {\n"
-"var m = d.match(MuPDF.monthPattern);\n"
-"return m ? MuPDF.shortMonthName.indexOf(m[0]) : null;\n"
+"function AFMatchMonth(date) {\n"
+"var names = ['jan','feb','mar','apr','may','jun','jul','aug','sep','oct','nov','dec'];\n"
+"var month = date.match(/Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec/i);\n"
+"if (month)\n"
+"return names.indexOf(month[0].toLowerCase());\n"
+"return null;\n"
 "}\n"
 "\n"
-"function AFParseTime(str, d) {\n"
-"if (!str)\n"
-"return d;\n"
+"function AFParseTime(string, date) {\n"
+"if (!date)\n"
+"date = new Date();\n"
+"if (!string)\n"
+"return date;\n"
+"var nums = AFExtractNums(string);\n"
+"if (!nums || nums.length < 2 || nums.length > 3)\n"
+"return null;\n"
+"var hour = nums[0];\n"
+"var min = nums[1];\n"
+"var sec = (nums.length == 3) ? nums[2] : 0;\n"
+"if (hour < 12 && (/pm/i).test(string))\n"
+"hour += 12;\n"
+"if (hour >= 12 && (/am/i).test(string))\n"
+"hour -= 12;\n"
+"date.setHours(hour, min, sec);\n"
+"if (date.getHours() != hour || date.getMinutes() != min || date.getSeconds() != sec)\n"
+"return null;\n"
+"return date;\n"
+"}\n"
 "\n"
-"if (!d)\n"
-"d = new Date();\n"
+"function AFMakeDate(out, year, month, date, time)\n"
+"{\n"
+"if (year < 50)\n"
+"year += 2000;\n"
+"if (year < 100)\n"
+"year += 1900;\n"
+"out.setFullYear(year, month, date);\n"
+"if (out.getFullYear() != year || out.getMonth() != month || out.getDate() != date)\n"
+"return null;\n"
+"if (time)\n"
+"out = AFParseTime(time, out);\n"
+"else\n"
+"out.setHours(0, 0, 0);\n"
+"return out;\n"
+"}\n"
 "\n"
-"var ampm = str.match(/(am|pm)/);\n"
-"var nums = str.match(/\\d+/g);\n"
-"var hour, min, sec;\n"
+"function AFParseDateEx(string, fmt) {\n"
+"var out = new Date();\n"
+"var year = out.getFullYear();\n"
+"var month;\n"
+"var date;\n"
+"var i;\n"
 "\n"
+"out.setHours(12, 0, 0);\n"
+"\n"
+"var order = AFParseDateOrder(fmt);\n"
+"\n"
+"var time = AFExtractTime(string);\n"
+"if (time) {\n"
+"string = time[0];\n"
+"time = time[1];\n"
+"}\n"
+"\n"
+"var nums = AFExtractNums(string);\n"
 "if (!nums)\n"
 "return null;\n"
 "\n"
-"sec = 0;\n"
-"\n"
-"switch (nums.length) {\n"
-"case 3:\n"
-"sec = parseInt(nums[2]);\n"
-"case 2:\n"
-"hour = parseInt(nums[0]);\n"
-"min = parseInt(nums[1]);\n"
-"break;\n"
-"default:\n"
-"return null;\n"
+"if (nums.length == 3) {\n"
+"year = nums[order.indexOf('y')];\n"
+"month = nums[order.indexOf('m')];\n"
+"date = nums[order.indexOf('d')];\n"
+"return AFMakeDate(out, year, month-1, date, time);\n"
 "}\n"
 "\n"
-"ampm = ampm && ampm[0]\n"
+"month = AFMatchMonth(string);\n"
 "\n"
-"if (ampm === 'am' && hour < 12)\n"
-"hour = 12 + hour;\n"
-"if (ampm === 'pm' && hour >= 12)\n"
-"hour = 0 + hour - 12;\n"
+"if (nums.length == 2) {\n"
 "\n"
-"d.setHours(hour, min, sec);\n"
-"\n"
-"if (d.getHours() !== hour || d.getMinutes() !== min || d.getSeconds() !== sec)\n"
-"return null;\n"
-"\n"
-"return d;\n"
+"if (month) {\n"
+"if (order.indexOf('y') < order.indexOf('d')) {\n"
+"year = nums[0];\n"
+"date = nums[1];\n"
+"} else {\n"
+"year = nums[1];\n"
+"date = nums[0];\n"
+"}\n"
 "}\n"
 "\n"
-"function AFParseDateEx(d, fmt) {\n"
-"var i;\n"
-"var dt = AFExtractTime(d);\n"
-"var nums = dt[0].match(/\\d+/g);\n"
-"var order = AFParseDateOrder(fmt);\n"
-"var text_month = AFMatchMonth(dt[0]);\n"
-"var dout = new Date();\n"
-"var year = dout.getFullYear();\n"
-"var month = dout.getMonth();\n"
-"var date = dout.getDate();\n"
 "\n"
-"dout.setHours(12, 0, 0);\n"
-"\n"
-"if (!nums || nums.length < 1 || nums.length > 3)\n"
-"return null;\n"
-"\n"
-"if (nums.length < 3 && text_month) {\n"
-"\n"
-"month = text_month;\n"
-"order = order.replace('m', '');\n"
+"else if (order.indexOf('y') < order.indexOf('d')) {\n"
+"if (order.indexOf('y') < order.indexOf('m')) {\n"
+"year = nums[0];\n"
+"month = nums[1];\n"
+"} else {\n"
+"year = nums[1];\n"
+"month = nums[0];\n"
+"}\n"
 "}\n"
 "\n"
-"order = order.substring(0, nums.length);\n"
 "\n"
+"else {\n"
+"if (order.indexOf('d') < order.indexOf('m')) {\n"
+"date = nums[0];\n"
+"month = nums[1];\n"
+"} else {\n"
+"date = nums[1];\n"
+"month = nums[0];\n"
+"}\n"
+"}\n"
 "\n"
-"if (order === 'ym' || (order === 'y' && text_month))\n"
+"return AFMakeDate(out, year, month-1, date, time);\n"
+"}\n"
+"\n"
+"if (nums.length == 1) {\n"
+"if (month) {\n"
+"if (order.indexOf('y') < order.indexOf('d')) {\n"
+"year = nums[0];\n"
 "date = 1;\n"
+"} else {\n"
+"date = nums[0];\n"
+"}\n"
+"return AFMakeDate(out, year, month-1, date, time);\n"
+"}\n"
 "\n"
-"for (i = 0; i < nums.length; i++) {\n"
-"switch (order.charAt(i)) {\n"
-"case 'y': year = parseInt(nums[i]); break;\n"
-"case 'm': month = parseInt(nums[i]) - 1; break;\n"
-"case 'd': date = parseInt(nums[i]); break;\n"
+"\n"
+"if (string.length == fmt.length) {\n"
+"year = month = date = '';\n"
+"for (i = 0; i < fmt.length; ++i) {\n"
+"switch (fmt.charAt(i)) {\n"
+"case '\\\\': ++i; break;\n"
+"case 'y': year += string.charAt(i); break;\n"
+"case 'm': month += string.charAt(i); break;\n"
+"case 'd': date += string.charAt(i); break;\n"
+"}\n"
+"}\n"
+"return AFMakeDate(out, year, month-1, date, time);\n"
 "}\n"
 "}\n"
 "\n"
-"if (year < 100) {\n"
-"if (fmt.search('yyyy') !== -1)\n"
 "return null;\n"
-"if (year >= 50)\n"
-"year = 1900 + year;\n"
-"else if (year >= 0)\n"
-"year = 2000 + year;\n"
 "}\n"
 "\n"
-"dout.setFullYear(year, month, date);\n"
-"\n"
-"if (dout.getFullYear() !== year || dout.getMonth() !== month || dout.getDate() !== date)\n"
-"return null;\n"
-"\n"
-"return AFParseTime(dt[1], dout);\n"
-"}\n"
+"var AFDate_oldFormats = [\n"
+"'m/d',\n"
+"'m/d/yy',\n"
+"'mm/dd/yy',\n"
+"'mm/yy',\n"
+"'d-mmm',\n"
+"'d-mmm-yy',\n"
+"'dd-mm-yy',\n"
+"'yy-mm-dd',\n"
+"'mmm-yy',\n"
+"'mmmm-yy',\n"
+"'mmm d, yyyy',\n"
+"'mmmm d, yyyy',\n"
+"'m/d/yy h:MM tt',\n"
+"'m/d/yy HH:MM'\n"
+"];\n"
 "\n"
 "function AFDate_KeystrokeEx(fmt) {\n"
 "if (event.willCommit && !AFParseDateEx(event.value, fmt)) {\n"
@@ -602,7 +504,7 @@
 "}\n"
 "\n"
 "function AFDate_Keystroke(index) {\n"
-"AFDate_KeystrokeEx(MuPDF.dateFormats[index]);\n"
+"AFDate_KeystrokeEx(AFDate_oldFormats[index]);\n"
 "}\n"
 "\n"
 "function AFDate_FormatEx(fmt) {\n"
@@ -611,7 +513,7 @@
 "}\n"
 "\n"
 "function AFDate_Format(index) {\n"
-"AFDate_FormatEx(MuPDF.dateFormats[index]);\n"
+"AFDate_FormatEx(AFDate_oldFormats[index]);\n"
 "}\n"
 "\n"
 "function AFTime_Keystroke(index) {\n"
@@ -627,25 +529,20 @@
 "}\n"
 "\n"
 "function AFTime_Format(index) {\n"
-"AFTime_FormatEx(MuPDF.timeFormats[index]);\n"
+"var formats = [ 'HH:MM', 'h:MM tt', 'HH:MM:ss', 'h:MM:ss tt' ];\n"
+"AFTime_FormatEx(formats[index]);\n"
 "}\n"
 "\n"
 "function AFSpecial_KeystrokeEx(fmt) {\n"
-"var cs = '=';\n"
+"function toUpper(str) { return str.toUpperCase(); }\n"
+"function toLower(str) { return str.toLowerCase(); }\n"
+"function toSame(str) { return str; }\n"
+"var convertCase = toSame;\n"
 "var val = event.value;\n"
 "var res = '';\n"
 "var i = 0;\n"
 "var m;\n"
 "var length = fmt ? fmt.length : 0;\n"
-"\n"
-"function convertCase(str, cmd) {\n"
-"switch (cmd) {\n"
-"case '>': return str.toUpperCase();\n"
-"case '<': return str.toLowerCase();\n"
-"default: return str;\n"
-"}\n"
-"}\n"
-"\n"
 "while (i < length) {\n"
 "switch (fmt.charAt(i)) {\n"
 "case '\\\\':\n"
@@ -663,7 +560,7 @@
 "event.rc = false;\n"
 "break;\n"
 "}\n"
-"res += convertCase(m[0], cs);\n"
+"res += convertCase(m[0]);\n"
 "val = val.substring(1);\n"
 "break;\n"
 "\n"
@@ -673,7 +570,7 @@
 "event.rc = false;\n"
 "break;\n"
 "}\n"
-"res += convertCase(m[0], cs);\n"
+"res += convertCase(m[0]);\n"
 "val = val.substring(1);\n"
 "break;\n"
 "\n"
@@ -688,23 +585,27 @@
 "break;\n"
 "\n"
 "case '*':\n"
-"res += val;\n"
+"res += convertCase(val);\n"
 "val = '';\n"
 "break;\n"
 "\n"
 "case '?':\n"
-"if (!val) {\n"
+"if (val === '') {\n"
 "event.rc = false;\n"
 "break;\n"
 "}\n"
-"res += convertCase(val.charAt(0), cs);\n"
+"res += convertCase(val.charAt(0));\n"
 "val = val.substring(1);\n"
 "break;\n"
 "\n"
 "case '=':\n"
+"convertCase = toSame;\n"
+"break;\n"
 "case '>':\n"
+"convertCase = toUpper;\n"
+"break;\n"
 "case '<':\n"
-"cs = fmt.charAt(i);\n"
+"convertCase = toLower;\n"
 "break;\n"
 "\n"
 "default:\n"
@@ -789,78 +690,31 @@
 "}\n"
 "\n"
 "function AFNumber_Format(nDec, sepStyle, negStyle, currStyle, strCurrency, bCurrencyPrepend) {\n"
-"var val = event.value;\n"
-"var fracpart;\n"
-"var intpart;\n"
-"var point = sepStyle&2 ? ',' : '.';\n"
-"var separator = sepStyle&2 ? '.' : ',';\n"
-"\n"
-"if (/^\\D*\\./.test(val))\n"
-"val = '0'+val;\n"
-"\n"
-"var groups = val.match(/\\d+/g);\n"
-"\n"
-"if (!groups)\n"
+"var value = AFMakeNumber(event.value);\n"
+"var fmt = '%,' + sepStyle + '.' + nDec + 'f';\n"
+"if (value == null) {\n"
+"event.value = '';\n"
 "return;\n"
-"\n"
-"switch (groups.length) {\n"
-"case 0:\n"
-"return;\n"
-"case 1:\n"
-"fracpart = '';\n"
-"intpart = groups[0];\n"
-"break;\n"
-"default:\n"
-"fracpart = groups.pop();\n"
-"intpart = groups.join('');\n"
-"break;\n"
 "}\n"
-"\n"
-"\n"
-"intpart = intpart.replace(/^0*/, '');\n"
-"if (!intpart)\n"
-"intpart = '0';\n"
-"\n"
-"if ((sepStyle & 1) === 0) {\n"
-"\n"
-"\n"
-"intpart = new Array(2-(intpart.length+2)%3+1).join('0') + intpart;\n"
-"intpart = intpart.match(/.../g).join(separator).replace(/^0*/, '');\n"
-"}\n"
-"\n"
-"if (!intpart)\n"
-"intpart = '0';\n"
-"\n"
-"\n"
-"fracpart += new Array(nDec+1).join('0');\n"
-"fracpart = fracpart.substring(0, nDec);\n"
-"\n"
-"if (fracpart)\n"
-"intpart += point+fracpart;\n"
-"\n"
 "if (bCurrencyPrepend)\n"
-"intpart = strCurrency+intpart;\n"
+"fmt = strCurrency + fmt;\n"
 "else\n"
-"intpart += strCurrency;\n"
-"\n"
-"if (/-/.test(val)) {\n"
-"switch (negStyle) {\n"
-"case 0:\n"
-"intpart = '-'+intpart;\n"
-"break;\n"
-"case 1:\n"
-"break;\n"
-"case 2:\n"
-"case 3:\n"
-"intpart = '('+intpart+')';\n"
-"break;\n"
+"fmt = fmt + strCurrency;\n"
+"if (value < 0) {\n"
+"/* negStyle: 0=MinusBlack, 1=Red, 2=ParensBlack, 3=ParensRed */\n"
+"value = Math.abs(value);\n"
+"if (negStyle == 2 || negStyle == 3)\n"
+"fmt = '(' + fmt + ')';\n"
+"else if (negStyle == 0)\n"
+"fmt = '-' + fmt;\n"
+"if (negStyle == 1 || negStyle == 3)\n"
+"event.target.textColor = color.red;\n"
+"else\n"
+"event.target.textColor = color.black;\n"
+"} else {\n"
+"event.target.textColor = color.black;\n"
 "}\n"
-"}\n"
-"\n"
-"if (negStyle&1)\n"
-"event.target.textColor = /-/.test(val) ? color.red : color.black;\n"
-"\n"
-"event.value = intpart;\n"
+"event.value = util.printf(fmt, value);\n"
 "}\n"
 "\n"
 "function AFPercent_Keystroke(nDec, sepStyle) {\n"
@@ -890,7 +744,7 @@
 "list = list.split(/ *, */);\n"
 "\n"
 "for (i = 0; i < list.length; i++) {\n"
-"var field = MuPDF_Doc.getField(list[i]);\n"
+"var field = this.getField(list[i]);\n"
 "var value = Number(field.value);\n"
 "switch (op) {\n"
 "case 'SUM': res += value; break;\n"
@@ -922,7 +776,24 @@
 "}\n"
 "}\n"
 "\n"
+"/* Compatibility ECMAScript functions */\n"
+"String.prototype.substr = function (start, length) {\n"
+"if (start < 0)\n"
+"start = this.length + start;\n"
+"if (length === undefined)\n"
+"return this.substring(start, this.length);\n"
+"return this.substring(start, start + length);\n"
+"}\n"
+"Date.prototype.getYear = Date.prototype.getFullYear;\n"
+"Date.prototype.setYear = Date.prototype.setFullYear;\n"
+"Date.prototype.toGMTString = Date.prototype.toUTCString;\n"
+"\n"
+"console.clear = function() { console.println('--- clear console ---\\n'); };\n"
+"console.show = function(){};\n"
+"console.hide = function(){};\n"
+"\n"
 "app.plugIns = [];\n"
 "app.viewerType = 'Reader';\n"
 "app.language = 'ENU';\n"
 "app.viewerVersion = NaN;\n"
+"app.execDialog = function () { return 'cancel'; }\n"
