@@ -27,6 +27,7 @@ struct gstate_s
 	float alpha[2];
 	fz_stroke_state *stroke_state;
 	int font;
+	float font_size;
 	int text_rendering_mode;
 	int knockout;
 };
@@ -212,7 +213,7 @@ pdf_dev_ctm(fz_context *ctx, pdf_device *pdev, fz_matrix ctm)
 }
 
 static void
-pdf_dev_color(fz_context *ctx, pdf_device *pdev, fz_colorspace *colorspace, const float *color, int stroke, const fz_color_params *color_params)
+pdf_dev_color(fz_context *ctx, pdf_device *pdev, fz_colorspace *colorspace, const float *color, int stroke, fz_color_params color_params)
 {
 	int diff = 0;
 	int i;
@@ -230,7 +231,7 @@ pdf_dev_color(fz_context *ctx, pdf_device *pdev, fz_colorspace *colorspace, cons
 	if (cspace == 0)
 	{
 		/* If it's an unknown colorspace, fallback to rgb */
-		fz_convert_color(ctx, color_params, NULL, fz_device_rgb(ctx), rgb, colorspace, color);
+		fz_convert_color(ctx, colorspace, color, fz_device_rgb(ctx), rgb, NULL, color_params);
 		color = rgb;
 		colorspace = fz_device_rgb(ctx);
 		cspace = 3;
@@ -301,7 +302,7 @@ pdf_dev_alpha(fz_context *ctx, pdf_device *pdev, float alpha, int stroke)
 			int newmax = pdev->max_alphas * 2;
 			if (newmax == 0)
 				newmax = 4;
-			pdev->alphas = fz_resize_array(ctx, pdev->alphas, newmax, sizeof(*pdev->alphas));
+			pdev->alphas = fz_realloc_array(ctx, pdev->alphas, newmax, alpha_entry);
 			pdev->max_alphas = newmax;
 		}
 		pdev->alphas[i].alpha = alpha;
@@ -355,7 +356,7 @@ pdf_dev_add_font_res(fz_context *ctx, pdf_device *pdev, fz_font *font)
 		int newmax = pdev->max_cid_fonts * 2;
 		if (newmax == 0)
 			newmax = 4;
-		pdev->cid_fonts = fz_resize_array(ctx, pdev->cid_fonts, newmax, sizeof(*pdev->cid_fonts));
+		pdev->cid_fonts = fz_realloc_array(ctx, pdev->cid_fonts, newmax, fz_font*);
 		pdev->max_cid_fonts = newmax;
 	}
 	num = pdev->num_cid_fonts++;
@@ -364,12 +365,13 @@ pdf_dev_add_font_res(fz_context *ctx, pdf_device *pdev, fz_font *font)
 }
 
 static void
-pdf_dev_font(fz_context *ctx, pdf_device *pdev, fz_font *font)
+pdf_dev_font(fz_context *ctx, pdf_device *pdev, fz_font *font, fz_matrix trm)
 {
 	gstate *gs = CURRENT_GSTATE(pdev);
+	float font_size = fz_matrix_expansion(trm);
 
 	/* If the font is unchanged, nothing to do */
-	if (gs->font >= 0 && pdev->cid_fonts[gs->font] == font)
+	if (gs->font >= 0 && pdev->cid_fonts[gs->font] == font && gs->font_size == font_size)
 		return;
 
 	if (fz_font_t3_procs(ctx, font))
@@ -380,8 +382,9 @@ pdf_dev_font(fz_context *ctx, pdf_device *pdev, fz_font *font)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "pdf device does not support font types found in this file");
 
 	gs->font = pdf_dev_add_font_res(ctx, pdev, font);
+	gs->font_size = font_size;
 
-	fz_append_printf(ctx, gs->buf, "/F%d 1 Tf\n", gs->font);
+	fz_append_printf(ctx, gs->buf, "/F%d %g Tf\n", gs->font, gs->font_size);
 }
 
 static void
@@ -390,8 +393,7 @@ pdf_dev_push_new_buf(fz_context *ctx, pdf_device *pdev, fz_buffer *buf, void (*o
 	if (pdev->num_gstates == pdev->max_gstates)
 	{
 		int newmax = pdev->max_gstates*2;
-
-		pdev->gstates = fz_resize_array(ctx, pdev->gstates, newmax, sizeof(*pdev->gstates));
+		pdev->gstates = fz_realloc_array(ctx, pdev->gstates, newmax, gstate);
 		pdev->max_gstates = newmax;
 	}
 	memcpy(&pdev->gstates[pdev->num_gstates], &pdev->gstates[pdev->num_gstates-1], sizeof(*pdev->gstates));
@@ -431,7 +433,8 @@ static void
 pdf_dev_text_span(fz_context *ctx, pdf_device *pdev, fz_text_span *span)
 {
 	gstate *gs = CURRENT_GSTATE(pdev);
-	fz_matrix tm, inv_tm;
+	fz_matrix trm, tm, tlm, inv_trm, inv_tm;
+	fz_matrix inv_tfs;
 	fz_point d;
 	float adv;
 	int dx, dy;
@@ -440,11 +443,17 @@ pdf_dev_text_span(fz_context *ctx, pdf_device *pdev, fz_text_span *span)
 	if (span->len == 0)
 		return;
 
-	tm = span->trm;
-	tm.e = span->items[0].x;
-	tm.f = span->items[0].y;
+	inv_tfs = fz_scale(1 / gs->font_size, 1 / gs->font_size);
+
+	trm = span->trm;
+	trm.e = span->items[0].x;
+	trm.f = span->items[0].y;
+
+	tm = fz_concat(inv_tfs, trm);
+	tlm = tm;
 
 	inv_tm = fz_invert_matrix(tm);
+	inv_trm = fz_invert_matrix(trm);
 
 	fz_append_printf(ctx, gs->buf, "%M Tm\n[<", &tm);
 
@@ -455,14 +464,14 @@ pdf_dev_text_span(fz_context *ctx, pdf_device *pdev, fz_text_span *span)
 			continue;
 
 		/* transform difference from expected pen position into font units. */
-		d.x = it->x - tm.e;
-		d.y = it->y - tm.f;
-		d = fz_transform_vector(d, inv_tm);
+		d.x = it->x - trm.e;
+		d.y = it->y - trm.f;
+		d = fz_transform_vector(d, inv_trm);
 		dx = (int)(d.x * 1000 + (d.x < 0 ? -0.5f : 0.5f));
 		dy = (int)(d.y * 1000 + (d.y < 0 ? -0.5f : 0.5f));
 
-		tm.e = it->x;
-		tm.f = it->y;
+		trm.e = it->x;
+		trm.f = it->y;
 
 		if (dx != 0 || dy != 0)
 		{
@@ -471,7 +480,15 @@ pdf_dev_text_span(fz_context *ctx, pdf_device *pdev, fz_text_span *span)
 			else if (span->wmode == 1 && dx == 0)
 				fz_append_printf(ctx, gs->buf, ">%d<", -dy);
 			else
-				fz_append_printf(ctx, gs->buf, ">]TJ\n%M Tm\n[<", &tm);
+			{
+				/* Calculate offset from start of the previous line */
+				tm = fz_concat(inv_tfs, trm);
+				d.x = tm.e - tlm.e;
+				d.y = tm.f - tlm.f;
+				d = fz_transform_vector(d, inv_tm);
+				fz_append_printf(ctx, gs->buf, ">]TJ\n%g %g Td\n[<", d.x, d.y);
+				tlm = tm;
+			}
 		}
 
 		if (fz_font_t3_procs(ctx, span->font))
@@ -481,9 +498,9 @@ pdf_dev_text_span(fz_context *ctx, pdf_device *pdev, fz_text_span *span)
 
 		adv = fz_advance_glyph(ctx, span->font, it->gid, span->wmode);
 		if (span->wmode == 0)
-			tm = fz_pre_translate(tm, adv, 0);
+			trm = fz_pre_translate(trm, adv, 0);
 		else
-			tm = fz_pre_translate(tm, 0, adv);
+			trm = fz_pre_translate(trm, 0, adv);
 	}
 
 	fz_append_string(ctx, gs->buf, ">]TJ\n");
@@ -501,7 +518,7 @@ pdf_dev_trm(fz_context *ctx, pdf_device *pdev, int trm)
 }
 
 static void
-pdf_dev_begin_text(fz_context *ctx, pdf_device *pdev, fz_matrix tm, int trm)
+pdf_dev_begin_text(fz_context *ctx, pdf_device *pdev, int trm)
 {
 	pdf_dev_trm(ctx, pdev, trm);
 	if (!pdev->in_text)
@@ -553,7 +570,7 @@ pdf_dev_new_form(fz_context *ctx, pdf_obj **form_ref, pdf_device *pdev, fz_rect
 			int newmax = pdev->max_groups * 2;
 			if (newmax == 0)
 				newmax = 4;
-			pdev->groups = fz_resize_array(ctx, pdev->groups, newmax, sizeof(*pdev->groups));
+			pdev->groups = fz_realloc_array(ctx, pdev->groups, newmax, group_entry);
 			pdev->max_groups = newmax;
 		}
 		pdev->num_groups++;
@@ -630,7 +647,7 @@ pdf_dev_new_form(fz_context *ctx, pdf_obj **form_ref, pdf_device *pdev, fz_rect
 
 static void
 pdf_dev_fill_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even_odd, fz_matrix ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+	fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	gstate *gs = CURRENT_GSTATE(pdev);
@@ -645,7 +662,7 @@ pdf_dev_fill_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even
 
 static void
 pdf_dev_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, const fz_stroke_state *stroke, fz_matrix ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+	fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	gstate *gs = CURRENT_GSTATE(pdev);
@@ -693,37 +710,38 @@ pdf_dev_clip_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, c
 
 static void
 pdf_dev_fill_text(fz_context *ctx, fz_device *dev, const fz_text *text, fz_matrix ctm,
-		fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+		fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	fz_text_span *span;
 
+	pdf_dev_ctm(ctx, pdev, ctm);
+	pdf_dev_alpha(ctx, pdev, alpha, 0);
+	pdf_dev_color(ctx, pdev, colorspace, color, 0, color_params);
+
 	for (span = text->head; span; span = span->next)
 	{
-		fz_matrix trm = span->trm;
-		pdf_dev_begin_text(ctx, pdev, trm, 0);
-		pdf_dev_font(ctx, pdev, span->font);
-		pdf_dev_ctm(ctx, pdev, ctm);
-		pdf_dev_alpha(ctx, pdev, alpha, 0);
-		pdf_dev_color(ctx, pdev, colorspace, color, 0, color_params);
+		pdf_dev_begin_text(ctx, pdev, 0);
+		pdf_dev_font(ctx, pdev, span->font, span->trm);
 		pdf_dev_text_span(ctx, pdev, span);
 	}
 }
 
 static void
 pdf_dev_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, fz_matrix ctm,
-		fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+		fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	fz_text_span *span;
 
+	pdf_dev_ctm(ctx, pdev, ctm);
+	pdf_dev_alpha(ctx, pdev, alpha, 1);
+	pdf_dev_color(ctx, pdev, colorspace, color, 1, color_params);
+
 	for (span = text->head; span; span = span->next)
 	{
-		pdf_dev_begin_text(ctx, pdev, span->trm, 1);
-		pdf_dev_font(ctx, pdev, span->font);
-		pdf_dev_ctm(ctx, pdev, ctm);
-		pdf_dev_alpha(ctx, pdev, alpha, 1);
-		pdf_dev_color(ctx, pdev, colorspace, color, 1, color_params);
+		pdf_dev_begin_text(ctx, pdev, 1);
+		pdf_dev_font(ctx, pdev, span->font, span->trm);
 		pdf_dev_text_span(ctx, pdev, span);
 	}
 }
@@ -737,11 +755,12 @@ pdf_dev_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, fz_matri
 	pdf_dev_end_text(ctx, pdev);
 	pdf_dev_push(ctx, pdev);
 
+	pdf_dev_ctm(ctx, pdev, ctm);
+
 	for (span = text->head; span; span = span->next)
 	{
-		pdf_dev_begin_text(ctx, pdev, span->trm, 7);
-		pdf_dev_ctm(ctx, pdev, ctm);
-		pdf_dev_font(ctx, pdev, span->font);
+		pdf_dev_begin_text(ctx, pdev, 7);
+		pdf_dev_font(ctx, pdev, span->font, span->trm);
 		pdf_dev_text_span(ctx, pdev, span);
 	}
 }
@@ -755,11 +774,12 @@ pdf_dev_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, c
 	pdf_dev_end_text(ctx, pdev);
 	pdf_dev_push(ctx, pdev);
 
+	pdf_dev_ctm(ctx, pdev, ctm);
+
 	for (span = text->head; span; span = span->next)
 	{
-		pdf_dev_begin_text(ctx, pdev, span->trm, 7);
-		pdf_dev_font(ctx, pdev, span->font);
-		pdf_dev_ctm(ctx, pdev, ctm);
+		pdf_dev_begin_text(ctx, pdev, 7);
+		pdf_dev_font(ctx, pdev, span->font, span->trm);
 		pdf_dev_text_span(ctx, pdev, span);
 	}
 }
@@ -769,11 +789,13 @@ pdf_dev_ignore_text(fz_context *ctx, fz_device *dev, const fz_text *text, fz_mat
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	fz_text_span *span;
+
+	pdf_dev_ctm(ctx, pdev, ctm);
+
 	for (span = text->head; span; span = span->next)
 	{
-		pdf_dev_begin_text(ctx, pdev, span->trm, 0);
-		pdf_dev_ctm(ctx, pdev, ctm);
-		pdf_dev_font(ctx, pdev, span->font);
+		pdf_dev_begin_text(ctx, pdev, 0);
+		pdf_dev_font(ctx, pdev, span->font, span->trm);
 		pdf_dev_text_span(ctx, pdev, span);
 	}
 }
@@ -803,7 +825,7 @@ pdf_dev_add_image_res(fz_context *ctx, fz_device *dev, pdf_obj *im_res)
 		int newmax = pdev->max_imgs * 2;
 		if (newmax == 0)
 			newmax = 4;
-		pdev->image_indices = fz_resize_array(ctx, pdev->image_indices, newmax, sizeof(*pdev->image_indices));
+		pdev->image_indices = fz_realloc_array(ctx, pdev->image_indices, newmax, int);
 		pdev->max_imgs = newmax;
 	}
 	num = pdev->num_imgs++;
@@ -811,7 +833,7 @@ pdf_dev_add_image_res(fz_context *ctx, fz_device *dev, pdf_obj *im_res)
 }
 
 static void
-pdf_dev_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, fz_matrix ctm, float alpha, const fz_color_params *color_params)
+pdf_dev_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, fz_matrix ctm, float alpha, fz_color_params color_params)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	pdf_obj *im_res;
@@ -845,7 +867,7 @@ pdf_dev_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, fz_matrix c
 }
 
 static void
-pdf_dev_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, fz_matrix ctm, float alpha, const fz_color_params *color_params)
+pdf_dev_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, fz_matrix ctm, float alpha, fz_color_params color_params)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 
@@ -855,7 +877,7 @@ pdf_dev_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, fz_matrix c
 
 static void
 pdf_dev_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, fz_matrix ctm,
-		fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+		fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	pdf_obj *im_res = NULL;
@@ -911,7 +933,7 @@ pdf_dev_pop_clip(fz_context *ctx, fz_device *dev)
 }
 
 static void
-pdf_dev_begin_mask(fz_context *ctx, fz_device *dev, fz_rect bbox, int luminosity, fz_colorspace *colorspace, const float *color, const fz_color_params *color_params)
+pdf_dev_begin_mask(fz_context *ctx, fz_device *dev, fz_rect bbox, int luminosity, fz_colorspace *colorspace, const float *color, fz_color_params color_params)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	gstate *gs;
@@ -936,6 +958,8 @@ pdf_dev_begin_mask(fz_context *ctx, fz_device *dev, fz_rect bbox, int luminosity
 	{
 		fz_snprintf(egsname, sizeof(egsname), "SM%d", pdev->num_smasks++);
 		egss = pdf_dict_get(ctx, pdev->resources, PDF_NAME(ExtGState));
+		if (!egss)
+			egss = pdf_dict_put_dict(ctx, pdev->resources, PDF_NAME(ExtGState), 10);
 		egs = pdf_dict_puts_dict(ctx, egss, egsname, 1);
 
 		pdf_dict_put(ctx, egs, PDF_NAME(Type), PDF_NAME(ExtGState));
