@@ -2,12 +2,17 @@
 #include "mupdf/pdf.h"
 #include "../fitz/fitz-imp.h"
 
-#include "pdf-name-table.h"
-
 #include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
 
+#define PDF_MAKE_NAME(STRING,NAME) STRING,
+static const char *PDF_NAME_LIST[] = {
+	"", "", "", /* dummy slots for null, true, and false */
+#include "mupdf/pdf/name-table.h"
+};
+#undef PDF_MAKE_NAME
+
 typedef enum pdf_objkind_e
 {
 	PDF_INT = 'i',
@@ -54,6 +59,7 @@ typedef struct pdf_obj_num_s
 typedef struct pdf_obj_string_s
 {
 	pdf_obj super;
+	char *text; /* utf8 encoded text string */
 	unsigned int len;
 	char buf[1];
 } pdf_obj_string;
@@ -100,19 +106,7 @@ typedef struct pdf_obj_ref_s
 #define REF(obj) ((pdf_obj_ref *)(obj))
 
 pdf_obj *
-pdf_new_null(fz_context *ctx, pdf_document *doc)
-{
-	return PDF_OBJ_NULL;
-}
-
-pdf_obj *
-pdf_new_bool(fz_context *ctx, pdf_document *doc, int b)
-{
-	return b ? PDF_OBJ_TRUE : PDF_OBJ_FALSE;
-}
-
-pdf_obj *
-pdf_new_int(fz_context *ctx, pdf_document *doc, int64_t i)
+pdf_new_int(fz_context *ctx, int64_t i)
 {
 	pdf_obj_num *obj;
 	obj = Memento_label(fz_malloc(ctx, sizeof(pdf_obj_num)), "pdf_obj(int)");
@@ -124,7 +118,7 @@ pdf_new_int(fz_context *ctx, pdf_document *doc, int64_t i)
 }
 
 pdf_obj *
-pdf_new_real(fz_context *ctx, pdf_document *doc, float f)
+pdf_new_real(fz_context *ctx, float f)
 {
 	pdf_obj_num *obj;
 	obj = Memento_label(fz_malloc(ctx, sizeof(pdf_obj_num)), "pdf_obj(real)");
@@ -136,7 +130,7 @@ pdf_new_real(fz_context *ctx, pdf_document *doc, float f)
 }
 
 pdf_obj *
-pdf_new_string(fz_context *ctx, pdf_document *doc, const char *str, size_t len)
+pdf_new_string(fz_context *ctx, const char *str, size_t len)
 {
 	pdf_obj_string *obj;
 	unsigned int l = (unsigned int)len;
@@ -148,27 +142,30 @@ pdf_new_string(fz_context *ctx, pdf_document *doc, const char *str, size_t len)
 	obj->super.refs = 1;
 	obj->super.kind = PDF_STRING;
 	obj->super.flags = 0;
+	obj->text = NULL;
 	obj->len = l;
 	memcpy(obj->buf, str, len);
 	obj->buf[len] = '\0';
 	return &obj->super;
 }
 
-static int
-namecmp(const void *key, const void *name)
-{
-	return strcmp((char *)key, *(char **)name);
-}
-
 pdf_obj *
-pdf_new_name(fz_context *ctx, pdf_document *doc, const char *str)
+pdf_new_name(fz_context *ctx, const char *str)
 {
 	pdf_obj_name *obj;
-	const char **stdname;
-
-	stdname = bsearch(str, &PDF_NAMES[1], PDF_OBJ_ENUM_NAME__LIMIT-1, sizeof(char *), namecmp);
-	if (stdname != NULL)
-		return (pdf_obj *)(intptr_t)(stdname - &PDF_NAMES[0]);
+	int l = 3; /* skip dummy slots */
+	int r = nelem(PDF_NAME_LIST) - 1;
+	while (l <= r)
+	{
+		int m = (l + r) >> 1;
+		int c = strcmp(str, PDF_NAME_LIST[m]);
+		if (c < 0)
+			r = m - 1;
+		else if (c > 0)
+			l = m + 1;
+		else
+			return (pdf_obj*)(intptr_t)m;
+	}
 
 	obj = Memento_label(fz_malloc(ctx, offsetof(pdf_obj_name, n) + strlen(str) + 1), "pdf_obj(name)");
 	obj->super.refs = 1;
@@ -192,36 +189,23 @@ pdf_new_indirect(fz_context *ctx, pdf_document *doc, int num, int gen)
 	return &obj->super;
 }
 
-pdf_obj *
-pdf_keep_obj(fz_context *ctx, pdf_obj *obj)
-{
-	if (obj >= PDF_OBJ__LIMIT)
-	{
-		(void)fz_keep_imp16(ctx, obj, &obj->refs);
-	}
-	return obj;
-}
-
-#define OBJ_IS_NULL(obj) \
-	(obj == PDF_OBJ_NULL)
-#define OBJ_IS_BOOL(obj) \
-	(obj == PDF_OBJ_TRUE || obj == PDF_OBJ_FALSE)
+#define OBJ_IS_NULL(obj) (obj == PDF_NULL)
+#define OBJ_IS_BOOL(obj) (obj == PDF_TRUE || obj == PDF_FALSE)
+#define OBJ_IS_NAME(obj) ((obj > PDF_FALSE && obj < PDF_LIMIT) || (obj >= PDF_LIMIT && obj->kind == PDF_NAME))
 #define OBJ_IS_INT(obj) \
-	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_INT : 0)
+	(obj >= PDF_LIMIT && obj->kind == PDF_INT)
 #define OBJ_IS_REAL(obj) \
-	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_REAL : 0)
+	(obj >= PDF_LIMIT && obj->kind == PDF_REAL)
 #define OBJ_IS_NUMBER(obj) \
-	(obj >= PDF_OBJ__LIMIT ? (obj->kind == PDF_REAL || obj->kind == PDF_INT) : 0)
+	(obj >= PDF_LIMIT && (obj->kind == PDF_REAL || obj->kind == PDF_INT))
 #define OBJ_IS_STRING(obj) \
-	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_STRING : 0)
-#define OBJ_IS_NAME(obj) \
-	((obj >= PDF_OBJ__LIMIT) ? (obj->kind == PDF_NAME) : (obj != NULL && obj < PDF_OBJ_NAME__LIMIT))
+	(obj >= PDF_LIMIT && obj->kind == PDF_STRING)
 #define OBJ_IS_ARRAY(obj) \
-	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_ARRAY : 0)
+	(obj >= PDF_LIMIT && obj->kind == PDF_ARRAY)
 #define OBJ_IS_DICT(obj) \
-	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_DICT : 0)
+	(obj >= PDF_LIMIT && obj->kind == PDF_DICT)
 #define OBJ_IS_INDIRECT(obj) \
-	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_INDIRECT : 0)
+	(obj >= PDF_LIMIT && obj->kind == PDF_INDIRECT)
 
 #define RESOLVE(obj) \
 	if (OBJ_IS_INDIRECT(obj)) \
@@ -286,16 +270,17 @@ int pdf_is_dict(fz_context *ctx, pdf_obj *obj)
 	return OBJ_IS_DICT(obj);
 }
 
+/* safe, silent failure, no error reporting on type mismatches */
 int pdf_to_bool(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	return obj == PDF_OBJ_TRUE;
+	return obj == PDF_TRUE;
 }
 
 int pdf_to_int(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return 0;
 	if (obj->kind == PDF_INT)
 		return (int)NUM(obj)->u.i;
@@ -307,7 +292,7 @@ int pdf_to_int(fz_context *ctx, pdf_obj *obj)
 int64_t pdf_to_int64(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return 0;
 	if (obj->kind == PDF_INT)
 		return NUM(obj)->u.i;
@@ -319,7 +304,7 @@ int64_t pdf_to_int64(fz_context *ctx, pdf_obj *obj)
 float pdf_to_real(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return 0;
 	if (obj->kind == PDF_REAL)
 		return NUM(obj)->u.f;
@@ -331,34 +316,59 @@ float pdf_to_real(fz_context *ctx, pdf_obj *obj)
 const char *pdf_to_name(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (!OBJ_IS_NAME(obj))
-		return "";
-	if (obj < PDF_OBJ_NAME__LIMIT)
-		return PDF_NAMES[(intptr_t)obj];
-	return NAME(obj)->n;
+	if (obj < PDF_LIMIT)
+		return PDF_NAME_LIST[((intptr_t)obj)];
+	if (obj->kind == PDF_NAME)
+		return NAME(obj)->n;
+	return "";
 }
 
 char *pdf_to_str_buf(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (!OBJ_IS_STRING(obj))
-		return "";
-	return STRING(obj)->buf;
+	if (OBJ_IS_STRING(obj))
+		return STRING(obj)->buf;
+	return "";
 }
 
 int pdf_to_str_len(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (!OBJ_IS_STRING(obj))
-		return 0;
-	return STRING(obj)->len;
+	if (OBJ_IS_STRING(obj))
+		return STRING(obj)->len;
+	return 0;
+}
+
+const char *pdf_to_string(fz_context *ctx, pdf_obj *obj, size_t *sizep)
+{
+	RESOLVE(obj);
+	if (OBJ_IS_STRING(obj))
+	{
+		if (sizep)
+			*sizep = STRING(obj)->len;
+		return STRING(obj)->buf;
+	}
+	if (sizep)
+		*sizep = 0;
+	return "";
+}
+
+const char *pdf_to_text_string(fz_context *ctx, pdf_obj *obj)
+{
+	RESOLVE(obj);
+	if (OBJ_IS_STRING(obj))
+	{
+		if (!STRING(obj)->text)
+			STRING(obj)->text = pdf_new_utf8_from_pdf_string(ctx, STRING(obj)->buf, STRING(obj)->len);
+		return STRING(obj)->text;
+	}
+	return "";
 }
 
 void pdf_set_int(fz_context *ctx, pdf_obj *obj, int64_t i)
 {
-	if (!OBJ_IS_INT(obj))
-		return;
-	NUM(obj)->u.i = i;
+	if (OBJ_IS_INT(obj))
+		NUM(obj)->u.i = i;
 }
 
 /* for use by pdf_crypt_obj_imp to decrypt AES string in place */
@@ -369,33 +379,34 @@ void pdf_set_str_len(fz_context *ctx, pdf_obj *obj, int newlen)
 		return; /* This should never happen */
 	if (newlen < 0 || (unsigned int)newlen > STRING(obj)->len)
 		return; /* This should never happen */
+	STRING(obj)->buf[newlen] = 0;
 	STRING(obj)->len = newlen;
 }
 
 int pdf_to_num(fz_context *ctx, pdf_obj *obj)
 {
-	if (!OBJ_IS_INDIRECT(obj))
-		return 0;
-	return REF(obj)->num;
+	if (OBJ_IS_INDIRECT(obj))
+		return REF(obj)->num;
+	return 0;
 }
 
 int pdf_to_gen(fz_context *ctx, pdf_obj *obj)
 {
-	if (!OBJ_IS_INDIRECT(obj))
-		return 0;
-	return REF(obj)->gen;
+	if (OBJ_IS_INDIRECT(obj))
+		return REF(obj)->gen;
+	return 0;
 }
 
 pdf_document *pdf_get_indirect_document(fz_context *ctx, pdf_obj *obj)
 {
-	if (!OBJ_IS_INDIRECT(obj))
-		return NULL;
-	return REF(obj)->doc;
+	if (OBJ_IS_INDIRECT(obj))
+		return REF(obj)->doc;
+	return NULL;
 }
 
 pdf_document *pdf_get_bound_document(fz_context *ctx, pdf_obj *obj)
 {
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return NULL;
 	if (obj->kind == PDF_INDIRECT)
 		return REF(obj)->doc;
@@ -421,32 +432,29 @@ pdf_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b)
 	if (a == b)
 		return 0;
 
-	if (!a || !b)
+	/* a or b is null, true, or false */
+	if (a <= PDF_FALSE || b <= PDF_FALSE)
 		return 1;
 
-	if (a < PDF_OBJ_NAME__LIMIT)
+	/* a is a constant name */
+	if (a < PDF_LIMIT)
 	{
-		if (b < PDF_OBJ_NAME__LIMIT)
+		if (b < PDF_LIMIT)
 			return a != b;
-		if (b < PDF_OBJ__LIMIT)
-			return 1;
 		if (b->kind != PDF_NAME)
 			return 1;
-		return strcmp(NAME(b)->n, PDF_NAMES[(intptr_t)a]);
+		return strcmp(PDF_NAME_LIST[(intptr_t)a], NAME(b)->n);
 	}
 
-	if (b < PDF_OBJ_NAME__LIMIT)
+	/* b is a constant name */
+	if (b < PDF_LIMIT)
 	{
-		if (a < PDF_OBJ__LIMIT)
-			return 1;
 		if (a->kind != PDF_NAME)
 			return 1;
-		return strcmp(NAME(a)->n, PDF_NAMES[(intptr_t)b]);
+		return strcmp(NAME(a)->n, PDF_NAME_LIST[(intptr_t)b]);
 	}
 
-	if (a < PDF_OBJ__LIMIT || b < PDF_OBJ__LIMIT)
-		return a != b;
-
+	/* both a and b are allocated objects */
 	if (a->kind != b->kind)
 		return 1;
 
@@ -508,18 +516,28 @@ pdf_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b)
 	return 1;
 }
 
+int pdf_name_eq(fz_context *ctx, pdf_obj *a, pdf_obj *b)
+{
+	RESOLVE(a);
+	RESOLVE(b);
+	if (a <= PDF_FALSE || b <= PDF_FALSE)
+		return 0;
+	if (a < PDF_LIMIT || b < PDF_LIMIT)
+		return (a == b);
+	if (a->kind == PDF_NAME && b->kind == PDF_NAME)
+		return !strcmp(NAME(a)->n, NAME(b)->n);
+	return 0;
+}
+
 static char *
 pdf_objkindstr(pdf_obj *obj)
 {
-	if (!obj)
-		return "<NULL>";
-	if (obj < PDF_OBJ_NAME__LIMIT)
-		return "name";
-	if (obj == PDF_OBJ_TRUE || obj == PDF_OBJ_FALSE)
-		return "boolean";
-	if (obj == PDF_OBJ_NULL)
+	if (obj == PDF_NULL)
 		return "null";
-
+	if (obj == PDF_TRUE || obj == PDF_FALSE)
+		return "boolean";
+	if (obj < PDF_LIMIT)
+		return "name";
 	switch (obj->kind)
 	{
 	case PDF_INT: return "integer";
@@ -551,7 +569,7 @@ pdf_new_array(fz_context *ctx, pdf_document *doc, int initialcap)
 
 	fz_try(ctx)
 	{
-		obj->items = Memento_label(fz_malloc_array(ctx, obj->cap, sizeof(pdf_obj*)), "pdf_obj(array items)");
+		obj->items = fz_malloc_array(ctx, obj->cap, pdf_obj*);
 	}
 	fz_catch(ctx)
 	{
@@ -570,7 +588,7 @@ pdf_array_grow(fz_context *ctx, pdf_obj_array *obj)
 	int i;
 	int new_cap = (obj->cap * 3) / 2;
 
-	obj->items = fz_resize_array(ctx, obj->items, new_cap, sizeof(pdf_obj*));
+	obj->items = fz_realloc_array(ctx, obj->items, new_cap, pdf_obj*);
 	obj->cap = new_cap;
 
 	for (i = obj->len ; i < obj->cap; i++)
@@ -634,7 +652,7 @@ static void prepare_object_for_alteration(fz_context *ctx, pdf_obj *obj, pdf_obj
 		obj should be a dict or an array. We don't care about
 		any other types, as they aren't 'containers'.
 	*/
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return;
 
 	switch (obj->kind)
@@ -687,10 +705,6 @@ pdf_array_put(fz_context *ctx, pdf_obj *obj, int i, pdf_obj *item)
 	}
 	if (i < 0 || i > ARRAY(obj)->len)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "index out of bounds");
-
-	if (!item)
-		item = PDF_OBJ_NULL;
-
 	prepare_object_for_alteration(ctx, obj, item);
 	pdf_drop_obj(ctx, ARRAY(obj)->items[i]);
 	ARRAY(obj)->items[i] = pdf_keep_obj(ctx, item);
@@ -713,10 +727,6 @@ pdf_array_push(fz_context *ctx, pdf_obj *obj, pdf_obj *item)
 	RESOLVE(obj);
 	if (!OBJ_IS_ARRAY(obj))
 		fz_throw(ctx, FZ_ERROR_GENERIC, "not an array (%s)", pdf_objkindstr(obj));
-
-	if (!item)
-		item = PDF_OBJ_NULL;
-
 	prepare_object_for_alteration(ctx, obj, item);
 	if (ARRAY(obj)->len + 1 > ARRAY(obj)->cap)
 		pdf_array_grow(ctx, ARRAY(obj));
@@ -727,16 +737,12 @@ pdf_array_push(fz_context *ctx, pdf_obj *obj, pdf_obj *item)
 void
 pdf_array_push_drop(fz_context *ctx, pdf_obj *obj, pdf_obj *item)
 {
-	RESOLVE(obj);
-	if (obj >= PDF_OBJ__LIMIT)
-	{
-		fz_try(ctx)
-			pdf_array_push(ctx, obj, item);
-		fz_always(ctx)
-			pdf_drop_obj(ctx, item);
-		fz_catch(ctx)
-			fz_rethrow(ctx);
-	}
+	fz_try(ctx)
+		pdf_array_push(ctx, obj, item);
+	fz_always(ctx)
+		pdf_drop_obj(ctx, item);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 void
@@ -747,10 +753,6 @@ pdf_array_insert(fz_context *ctx, pdf_obj *obj, pdf_obj *item, int i)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "not an array (%s)", pdf_objkindstr(obj));
 	if (i < 0 || i > ARRAY(obj)->len)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "index out of bounds");
-
-	if (!item)
-		item = PDF_OBJ_NULL;
-
 	prepare_object_for_alteration(ctx, obj, item);
 	if (ARRAY(obj)->len + 1 > ARRAY(obj)->cap)
 		pdf_array_grow(ctx, ARRAY(obj));
@@ -762,16 +764,12 @@ pdf_array_insert(fz_context *ctx, pdf_obj *obj, pdf_obj *item, int i)
 void
 pdf_array_insert_drop(fz_context *ctx, pdf_obj *obj, pdf_obj *item, int i)
 {
-	RESOLVE(obj);
-	if (obj >= PDF_OBJ__LIMIT)
-	{
-		fz_try(ctx)
-			pdf_array_insert(ctx, obj, item, i);
-		fz_always(ctx)
-			pdf_drop_obj(ctx, item);
-		fz_catch(ctx)
-			fz_rethrow(ctx);
-	}
+	fz_try(ctx)
+		pdf_array_insert(ctx, obj, item, i);
+	fz_always(ctx)
+		pdf_drop_obj(ctx, item);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 void
@@ -782,7 +780,6 @@ pdf_array_delete(fz_context *ctx, pdf_obj *obj, int i)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "not an array (%s)", pdf_objkindstr(obj));
 	if (i < 0 || i >= ARRAY(obj)->len)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "index out of bounds");
-
 	prepare_object_for_alteration(ctx, obj, NULL);
 	pdf_drop_obj(ctx, ARRAY(obj)->items[i]);
 	ARRAY(obj)->items[i] = 0;
@@ -816,15 +813,15 @@ pdf_array_find(fz_context *ctx, pdf_obj *arr, pdf_obj *obj)
 	return -1;
 }
 
-pdf_obj *pdf_new_rect(fz_context *ctx, pdf_document *doc, const fz_rect *rect)
+pdf_obj *pdf_new_rect(fz_context *ctx, pdf_document *doc, fz_rect rect)
 {
 	pdf_obj *arr = pdf_new_array(ctx, doc, 4);
 	fz_try(ctx)
 	{
-		pdf_array_push_real(ctx, arr, rect->x0);
-		pdf_array_push_real(ctx, arr, rect->y0);
-		pdf_array_push_real(ctx, arr, rect->x1);
-		pdf_array_push_real(ctx, arr, rect->y1);
+		pdf_array_push_real(ctx, arr, rect.x0);
+		pdf_array_push_real(ctx, arr, rect.y0);
+		pdf_array_push_real(ctx, arr, rect.x1);
+		pdf_array_push_real(ctx, arr, rect.y1);
 	}
 	fz_catch(ctx)
 	{
@@ -834,17 +831,17 @@ pdf_obj *pdf_new_rect(fz_context *ctx, pdf_document *doc, const fz_rect *rect)
 	return arr;
 }
 
-pdf_obj *pdf_new_matrix(fz_context *ctx, pdf_document *doc, const fz_matrix *mtx)
+pdf_obj *pdf_new_matrix(fz_context *ctx, pdf_document *doc, fz_matrix mtx)
 {
 	pdf_obj *arr = pdf_new_array(ctx, doc, 6);
 	fz_try(ctx)
 	{
-		pdf_array_push_real(ctx, arr, mtx->a);
-		pdf_array_push_real(ctx, arr, mtx->b);
-		pdf_array_push_real(ctx, arr, mtx->c);
-		pdf_array_push_real(ctx, arr, mtx->d);
-		pdf_array_push_real(ctx, arr, mtx->e);
-		pdf_array_push_real(ctx, arr, mtx->f);
+		pdf_array_push_real(ctx, arr, mtx.a);
+		pdf_array_push_real(ctx, arr, mtx.b);
+		pdf_array_push_real(ctx, arr, mtx.c);
+		pdf_array_push_real(ctx, arr, mtx.d);
+		pdf_array_push_real(ctx, arr, mtx.e);
+		pdf_array_push_real(ctx, arr, mtx.f);
 	}
 	fz_catch(ctx)
 	{
@@ -865,16 +862,16 @@ static int keyvalcmp(const void *ap, const void *bp)
 
 	/* We should never get a->k == NULL or b->k == NULL. If we
 	 * do, then they match. */
-	if (a->k < PDF_OBJ_NAME__LIMIT)
-		an = PDF_NAMES[(intptr_t)a->k];
-	else if (a->k >= PDF_OBJ__LIMIT && a->k->kind == PDF_NAME)
+	if (a->k < PDF_LIMIT)
+		an = PDF_NAME_LIST[(intptr_t)a->k];
+	else if (a->k >= PDF_LIMIT && a->k->kind == PDF_NAME)
 		an = NAME(a->k)->n;
 	else
 		return 0;
 
-	if (b->k < PDF_OBJ_NAME__LIMIT)
-		bn = PDF_NAMES[(intptr_t)b->k];
-	else if (b->k >= PDF_OBJ__LIMIT && b->k->kind == PDF_NAME)
+	if (b->k < PDF_LIMIT)
+		bn = PDF_NAME_LIST[(intptr_t)b->k];
+	else if (b->k >= PDF_LIMIT && b->k->kind == PDF_NAME)
 		bn = NAME(b->k)->n;
 	else
 		return 0;
@@ -900,7 +897,7 @@ pdf_new_dict(fz_context *ctx, pdf_document *doc, int initialcap)
 
 	fz_try(ctx)
 	{
-		DICT(obj)->items = Memento_label(fz_malloc_array(ctx, DICT(obj)->cap, sizeof(struct keyval)), "pdf_obj(dict items)");
+		DICT(obj)->items = fz_malloc_array(ctx, DICT(obj)->cap, struct keyval);
 	}
 	fz_catch(ctx)
 	{
@@ -922,7 +919,7 @@ pdf_dict_grow(fz_context *ctx, pdf_obj *obj)
 	int i;
 	int new_cap = (DICT(obj)->cap * 3) / 2;
 
-	DICT(obj)->items = fz_resize_array(ctx, DICT(obj)->items, new_cap, sizeof(struct keyval));
+	DICT(obj)->items = fz_realloc_array(ctx, DICT(obj)->items, new_cap, struct keyval);
 	DICT(obj)->cap = new_cap;
 
 	for (i = DICT(obj)->len; i < DICT(obj)->cap; i++)
@@ -1000,7 +997,7 @@ pdf_dict_put_val_null(fz_context *ctx, pdf_obj *obj, int idx)
 
 	prepare_object_for_alteration(ctx, obj, NULL);
 	pdf_drop_obj(ctx, DICT(obj)->items[idx].v);
-	DICT(obj)->items[idx].v = PDF_OBJ_NULL;
+	DICT(obj)->items[idx].v = PDF_NULL;
 }
 
 /* Returns 0 <= i < len for key found. Returns -1-len < i <= -1 for key
@@ -1054,7 +1051,7 @@ pdf_dict_find(fz_context *ctx, pdf_obj *obj, pdf_obj *key)
 		int r = len - 1;
 		pdf_obj *k = DICT(obj)->items[r].k;
 
-		if (k == key || (k >= PDF_OBJ__LIMIT && strcmp(NAME(k)->n, PDF_NAMES[(intptr_t)key]) < 0))
+		if (k == key || (k >= PDF_LIMIT && strcmp(NAME(k)->n, PDF_NAME_LIST[(intptr_t)key]) < 0))
 		{
 			return -1 - (r+1);
 		}
@@ -1065,7 +1062,7 @@ pdf_dict_find(fz_context *ctx, pdf_obj *obj, pdf_obj *key)
 			int c;
 
 			k = DICT(obj)->items[m].k;
-			c = (k < PDF_OBJ__LIMIT ? (char *)key-(char *)k : -strcmp(NAME(k)->n, PDF_NAMES[(intptr_t)key]));
+			c = (k < PDF_LIMIT ? (char *)key-(char *)k : -strcmp(NAME(k)->n, PDF_NAME_LIST[(intptr_t)key]));
 			if (c < 0)
 				r = m - 1;
 			else if (c > 0)
@@ -1081,14 +1078,14 @@ pdf_dict_find(fz_context *ctx, pdf_obj *obj, pdf_obj *key)
 		for (i = 0; i < len; i++)
 		{
 			pdf_obj *k = DICT(obj)->items[i].k;
-			if (k < PDF_OBJ__LIMIT)
+			if (k < PDF_LIMIT)
 			{
 				if (k == key)
 					return i;
 			}
 			else
 			{
-				if (!strcmp(PDF_NAMES[(intptr_t)key], NAME(k)->n))
+				if (!strcmp(PDF_NAME_LIST[(intptr_t)key], NAME(k)->n))
 					return i;
 			}
 		}
@@ -1175,7 +1172,7 @@ pdf_dict_get(fz_context *ctx, pdf_obj *obj, pdf_obj *key)
 	if (!OBJ_IS_NAME(key))
 		return NULL;
 
-	if (key < PDF_OBJ_NAME__LIMIT)
+	if (key < PDF_LIMIT)
 		i = pdf_dict_find(ctx, obj, key);
 	else
 		i = pdf_dict_finds(ctx, obj, pdf_to_name(ctx, key));
@@ -1218,13 +1215,10 @@ pdf_dict_get_put(fz_context *ctx, pdf_obj *obj, pdf_obj *key, pdf_obj *val, pdf_
 	if (!OBJ_IS_NAME(key))
 		fz_throw(ctx, FZ_ERROR_GENERIC, "key is not a name (%s)", pdf_objkindstr(obj));
 
-	if (!val)
-		val = PDF_OBJ_NULL;
-
 	if (DICT(obj)->len > 100 && !(obj->flags & PDF_FLAGS_SORTED))
 		pdf_sort_dict(ctx, obj);
 
-	if (key < PDF_OBJ_NAME__LIMIT)
+	if (key < PDF_LIMIT)
 		i = pdf_dict_find(ctx, obj, key);
 	else
 		i = pdf_dict_finds(ctx, obj, pdf_to_name(ctx, key));
@@ -1291,15 +1285,13 @@ pdf_dict_get_put_drop(fz_context *ctx, pdf_obj *obj, pdf_obj *key, pdf_obj *val,
 void
 pdf_dict_puts(fz_context *ctx, pdf_obj *obj, const char *key, pdf_obj *val)
 {
-	pdf_document *doc;
 	pdf_obj *keyobj;
 
 	RESOLVE(obj);
 	if (!OBJ_IS_DICT(obj))
 		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dict (%s)", pdf_objkindstr(obj));
 
-	doc = DICT(obj)->doc;
-	keyobj = pdf_new_name(ctx, doc, key);
+	keyobj = pdf_new_name(ctx, key);
 
 	fz_try(ctx)
 		pdf_dict_put(ctx, obj, keyobj, val);
@@ -1312,15 +1304,13 @@ pdf_dict_puts(fz_context *ctx, pdf_obj *obj, const char *key, pdf_obj *val)
 void
 pdf_dict_puts_drop(fz_context *ctx, pdf_obj *obj, const char *key, pdf_obj *val)
 {
-	pdf_document *doc;
 	pdf_obj *keyobj;
 
 	RESOLVE(obj);
 	if (!OBJ_IS_DICT(obj))
 		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dict (%s)", pdf_objkindstr(obj));
 
-	doc = DICT(obj)->doc;
-	keyobj = pdf_new_name(ctx, doc, key);
+	keyobj = pdf_new_name(ctx, key);
 
 	fz_var(keyobj);
 
@@ -1512,8 +1502,8 @@ pdf_dict_del(fz_context *ctx, pdf_obj *obj, pdf_obj *key)
 	if (!OBJ_IS_NAME(key))
 		fz_throw(ctx, FZ_ERROR_GENERIC, "key is not a name (%s)", pdf_objkindstr(key));
 
-	if (key < PDF_OBJ_NAME__LIMIT)
-		pdf_dict_dels(ctx, obj, PDF_NAMES[(intptr_t)key]);
+	if (key < PDF_LIMIT)
+		pdf_dict_dels(ctx, obj, PDF_NAME_LIST[(intptr_t)key]);
 	else
 		pdf_dict_dels(ctx, obj, NAME(key)->n);
 }
@@ -1534,9 +1524,9 @@ pdf_sort_dict(fz_context *ctx, pdf_obj *obj)
 pdf_obj *
 pdf_deep_copy_obj(fz_context *ctx, pdf_obj *obj)
 {
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 	{
-		return pdf_keep_obj(ctx, obj);
+		return obj;
 	}
 	if (obj->kind == PDF_DICT)
 	{
@@ -1586,11 +1576,12 @@ pdf_deep_copy_obj(fz_context *ctx, pdf_obj *obj)
 	}
 }
 
+/* obj marking and unmarking functions - to avoid infinite recursions. */
 int
 pdf_obj_marked(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return 0;
 	return !!(obj->flags & PDF_FLAGS_MARKED);
 }
@@ -1600,7 +1591,7 @@ pdf_mark_obj(fz_context *ctx, pdf_obj *obj)
 {
 	int marked;
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return 0;
 	marked = !!(obj->flags & PDF_FLAGS_MARKED);
 	obj->flags |= PDF_FLAGS_MARKED;
@@ -1611,7 +1602,7 @@ void
 pdf_unmark_obj(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return;
 	obj->flags &= ~PDF_FLAGS_MARKED;
 }
@@ -1619,9 +1610,8 @@ pdf_unmark_obj(fz_context *ctx, pdf_obj *obj)
 void
 pdf_set_obj_memo(fz_context *ctx, pdf_obj *obj, int bit, int memo)
 {
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return;
-
 	bit <<= 1;
 	obj->flags |= PDF_FLAGS_MEMO_BASE << bit;
 	if (memo)
@@ -1633,7 +1623,7 @@ pdf_set_obj_memo(fz_context *ctx, pdf_obj *obj, int bit, int memo)
 int
 pdf_obj_memo(fz_context *ctx, pdf_obj *obj, int bit, int *memo)
 {
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return 0;
 	bit <<= 1;
 	if (!(obj->flags & (PDF_FLAGS_MEMO_BASE<<bit)))
@@ -1642,10 +1632,11 @@ pdf_obj_memo(fz_context *ctx, pdf_obj *obj, int bit, int *memo)
 	return 1;
 }
 
+/* obj dirty bit support. */
 int pdf_obj_is_dirty(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return 0;
 	return !!(obj->flags & PDF_FLAGS_DIRTY);
 }
@@ -1653,14 +1644,15 @@ int pdf_obj_is_dirty(fz_context *ctx, pdf_obj *obj)
 void pdf_dirty_obj(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return;
 	obj->flags |= PDF_FLAGS_DIRTY;
 }
 
 void pdf_clean_obj(fz_context *ctx, pdf_obj *obj)
 {
-	if (obj < PDF_OBJ__LIMIT)
+	RESOLVE(obj);
+	if (obj < PDF_LIMIT)
 		return;
 	obj->flags &= ~PDF_FLAGS_DIRTY;
 }
@@ -1691,10 +1683,18 @@ pdf_drop_dict(fz_context *ctx, pdf_obj *obj)
 	fz_free(ctx, obj);
 }
 
+pdf_obj *
+pdf_keep_obj(fz_context *ctx, pdf_obj *obj)
+{
+	if (obj >= PDF_LIMIT)
+		return fz_keep_imp16(ctx, obj, &obj->refs);
+	return obj;
+}
+
 void
 pdf_drop_obj(fz_context *ctx, pdf_obj *obj)
 {
-	if (obj >= PDF_OBJ__LIMIT)
+	if (obj >= PDF_LIMIT)
 	{
 		if (fz_drop_imp16(ctx, obj, &obj->refs))
 		{
@@ -1702,21 +1702,32 @@ pdf_drop_obj(fz_context *ctx, pdf_obj *obj)
 				pdf_drop_array(ctx, obj);
 			else if (obj->kind == PDF_DICT)
 				pdf_drop_dict(ctx, obj);
+			else if (obj->kind == PDF_STRING)
+			{
+				fz_free(ctx, STRING(obj)->text);
+				fz_free(ctx, obj);
+			}
 			else
 				fz_free(ctx, obj);
 		}
 	}
 }
 
+/*
+	Recurse through the object structure setting the node's parent_num to num.
+	parent_num is used when a subobject is to be changed during a document edit.
+	The whole containing hierarchy is moved to the incremental xref section, so
+	to be later written out as an incremental file update.
+*/
 void
 pdf_set_obj_parent(fz_context *ctx, pdf_obj *obj, int num)
 {
 	int n, i;
 
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return;
 
-	switch(obj->kind)
+	switch (obj->kind)
 	{
 	case PDF_ARRAY:
 		ARRAY(obj)->parent_num = num;
@@ -1735,10 +1746,10 @@ pdf_set_obj_parent(fz_context *ctx, pdf_obj *obj, int num)
 
 int pdf_obj_parent_num(fz_context *ctx, pdf_obj *obj)
 {
-	if (obj < PDF_OBJ__LIMIT)
+	if (obj < PDF_LIMIT)
 		return 0;
 
-	switch(obj->kind)
+	switch (obj->kind)
 	{
 	case PDF_INDIRECT:
 		return REF(obj)->num;
@@ -1751,39 +1762,17 @@ int pdf_obj_parent_num(fz_context *ctx, pdf_obj *obj)
 	}
 }
 
-pdf_obj *pdf_new_obj_from_str(fz_context *ctx, pdf_document *doc, const char *src)
-{
-	pdf_obj *result = NULL;
-	pdf_lexbuf lexbuf;
-	fz_stream *stream = fz_open_memory(ctx, (unsigned char *)src, strlen(src));
-
-	pdf_lexbuf_init(ctx, &lexbuf, PDF_LEXBUF_SMALL);
-	fz_try(ctx)
-	{
-		result = pdf_parse_stm_obj(ctx, doc, stream, &lexbuf);
-	}
-	fz_always(ctx)
-	{
-		pdf_lexbuf_fin(ctx, &lexbuf);
-		fz_drop_stream(ctx, stream);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
-	}
-
-	return result;
-}
-
 /* Pretty printing objects */
 
 struct fmt
 {
-	char *buf;
+	char *buf; /* original static buffer */
+	char *ptr; /* buffer we're writing to, maybe dynamically reallocated */
 	int cap;
 	int len;
 	int indent;
 	int tight;
+	int ascii;
 	int col;
 	int sep;
 	int last;
@@ -1824,8 +1813,21 @@ static inline void fmt_putc(fz_context *ctx, struct fmt *fmt, int c)
 	}
 	fmt->sep = 0;
 
-	if (fmt->buf && fmt->len < fmt->cap)
-		fmt->buf[fmt->len] = c;
+	if (fmt->len >= fmt->cap)
+	{
+		fmt->cap *= 2;
+		if (fmt->buf == fmt->ptr)
+		{
+			fmt->ptr = fz_malloc(ctx, fmt->cap);
+			memcpy(fmt->ptr, fmt->buf, fmt->len);
+		}
+		else
+		{
+			fmt->ptr = fz_realloc(ctx, fmt->ptr, fmt->cap);
+		}
+	}
+
+	fmt->ptr[fmt->len] = c;
 
 	if (c == '\n')
 		fmt->col = 0;
@@ -1857,6 +1859,39 @@ static inline void fmt_sep(fz_context *ctx, struct fmt *fmt)
 	fmt->sep = 1;
 }
 
+static int is_binary_string(fz_context *ctx, pdf_obj *obj)
+{
+	unsigned char *s = (unsigned char *)pdf_to_str_buf(ctx, obj);
+	int i, n = pdf_to_str_len(ctx, obj);
+	for (i = 0; i < n; ++i)
+	{
+		if (s[i] > 126) return 1;
+		if (s[i] < 32 && (s[i] != '\t' && s[i] != '\n' && s[i] != '\r')) return 1;
+	}
+	return 0;
+}
+
+static int is_longer_than_hex(fz_context *ctx, pdf_obj *obj)
+{
+	unsigned char *s = (unsigned char *)pdf_to_str_buf(ctx, obj);
+	int i, n = pdf_to_str_len(ctx, obj);
+	int m = 0;
+	for (i = 0; i < n; ++i)
+	{
+		if (s[i] > 126)
+			m += 4;
+		else if (s[i] == 0)
+			m += 4;
+		else if (strchr("\n\r\t\b\f()\\", s[i]))
+			m += 2;
+		else if (s[i] < 32)
+			m += 4;
+		else
+			m += 1;
+	}
+	return m > (n * 2);
+}
+
 static void fmt_str_out(fz_context *ctx, void *fmt_, const unsigned char *s, int n)
 {
 	struct fmt *fmt = (struct fmt *)fmt_;
@@ -2021,38 +2056,21 @@ static void fmt_dict(fz_context *ctx, struct fmt *fmt, pdf_obj *obj)
 	}
 }
 
-static void count_encrypted_data(fz_context *ctx, void *arg, const unsigned char *str, int len)
-{
-	int *encrypted_len = (int *)arg;
-	int added = 0;
-	int i;
-	unsigned char c;
-
-	for (i = 0; i < len; i++) {
-		c = (unsigned char)str[i];
-		if (c != 0 && strchr("()\\\n\r\t\b\f", c))
-			added ++;
-		else if (c < 32 || c >= 127)
-			added += 3;
-	}
-	*encrypted_len += added;
-}
-
 static void fmt_obj(fz_context *ctx, struct fmt *fmt, pdf_obj *obj)
 {
 	char buf[256];
 
-	if (!obj)
-		fmt_puts(ctx, fmt, "<NULL>");
+	if (obj == PDF_NULL)
+		fmt_puts(ctx, fmt, "null");
+	else if (obj == PDF_TRUE)
+		fmt_puts(ctx, fmt, "true");
+	else if (obj == PDF_FALSE)
+		fmt_puts(ctx, fmt, "false");
 	else if (pdf_is_indirect(ctx, obj))
 	{
 		fz_snprintf(buf, sizeof buf, "%d %d R", pdf_to_num(ctx, obj), pdf_to_gen(ctx, obj));
 		fmt_puts(ctx, fmt, buf);
 	}
-	else if (pdf_is_null(ctx, obj))
-		fmt_puts(ctx, fmt, "null");
-	else if (pdf_is_bool(ctx, obj))
-		fmt_puts(ctx, fmt, pdf_to_bool(ctx, obj) ? "true" : "false");
 	else if (pdf_is_int(ctx, obj))
 	{
 		fz_snprintf(buf, sizeof buf, "%d", pdf_to_int(ctx, obj));
@@ -2066,14 +2084,15 @@ static void fmt_obj(fz_context *ctx, struct fmt *fmt, pdf_obj *obj)
 	else if (pdf_is_string(ctx, obj))
 	{
 		unsigned char *str = (unsigned char *)pdf_to_str_buf(ctx, obj);
-		int len = pdf_to_str_len(ctx, obj);
-		int encoded_len = 0;
-
-		pdf_encrypt_data(ctx, fmt->crypt, fmt->num, fmt->gen, count_encrypted_data, &encoded_len, str, len);
-		if (encoded_len < 2*len)
-			fmt_str(ctx, fmt, obj);
-		else
+		if (fmt->crypt
+			|| (fmt->ascii && is_binary_string(ctx, obj))
+			|| (str[0]==0xff && str[1]==0xfe)
+			|| (str[0]==0xfe && str[1] == 0xff)
+			|| is_longer_than_hex(ctx, obj)
+			)
 			fmt_hex(ctx, fmt, obj);
+		else
+			fmt_str(ctx, fmt, obj);
 	}
 	else if (pdf_is_name(ctx, obj))
 		fmt_name(ctx, fmt, obj);
@@ -2085,8 +2104,8 @@ static void fmt_obj(fz_context *ctx, struct fmt *fmt, pdf_obj *obj)
 		fmt_puts(ctx, fmt, "<unknown object>");
 }
 
-int
-pdf_sprint_encrypted_obj(fz_context *ctx, char *s, int n, pdf_obj *obj, int tight, pdf_crypt *crypt, int num, int gen)
+static char *
+pdf_sprint_encrypted_obj(fz_context *ctx, char *buf, int cap, int *len, pdf_obj *obj, int tight, int ascii, pdf_crypt *crypt, int num, int gen)
 {
 	struct fmt fmt;
 
@@ -2095,124 +2114,177 @@ pdf_sprint_encrypted_obj(fz_context *ctx, char *s, int n, pdf_obj *obj, int tigh
 	fmt.sep = 0;
 	fmt.last = 0;
 
+	if (!buf || cap == 0)
+	{
+		fmt.cap = 1024;
+		fmt.buf = NULL;
+		fmt.ptr = fz_malloc(ctx, fmt.cap);
+	}
+	else
+	{
+		fmt.cap = cap;
+		fmt.buf = buf;
+		fmt.ptr = buf;
+	}
+
 	fmt.tight = tight;
-	fmt.buf = s;
-	fmt.cap = n;
+	fmt.ascii = ascii;
 	fmt.len = 0;
 	fmt.crypt = crypt;
 	fmt.num = num;
 	fmt.gen = gen;
 	fmt_obj(ctx, &fmt, obj);
 
-	if (fmt.buf && fmt.len < fmt.cap)
-		fmt.buf[fmt.len] = '\0';
+	fmt_putc(ctx, &fmt, 0);
 
-	return fmt.len;
+	return *len = fmt.len-1, fmt.ptr;
 }
 
-int
-pdf_sprint_obj(fz_context *ctx, char *s, int n, pdf_obj *obj, int tight)
+char *
+pdf_sprint_obj(fz_context *ctx, char *buf, int cap, int *len, pdf_obj *obj, int tight, int ascii)
 {
-	return pdf_sprint_encrypted_obj(ctx, s, n, obj, tight, NULL, 0, 0);
+	return pdf_sprint_encrypted_obj(ctx, buf, cap, len, obj, tight, ascii, NULL, 0, 0);
 }
 
-int pdf_print_encrypted_obj(fz_context *ctx, fz_output *out, pdf_obj *obj, int tight, pdf_crypt *crypt, int num, int gen)
+void pdf_print_encrypted_obj(fz_context *ctx, fz_output *out, pdf_obj *obj, int tight, int ascii, pdf_crypt *crypt, int num, int gen)
 {
 	char buf[1024];
 	char *ptr;
 	int n;
 
-	n = pdf_sprint_encrypted_obj(ctx, buf, sizeof buf, obj, tight, crypt, num, gen);
-	if (n <= sizeof buf)
-	{
-		fz_write_data(ctx, out, buf, n);
-	}
-	else
-	{
-		ptr = fz_malloc(ctx, n + 1);
-		pdf_sprint_encrypted_obj(ctx, ptr, n + 1, obj, tight, crypt, num, gen);
+	ptr = pdf_sprint_encrypted_obj(ctx, buf, sizeof buf, &n, obj, tight, ascii,crypt, num, gen);
+	fz_try(ctx)
 		fz_write_data(ctx, out, ptr, n);
-		fz_free(ctx, ptr);
-	}
-	return n;
+	fz_always(ctx)
+		if (ptr != buf)
+			fz_free(ctx, ptr);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
-int pdf_print_obj(fz_context *ctx, fz_output *out, pdf_obj *obj, int tight)
+void pdf_print_obj(fz_context *ctx, fz_output *out, pdf_obj *obj, int tight, int ascii)
 {
-	return pdf_print_encrypted_obj(ctx, out, obj, tight, NULL, 0, 0);
+	pdf_print_encrypted_obj(ctx, out, obj, tight, ascii, NULL, 0, 0);
 }
 
-static int pdf_debug_encrypted_obj(fz_context *ctx, pdf_obj *obj, int tight, pdf_crypt *crypt, int num, int gen)
+static void pdf_debug_encrypted_obj(fz_context *ctx, pdf_obj *obj, int tight, pdf_crypt *crypt, int num, int gen)
 {
 	char buf[1024];
 	char *ptr;
 	int n;
+	int ascii = 1;
 
-	n = pdf_sprint_obj(ctx, NULL, 0, obj, tight);
-	if ((n + 1) < sizeof buf)
-	{
-		pdf_sprint_encrypted_obj(ctx, buf, sizeof buf, obj, tight, crypt, num, gen);
-		fwrite(buf, 1, n, stdout);
-	}
-	else
-	{
-		ptr = fz_malloc(ctx, n + 1);
-		pdf_sprint_encrypted_obj(ctx, ptr, n + 1, obj, tight, crypt, num, gen);
-		fwrite(ptr, 1, n, stdout);
+	ptr = pdf_sprint_encrypted_obj(ctx, buf, sizeof buf, &n, obj, tight, ascii, crypt, num, gen);
+	fwrite(ptr, 1, n, stdout);
+	if (ptr != buf)
 		fz_free(ctx, ptr);
-	}
-	return n;
 }
 
 void pdf_debug_obj(fz_context *ctx, pdf_obj *obj)
+{
+	pdf_debug_encrypted_obj(ctx, pdf_resolve_indirect(ctx, obj), 0, NULL, 0, 0);
+	putchar('\n');
+}
+
+void pdf_debug_ref(fz_context *ctx, pdf_obj *obj)
 {
 	pdf_debug_encrypted_obj(ctx, obj, 0, NULL, 0, 0);
 	putchar('\n');
 }
 
-int pdf_obj_refs(fz_context *ctx, pdf_obj *ref)
+int pdf_obj_refs(fz_context *ctx, pdf_obj *obj)
 {
-	return (ref >= PDF_OBJ__LIMIT ? ref->refs : 0);
+	if (obj < PDF_LIMIT)
+		return 0;
+	return obj->refs;
 }
 
 /* Convenience functions */
 
+pdf_obj *
+pdf_dict_get_inheritable(fz_context *ctx, pdf_obj *node, pdf_obj *key)
+{
+	pdf_obj *node2 = node;
+	pdf_obj *val = NULL;
+	pdf_obj *marked = NULL;
+
+	fz_var(node);
+	fz_var(marked);
+	fz_try(ctx)
+	{
+		do
+		{
+			val = pdf_dict_get(ctx, node, key);
+			if (val)
+				break;
+			if (pdf_mark_obj(ctx, node))
+				fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in tree (parents)");
+			marked = node;
+			node = pdf_dict_get(ctx, node, PDF_NAME(Parent));
+		}
+		while (node);
+	}
+	fz_always(ctx)
+	{
+		/* We assume that if we have marked an object, without an exception
+		 * being thrown, that we can always unmark the same object again
+		 * without an exception being thrown. */
+		if (marked)
+		{
+			do
+			{
+				pdf_unmark_obj(ctx, node2);
+				if (node2 == marked)
+					break;
+				node2 = pdf_dict_get(ctx, node2, PDF_NAME(Parent));
+			}
+			while (node2);
+		}
+	}
+	fz_catch(ctx)
+	{
+		fz_rethrow(ctx);
+	}
+
+	return val;
+}
+
 void pdf_dict_put_bool(fz_context *ctx, pdf_obj *dict, pdf_obj *key, int x)
 {
-	pdf_dict_put_drop(ctx, dict, key, pdf_new_bool(ctx, NULL, x));
+	pdf_dict_put(ctx, dict, key, x ? PDF_TRUE : PDF_FALSE);
 }
 
 void pdf_dict_put_int(fz_context *ctx, pdf_obj *dict, pdf_obj *key, int64_t x)
 {
-	pdf_dict_put_drop(ctx, dict, key, pdf_new_int(ctx, NULL, x));
+	pdf_dict_put_drop(ctx, dict, key, pdf_new_int(ctx, x));
 }
 
 void pdf_dict_put_real(fz_context *ctx, pdf_obj *dict, pdf_obj *key, double x)
 {
-	pdf_dict_put_drop(ctx, dict, key, pdf_new_real(ctx, NULL, x));
+	pdf_dict_put_drop(ctx, dict, key, pdf_new_real(ctx, x));
 }
 
 void pdf_dict_put_name(fz_context *ctx, pdf_obj *dict, pdf_obj *key, const char *x)
 {
-	pdf_dict_put_drop(ctx, dict, key, pdf_new_name(ctx, NULL, x));
+	pdf_dict_put_drop(ctx, dict, key, pdf_new_name(ctx, x));
 }
 
 void pdf_dict_put_string(fz_context *ctx, pdf_obj *dict, pdf_obj *key, const char *x, size_t n)
 {
-	pdf_dict_put_drop(ctx, dict, key, pdf_new_string(ctx, NULL, x, n));
+	pdf_dict_put_drop(ctx, dict, key, pdf_new_string(ctx, x, n));
 }
 
 void pdf_dict_put_text_string(fz_context *ctx, pdf_obj *dict, pdf_obj *key, const char *x)
 {
-	pdf_dict_put_drop(ctx, dict, key, pdf_new_text_string(ctx, NULL, x));
+	pdf_dict_put_drop(ctx, dict, key, pdf_new_text_string(ctx, x));
 }
 
-void pdf_dict_put_rect(fz_context *ctx, pdf_obj *dict, pdf_obj *key, const fz_rect *x)
+void pdf_dict_put_rect(fz_context *ctx, pdf_obj *dict, pdf_obj *key, fz_rect x)
 {
 	pdf_dict_put_drop(ctx, dict, key, pdf_new_rect(ctx, NULL, x));
 }
 
-void pdf_dict_put_matrix(fz_context *ctx, pdf_obj *dict, pdf_obj *key, const fz_matrix *x)
+void pdf_dict_put_matrix(fz_context *ctx, pdf_obj *dict, pdf_obj *key, fz_matrix x)
 {
 	pdf_dict_put_drop(ctx, dict, key, pdf_new_matrix(ctx, NULL, x));
 }
@@ -2231,34 +2303,41 @@ pdf_obj *pdf_dict_put_dict(fz_context *ctx, pdf_obj *dict, pdf_obj *key, int ini
 	return obj;
 }
 
+pdf_obj *pdf_dict_puts_dict(fz_context *ctx, pdf_obj *dict, const char *key, int initial)
+{
+	pdf_obj *obj = pdf_new_dict(ctx, pdf_get_bound_document(ctx, dict), initial);
+	pdf_dict_puts_drop(ctx, dict, key, obj);
+	return obj;
+}
+
 void pdf_array_push_bool(fz_context *ctx, pdf_obj *array, int x)
 {
-	pdf_array_push_drop(ctx, array, pdf_new_bool(ctx, NULL, x));
+	pdf_array_push(ctx, array, x ? PDF_TRUE : PDF_FALSE);
 }
 
 void pdf_array_push_int(fz_context *ctx, pdf_obj *array, int64_t x)
 {
-	pdf_array_push_drop(ctx, array, pdf_new_int(ctx, NULL, x));
+	pdf_array_push_drop(ctx, array, pdf_new_int(ctx, x));
 }
 
 void pdf_array_push_real(fz_context *ctx, pdf_obj *array, double x)
 {
-	pdf_array_push_drop(ctx, array, pdf_new_real(ctx, NULL, x));
+	pdf_array_push_drop(ctx, array, pdf_new_real(ctx, x));
 }
 
 void pdf_array_push_name(fz_context *ctx, pdf_obj *array, const char *x)
 {
-	pdf_array_push_drop(ctx, array, pdf_new_name(ctx, NULL, x));
+	pdf_array_push_drop(ctx, array, pdf_new_name(ctx, x));
 }
 
 void pdf_array_push_string(fz_context *ctx, pdf_obj *array, const char *x, size_t n)
 {
-	pdf_array_push_drop(ctx, array, pdf_new_string(ctx, NULL, x, n));
+	pdf_array_push_drop(ctx, array, pdf_new_string(ctx, x, n));
 }
 
 void pdf_array_push_text_string(fz_context *ctx, pdf_obj *array, const char *x)
 {
-	pdf_array_push_drop(ctx, array, pdf_new_text_string(ctx, NULL, x));
+	pdf_array_push_drop(ctx, array, pdf_new_text_string(ctx, x));
 }
 
 pdf_obj *pdf_array_push_array(fz_context *ctx, pdf_obj *array, int initial)
@@ -2274,3 +2353,83 @@ pdf_obj *pdf_array_push_dict(fz_context *ctx, pdf_obj *array, int initial)
 	pdf_array_push_drop(ctx, array, obj);
 	return obj;
 }
+
+int pdf_dict_get_bool(fz_context *ctx, pdf_obj *dict, pdf_obj *key)
+{
+	return pdf_to_bool(ctx, pdf_dict_get(ctx, dict, key));
+}
+
+int pdf_dict_get_int(fz_context *ctx, pdf_obj *dict, pdf_obj *key)
+{
+	return pdf_to_int(ctx, pdf_dict_get(ctx, dict, key));
+}
+
+float pdf_dict_get_real(fz_context *ctx, pdf_obj *dict, pdf_obj *key)
+{
+	return pdf_to_real(ctx, pdf_dict_get(ctx, dict, key));
+}
+
+const char *pdf_dict_get_name(fz_context *ctx, pdf_obj *dict, pdf_obj *key)
+{
+	return pdf_to_name(ctx, pdf_dict_get(ctx, dict, key));
+}
+
+const char *pdf_dict_get_string(fz_context *ctx, pdf_obj *dict, pdf_obj *key, size_t *sizep)
+{
+	return pdf_to_string(ctx, pdf_dict_get(ctx, dict, key), sizep);
+}
+
+const char *pdf_dict_get_text_string(fz_context *ctx, pdf_obj *dict, pdf_obj *key)
+{
+	return pdf_to_text_string(ctx, pdf_dict_get(ctx, dict, key));
+}
+
+fz_rect pdf_dict_get_rect(fz_context *ctx, pdf_obj *dict, pdf_obj *key)
+{
+	return pdf_to_rect(ctx, pdf_dict_get(ctx, dict, key));
+}
+
+fz_matrix pdf_dict_get_matrix(fz_context *ctx, pdf_obj *dict, pdf_obj *key)
+{
+	return pdf_to_matrix(ctx, pdf_dict_get(ctx, dict, key));
+}
+
+int pdf_array_get_bool(fz_context *ctx, pdf_obj *array, int index)
+{
+	return pdf_to_bool(ctx, pdf_array_get(ctx, array, index));
+}
+
+int pdf_array_get_int(fz_context *ctx, pdf_obj *array, int index)
+{
+	return pdf_to_int(ctx, pdf_array_get(ctx, array, index));
+}
+
+float pdf_array_get_real(fz_context *ctx, pdf_obj *array, int index)
+{
+	return pdf_to_real(ctx, pdf_array_get(ctx, array, index));
+}
+
+const char *pdf_array_get_name(fz_context *ctx, pdf_obj *array, int index)
+{
+	return pdf_to_name(ctx, pdf_array_get(ctx, array, index));
+}
+
+const char *pdf_array_get_string(fz_context *ctx, pdf_obj *array, int index, size_t *sizep)
+{
+	return pdf_to_string(ctx, pdf_array_get(ctx, array, index), sizep);
+}
+
+const char *pdf_array_get_text_string(fz_context *ctx, pdf_obj *array, int index)
+{
+	return pdf_to_text_string(ctx, pdf_array_get(ctx, array, index));
+}
+
+fz_rect pdf_array_get_rect(fz_context *ctx, pdf_obj *array, int index)
+{
+	return pdf_to_rect(ctx, pdf_array_get(ctx, array, index));
+}
+
+fz_matrix pdf_array_get_matrix(fz_context *ctx, pdf_obj *array, int index)
+{
+	return pdf_to_matrix(ctx, pdf_array_get(ctx, array, index));
+}
