@@ -21,7 +21,7 @@ static void add_root(fz_context *ctx, pdf_obj *obj, pdf_obj ***roots, int *num_r
 		int new_max_roots = *max_roots * 2;
 		if (new_max_roots == 0)
 			new_max_roots = 4;
-		*roots = fz_resize_array(ctx, *roots, new_max_roots, sizeof(**roots));
+		*roots = fz_realloc_array(ctx, *roots, new_max_roots, pdf_obj*);
 		*max_roots = new_max_roots;
 	}
 	(*roots)[(*num_roots)++] = pdf_keep_obj(ctx, obj);
@@ -56,6 +56,7 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 		}
 		fz_catch(ctx)
 		{
+			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
 			/* Don't let a broken object at EOF overwrite a good one */
 			if (file->eof)
 				fz_rethrow(ctx);
@@ -103,6 +104,16 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 		if (!pdf_is_indirect(ctx, obj) && pdf_is_int(ctx, obj))
 			stm_len = pdf_to_int(ctx, obj);
 
+		if (doc->file_reading_linearly && page)
+		{
+			obj = pdf_dict_get(ctx, dict, PDF_NAME(Type));
+			if (!pdf_is_indirect(ctx, obj) && pdf_name_eq(ctx, obj, PDF_NAME(Page)))
+			{
+				pdf_drop_obj(ctx, *page);
+				*page = pdf_keep_obj(ctx, dict);
+			}
+		}
+
 		pdf_drop_obj(ctx, dict);
 	}
 
@@ -140,6 +151,7 @@ pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int64_t *stm
 			}
 			fz_catch(ctx)
 			{
+				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
 				fz_warn(ctx, "cannot find endstream token, falling back to scanning");
 			}
 			if (tok == PDF_TOK_ENDSTREAM)
@@ -257,7 +269,7 @@ orphan_object(fz_context *ctx, pdf_document *doc, pdf_obj *obj)
 
 		fz_try(ctx)
 		{
-			doc->orphans = fz_resize_array(ctx, doc->orphans, new_max, sizeof(*doc->orphans));
+			doc->orphans = fz_realloc_array(ctx, doc->orphans, new_max, pdf_obj*);
 			doc->orphans_max = new_max;
 		}
 		fz_catch(ctx)
@@ -330,7 +342,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 		pdf_xref_entry *entry;
 		listlen = 0;
 		listcap = 1024;
-		list = fz_malloc_array(ctx, listcap, sizeof(struct entry));
+		list = fz_malloc_array(ctx, listcap, struct entry);
 
 		/* look for '%PDF' version marker within first kilobyte of file */
 		n = fz_read(ctx, doc->file, (unsigned char *)buf->scratch, fz_mini(buf->size, 1024));
@@ -365,6 +377,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				tok = pdf_lex_no_string(ctx, doc->file, buf);
 			fz_catch(ctx)
 			{
+				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
 				fz_warn(ctx, "skipping ahead to next token");
 				do
 					c = fz_read_byte(ctx, doc->file);
@@ -412,6 +425,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				}
 				fz_catch(ctx)
 				{
+					fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
 					/* If we haven't seen a root yet, there is nothing
 					 * we can do, but give up. Otherwise, we'll make
 					 * do. */
@@ -432,7 +446,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				if (listlen + 1 == listcap)
 				{
 					listcap = (listcap * 3) / 2;
-					list = fz_resize_array(ctx, list, listcap, sizeof(struct entry));
+					list = fz_realloc_array(ctx, list, listcap, struct entry);
 				}
 
 				list[listlen].num = num;
@@ -461,6 +475,7 @@ pdf_repair_xref(fz_context *ctx, pdf_document *doc)
 				}
 				fz_catch(ctx)
 				{
+					fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
 					/* If this was the real trailer dict
 					 * it was broken, in which case we are
 					 * in trouble. Keep going though in
