@@ -17,7 +17,10 @@ pdf_obj_num_is_stream(fz_context *ctx, pdf_document *doc, int num)
 	fz_try(ctx)
 		entry = pdf_cache_object(ctx, doc, num);
 	fz_catch(ctx)
+	{
+		fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
 		return 0;
+	}
 
 	return entry->stm_ofs != 0 || entry->stm_buf;
 }
@@ -671,7 +674,14 @@ pdf_open_object_array(fz_context *ctx, pdf_document *doc, pdf_obj *list)
 		fz_try(ctx)
 			fz_concat_push_drop(ctx, stm, pdf_open_stream(ctx, obj));
 		fz_catch(ctx)
+		{
+			if (fz_caught(ctx) == FZ_ERROR_TRYLATER)
+			{
+				fz_drop_stream(ctx, stm);
+				fz_rethrow(ctx);
+			}
 			fz_warn(ctx, "cannot load content stream part %d/%d", i + 1, n);
+		}
 	}
 
 	return stm;
