@@ -1,6 +1,5 @@
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
-#include "pdf-imp.h"
 
 #include <assert.h>
 #include <limits.h>
@@ -14,6 +13,8 @@
 #define DEBUGMESS(A) do { } while (0)
 #endif
 
+#define isdigit(c) (c >= '0' && c <= '9')
+
 static inline int iswhite(int ch)
 {
 	return
@@ -84,7 +85,7 @@ extend_xref_index(fz_context *ctx, pdf_document *doc, int newlen)
 {
 	int i;
 
-	doc->xref_index = fz_resize_array(ctx, doc->xref_index, newlen, sizeof(int));
+	doc->xref_index = fz_realloc_array(ctx, doc->xref_index, newlen, int);
 	for (i = doc->max_xref_len; i < newlen; i++)
 	{
 		doc->xref_index[i] = 0;
@@ -106,7 +107,7 @@ static void pdf_resize_xref(fz_context *ctx, pdf_document *doc, int newlen)
 	assert(sub->next == NULL && sub->start == 0 && sub->len == xref->num_objects);
 	assert(newlen > xref->num_objects);
 
-	sub->table = fz_resize_array(ctx, sub->table, newlen, sizeof(pdf_xref_entry));
+	sub->table = fz_realloc_array(ctx, sub->table, newlen, pdf_xref_entry);
 	for (i = xref->num_objects; i < newlen; i++)
 	{
 		sub->table[i].type = 0;
@@ -126,7 +127,7 @@ static void pdf_resize_xref(fz_context *ctx, pdf_document *doc, int newlen)
 static void pdf_populate_next_xref_level(fz_context *ctx, pdf_document *doc)
 {
 	pdf_xref *xref;
-	doc->xref_sections = fz_resize_array(ctx, doc->xref_sections, doc->num_xref_sections + 1, sizeof(pdf_xref));
+	doc->xref_sections = fz_realloc_array(ctx, doc->xref_sections, doc->num_xref_sections + 1, pdf_xref);
 	doc->num_xref_sections++;
 
 	xref = &doc->xref_sections[doc->num_xref_sections - 1];
@@ -337,34 +338,37 @@ static void ensure_incremental_xref(fz_context *ctx, pdf_document *doc)
 		pdf_xref *xref = &doc->xref_sections[0];
 		pdf_xref *pxref;
 		pdf_xref_entry *new_table = fz_calloc(ctx, xref->num_objects, sizeof(pdf_xref_entry));
-		pdf_xref_subsec *sub;
+		pdf_xref_subsec *sub = NULL;
 		pdf_obj *trailer = NULL;
 		int i;
 
 		fz_var(trailer);
+		fz_var(sub);
 		fz_try(ctx)
 		{
 			sub = fz_malloc_struct(ctx, pdf_xref_subsec);
 			trailer = xref->trailer ? pdf_copy_dict(ctx, xref->trailer) : NULL;
-			doc->xref_sections = fz_resize_array(ctx, doc->xref_sections, doc->num_xref_sections + 1, sizeof(pdf_xref));
+			doc->xref_sections = fz_realloc_array(ctx, doc->xref_sections, doc->num_xref_sections + 1, pdf_xref);
 			xref = &doc->xref_sections[0];
 			pxref = &doc->xref_sections[1];
 			memmove(pxref, xref, doc->num_xref_sections * sizeof(pdf_xref));
 			/* xref->num_objects is already correct */
 			xref->subsec = sub;
+			sub = NULL;
 			xref->trailer = trailer;
 			xref->pre_repair_trailer = NULL;
 			xref->unsaved_sigs = NULL;
 			xref->unsaved_sigs_end = NULL;
-			sub->next = NULL;
-			sub->len = xref->num_objects;
-			sub->start = 0;
-			sub->table = new_table;
+			xref->subsec->next = NULL;
+			xref->subsec->len = xref->num_objects;
+			xref->subsec->start = 0;
+			xref->subsec->table = new_table;
 			doc->num_xref_sections++;
 			doc->num_incremental_sections++;
 		}
 		fz_catch(ctx)
 		{
+			fz_free(ctx, sub);
 			fz_free(ctx, new_table);
 			pdf_drop_obj(ctx, trailer);
 			fz_rethrow(ctx);
@@ -508,45 +512,46 @@ void pdf_xref_ensure_incremental_object(fz_context *ctx, pdf_document *doc, int
 
 void pdf_replace_xref(fz_context *ctx, pdf_document *doc, pdf_xref_entry *entries, int n)
 {
+	int *xref_index = NULL;
 	pdf_xref *xref = NULL;
 	pdf_xref_subsec *sub;
-	pdf_obj *trailer = pdf_keep_obj(ctx, pdf_trailer(ctx, doc));
 
+	fz_var(xref_index);
 	fz_var(xref);
+
 	fz_try(ctx)
 	{
-		fz_free(ctx, doc->xref_index);
-		doc->xref_index = NULL; /* In case the calloc fails */
-		doc->xref_index = fz_calloc(ctx, n, sizeof(int));
+		xref_index = fz_calloc(ctx, n, sizeof(int));
 		xref = fz_malloc_struct(ctx, pdf_xref);
 		sub = fz_malloc_struct(ctx, pdf_xref_subsec);
-
-		/* The new table completely replaces the previous separate sections */
-		pdf_drop_xref_sections(ctx, doc);
-
-		sub->table = entries;
-		sub->start = 0;
-		sub->len = n;
-		xref->subsec = sub;
-		xref->num_objects = n;
-		xref->trailer = trailer;
-		trailer = NULL;
-
-		doc->xref_sections = xref;
-		doc->num_xref_sections = 1;
-		doc->num_incremental_sections = 0;
-		doc->xref_base = 0;
-		doc->disallow_new_increments = 0;
-		doc->max_xref_len = n;
-
-		memset(doc->xref_index, 0, sizeof(int)*doc->max_xref_len);
 	}
 	fz_catch(ctx)
 	{
 		fz_free(ctx, xref);
-		pdf_drop_obj(ctx, trailer);
+		fz_free(ctx, xref_index);
 		fz_rethrow(ctx);
 	}
+
+	sub->table = entries;
+	sub->start = 0;
+	sub->len = n;
+
+	xref->subsec = sub;
+	xref->num_objects = n;
+	xref->trailer = pdf_keep_obj(ctx, pdf_trailer(ctx, doc));
+
+	/* The new table completely replaces the previous separate sections */
+	pdf_drop_xref_sections(ctx, doc);
+
+	doc->xref_sections = xref;
+	doc->num_xref_sections = 1;
+	doc->num_incremental_sections = 0;
+	doc->xref_base = 0;
+	doc->disallow_new_increments = 0;
+	doc->max_xref_len = n;
+
+	fz_free(ctx, doc->xref_index);
+	doc->xref_index = xref_index;
 }
 
 void pdf_forget_xref(fz_context *ctx, pdf_document *doc)
@@ -626,7 +631,7 @@ pdf_read_start_xref(fz_context *ctx, pdf_document *doc)
 			while (i < n && iswhite(buf[i]))
 				i ++;
 			doc->startxref = 0;
-			while (i < n && buf[i] >= '0' && buf[i] <= '9')
+			while (i < n && isdigit(buf[i]))
 			{
 				if (doc->startxref >= INT64_MAX/10)
 					fz_throw(ctx, FZ_ERROR_GENERIC, "startxref too large");
@@ -647,7 +652,7 @@ fz_skip_space(fz_context *ctx, fz_stream *stm)
 	do
 	{
 		int c = fz_peek_byte(ctx, stm);
-		if (c > 32 && c != EOF)
+		if (c == EOF || c > 32)
 			return;
 		(void)fz_read_byte(ctx, stm);
 	}
@@ -696,7 +701,7 @@ pdf_xref_size_from_old_trailer(fz_context *ctx, pdf_document *doc, pdf_lexbuf *b
 	while (1)
 	{
 		c = fz_peek_byte(ctx, doc->file);
-		if (!(c >= '0' && c <= '9'))
+		if (!isdigit(c))
 			break;
 
 		fz_read_line(ctx, doc->file, buf->scratch, buf->size);
@@ -747,7 +752,7 @@ pdf_xref_size_from_old_trailer(fz_context *ctx, pdf_document *doc, pdf_lexbuf *b
 
 		trailer = pdf_parse_dict(ctx, doc, doc->file, buf);
 
-		size = pdf_to_int(ctx, pdf_dict_get(ctx, trailer, PDF_NAME_Size));
+		size = pdf_dict_get_int(ctx, trailer, PDF_NAME(Size));
 		if (size < 0 || size > PDF_MAX_OBJECT_NUMBER + 1)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "trailer Size entry out of range");
 	}
@@ -831,7 +836,7 @@ pdf_read_old_xref(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf)
 	pdf_xref_entry *table;
 	pdf_token tok;
 	size_t n;
-	char *s;
+	char *s, *e;
 
 	xref_len = pdf_xref_size_from_old_trailer(ctx, doc, buf);
 
@@ -843,7 +848,7 @@ pdf_read_old_xref(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf)
 	while (1)
 	{
 		c = fz_peek_byte(ctx, file);
-		if (!(c >= '0' && c <= '9'))
+		if (!isdigit(c))
 			break;
 
 		fz_read_line(ctx, file, buf->scratch, buf->size);
@@ -883,24 +888,40 @@ pdf_read_old_xref(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf)
 			if (n != 20-carried)
 				fz_throw(ctx, FZ_ERROR_GENERIC, "unexpected EOF in xref table");
 			n += carried;
+			buf->scratch[n] = '\0';
 			if (!entry->type)
 			{
 				s = buf->scratch;
+				e = s + n;
+
+				entry->num = start + i;
 
 				/* broken pdfs where line start with white space */
-				while (*s != '\0' && iswhite(*s))
+				while (s < e && iswhite(*s))
 					s++;
 
-				entry->ofs = fz_atoi64(s);
-				entry->gen = fz_atoi(s + 11);
-				entry->num = start + i;
-				entry->type = s[17];
-				if (s[17] != 'f' && s[17] != 'n' && s[17] != 'o')
-					fz_throw(ctx, FZ_ERROR_GENERIC, "unexpected xref type: 0x%x (%d %d R)", s[17], entry->num, entry->gen);
+				if (s == e || !isdigit(*s))
+					fz_throw(ctx, FZ_ERROR_GENERIC, "xref offset missing");
+				while (s < e && isdigit(*s))
+					entry->ofs = entry->ofs * 10 + *s++ - '0';
+
+				while (s < e && iswhite(*s))
+					s++;
+				if (s == e || !isdigit(*s))
+					fz_throw(ctx, FZ_ERROR_GENERIC, "xref generation number missing");
+				while (s < e && isdigit(*s))
+					entry->gen = entry->gen * 10 + *s++ - '0';
+
+				while (s < e && iswhite(*s))
+					s++;
+				if (s == e || (*s != 'f' && *s != 'n' && *s != 'o'))
+					fz_throw(ctx, FZ_ERROR_GENERIC, "unexpected xref type: 0x%x (%d %d R)", s == e ? 0 : *s, entry->num, entry->gen);
+				entry->type = *s++;
+
 				/* If the last byte of our buffer isn't an EOL (or space), carry one byte forward */
-				carried = s[19] > 32;
+				carried = buf->scratch[19] > 32;
 				if (carried)
-					s[0] = s[19];
+					buf->scratch[0] = buf->scratch[19];
 			}
 		}
 		if (carried)
@@ -915,6 +936,8 @@ pdf_read_old_xref(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf)
 	if (tok != PDF_TOK_OPEN_DICT)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "expected trailer dictionary");
 
+	doc->has_old_style_xrefs = 1;
+
 	return pdf_parse_dict(ctx, doc, file, buf);
 }
 
@@ -989,18 +1012,18 @@ pdf_read_new_xref(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf)
 	{
 		pdf_xref_entry *entry;
 
-		obj = pdf_dict_get(ctx, trailer, PDF_NAME_Size);
+		obj = pdf_dict_get(ctx, trailer, PDF_NAME(Size));
 		if (!obj)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "xref stream missing Size entry (%d 0 R)", num);
 
 		size = pdf_to_int(ctx, obj);
 
-		obj = pdf_dict_get(ctx, trailer, PDF_NAME_W);
+		obj = pdf_dict_get(ctx, trailer, PDF_NAME(W));
 		if (!obj)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "xref stream missing W entry (%d  R)", num);
-		w0 = pdf_to_int(ctx, pdf_array_get(ctx, obj, 0));
-		w1 = pdf_to_int(ctx, pdf_array_get(ctx, obj, 1));
-		w2 = pdf_to_int(ctx, pdf_array_get(ctx, obj, 2));
+		w0 = pdf_array_get_int(ctx, obj, 0);
+		w1 = pdf_array_get_int(ctx, obj, 1);
+		w2 = pdf_array_get_int(ctx, obj, 2);
 
 		if (w0 < 0)
 			fz_warn(ctx, "xref stream objects have corrupt type");
@@ -1013,7 +1036,7 @@ pdf_read_new_xref(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf)
 		w1 = w1 < 0 ? 0 : w1;
 		w2 = w2 < 0 ? 0 : w2;
 
-		index = pdf_dict_get(ctx, trailer, PDF_NAME_Index);
+		index = pdf_dict_get(ctx, trailer, PDF_NAME(Index));
 
 		stm = pdf_open_stream_with_offset(ctx, doc, num, trailer, stm_ofs);
 
@@ -1026,8 +1049,8 @@ pdf_read_new_xref(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf)
 			int n = pdf_array_len(ctx, index);
 			for (t = 0; t < n; t += 2)
 			{
-				int i0 = pdf_to_int(ctx, pdf_array_get(ctx, index, t + 0));
-				int i1 = pdf_to_int(ctx, pdf_array_get(ctx, index, t + 1));
+				int i0 = pdf_array_get_int(ctx, index, t + 0);
+				int i1 = pdf_array_get_int(ctx, index, t + 1);
 				pdf_read_new_xref_section(ctx, doc, stm, i0, i1, w0, w1, w2);
 			}
 		}
@@ -1067,7 +1090,7 @@ pdf_read_xref(fz_context *ctx, pdf_document *doc, int64_t ofs, pdf_lexbuf *buf)
 	c = fz_peek_byte(ctx, doc->file);
 	if (c == 'x')
 		trailer = pdf_read_old_xref(ctx, doc, buf);
-	else if (c >= '0' && c <= '9')
+	else if (isdigit(c))
 		trailer = pdf_read_new_xref(ctx, doc, buf);
 	else
 		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot recognize xref format");
@@ -1079,20 +1102,18 @@ static int64_t
 read_xref_section(fz_context *ctx, pdf_document *doc, int64_t ofs, pdf_lexbuf *buf)
 {
 	pdf_obj *trailer = NULL;
+	pdf_obj *prevobj;
 	int64_t xrefstmofs = 0;
 	int64_t prevofs = 0;
 
-	fz_var(trailer);
-
+	trailer = pdf_read_xref(ctx, doc, ofs, buf);
 	fz_try(ctx)
 	{
-		trailer = pdf_read_xref(ctx, doc, ofs, buf);
-
 		pdf_set_populating_xref_trailer(ctx, doc, trailer);
 
 		/* FIXME: do we overwrite free entries properly? */
 		/* FIXME: Does this work properly with progression? */
-		xrefstmofs = pdf_to_int64(ctx, pdf_dict_get(ctx, trailer, PDF_NAME_XRefStm));
+		xrefstmofs = pdf_to_int64(ctx, pdf_dict_get(ctx, trailer, PDF_NAME(XRefStm)));
 		if (xrefstmofs)
 		{
 			if (xrefstmofs < 0)
@@ -1106,18 +1127,18 @@ read_xref_section(fz_context *ctx, pdf_document *doc, int64_t ofs, pdf_lexbuf *b
 			pdf_drop_obj(ctx, pdf_read_xref(ctx, doc, xrefstmofs, buf));
 		}
 
-		prevofs = pdf_to_int64(ctx, pdf_dict_get(ctx, trailer, PDF_NAME_Prev));
-		if (prevofs < 0)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "negative xref stream offset for previous xref stream");
+		prevobj = pdf_dict_get(ctx, trailer, PDF_NAME(Prev));
+		if (pdf_is_int(ctx, prevobj))
+		{
+			prevofs = pdf_to_int64(ctx, prevobj);
+			if (prevofs <= 0)
+				fz_throw(ctx, FZ_ERROR_GENERIC, "invalid offset for previous xref section");
+		}
 	}
 	fz_always(ctx)
-	{
 		pdf_drop_obj(ctx, trailer);
-	}
 	fz_catch(ctx)
-	{
 		fz_rethrow(ctx);
-	}
 
 	return prevofs;
 }
@@ -1130,7 +1151,7 @@ pdf_read_xref_sections(fz_context *ctx, pdf_document *doc, int64_t ofs, pdf_lexb
 
 	len = 0;
 	cap = 10;
-	offsets = fz_malloc_array(ctx, cap, sizeof(*offsets));
+	offsets = fz_malloc_array(ctx, cap, int64_t);
 
 	fz_try(ctx)
 	{
@@ -1143,13 +1164,13 @@ pdf_read_xref_sections(fz_context *ctx, pdf_document *doc, int64_t ofs, pdf_lexb
 			}
 			if (i < len)
 			{
-				fz_warn(ctx, "ignoring xref section recursion at offset %lu", ofs);
+				fz_warn(ctx, "ignoring xref section recursion at offset %d", (int)ofs);
 				break;
 			}
 			if (len == cap)
 			{
 				cap *= 2;
-				offsets = fz_resize_array(ctx, offsets, cap, sizeof(*offsets));
+				offsets = fz_realloc_array(ctx, offsets, cap, int64_t);
 			}
 			offsets[len++] = ofs;
 
@@ -1274,29 +1295,29 @@ pdf_load_linear(fz_context *ctx, pdf_document *doc)
 		dict = pdf_parse_ind_obj(ctx, doc, doc->file, &doc->lexbuf.base, &num, &gen, &stmofs, NULL);
 		if (!pdf_is_dict(ctx, dict))
 			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to read linearized dictionary");
-		o = pdf_dict_get(ctx, dict, PDF_NAME_Linearized);
+		o = pdf_dict_get(ctx, dict, PDF_NAME(Linearized));
 		if (o == NULL)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to read linearized dictionary");
 		lin = pdf_to_int(ctx, o);
 		if (lin != 1)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "Unexpected version of Linearized tag (%d)", lin);
-		len = pdf_to_int(ctx, pdf_dict_get(ctx, dict, PDF_NAME_L));
+		len = pdf_dict_get_int(ctx, dict, PDF_NAME(L));
 		if (len != doc->file_length)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "File has been updated since linearization");
 
 		pdf_read_xref_sections(ctx, doc, fz_tell(ctx, doc->file), &doc->lexbuf.base, 0);
 
-		doc->linear_page_count = pdf_to_int(ctx, pdf_dict_get(ctx, dict, PDF_NAME_N));
-		doc->linear_page_refs = fz_resize_array(ctx, doc->linear_page_refs, doc->linear_page_count, sizeof(pdf_obj *));
+		doc->linear_page_count = pdf_dict_get_int(ctx, dict, PDF_NAME(N));
+		doc->linear_page_refs = fz_realloc_array(ctx, doc->linear_page_refs, doc->linear_page_count, pdf_obj *);
 		memset(doc->linear_page_refs, 0, doc->linear_page_count * sizeof(pdf_obj*));
 		doc->linear_obj = dict;
 		doc->linear_pos = fz_tell(ctx, doc->file);
-		doc->linear_page1_obj_num = pdf_to_int(ctx, pdf_dict_get(ctx, dict, PDF_NAME_O));
+		doc->linear_page1_obj_num = pdf_dict_get_int(ctx, dict, PDF_NAME(O));
 		doc->linear_page_refs[0] = pdf_new_indirect(ctx, doc, doc->linear_page1_obj_num, 0);
 		doc->linear_page_num = 0;
-		hint = pdf_dict_get(ctx, dict, PDF_NAME_H);
-		doc->hint_object_offset = pdf_to_int(ctx, pdf_array_get(ctx, hint, 0));
-		doc->hint_object_length = pdf_to_int(ctx, pdf_array_get(ctx, hint, 1));
+		hint = pdf_dict_get(ctx, dict, PDF_NAME(H));
+		doc->hint_object_offset = pdf_array_get_int(ctx, hint, 0);
+		doc->hint_object_length = pdf_array_get_int(ctx, hint, 1);
 
 		entry = pdf_get_populating_xref_entry(ctx, doc, 0);
 		entry->type = 'f';
@@ -1329,15 +1350,18 @@ pdf_init_document(fz_context *ctx, pdf_document *doc)
 
 	fz_try(ctx)
 	{
-		pdf_load_version(ctx, doc);
-
-		doc->file_length = fz_stream_meta(ctx, doc->file, FZ_STREAM_META_LENGTH, 0, NULL);
-		if (doc->file_length < 0)
-			doc->file_length = 0;
-
 		/* Check to see if we should work in progressive mode */
-		if (fz_stream_meta(ctx, doc->file, FZ_STREAM_META_PROGRESSIVE, 0, NULL) > 0)
+		if (doc->file->progressive)
+		{
 			doc->file_reading_linearly = 1;
+			fz_seek(ctx, doc->file, 0, SEEK_END);
+			doc->file_length = fz_tell(ctx, doc->file);
+			if (doc->file_length < 0)
+				doc->file_length = 0;
+			fz_seek(ctx, doc->file, 0, SEEK_SET);
+		}
+
+		pdf_load_version(ctx, doc);
 
 		/* Try to load the linearized file if we are in progressive
 		 * mode. */
@@ -1365,13 +1389,14 @@ pdf_init_document(fz_context *ctx, pdf_document *doc)
 		if (repaired)
 		{
 			/* pdf_repair_xref may access xref_index, so reset it properly */
-			memset(doc->xref_index, 0, sizeof(int) * doc->max_xref_len);
+			if (doc->xref_index)
+				memset(doc->xref_index, 0, sizeof(int) * doc->max_xref_len);
 			pdf_repair_xref(ctx, doc);
 			pdf_prime_xref_index(ctx, doc);
 		}
 
-		encrypt = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Encrypt);
-		id = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_ID);
+		encrypt = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(Encrypt));
+		id = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(ID));
 		if (pdf_is_dict(ctx, encrypt))
 			doc->crypt = pdf_new_crypt(ctx, encrypt, id);
 
@@ -1383,10 +1408,10 @@ pdf_init_document(fz_context *ctx, pdf_document *doc)
 			int xref_len = pdf_xref_len(ctx, doc);
 			pdf_repair_obj_stms(ctx, doc);
 
-			hasroot = (pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Root) != NULL);
-			hasinfo = (pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Info) != NULL);
+			hasroot = (pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(Root)) != NULL);
+			hasinfo = (pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(Info)) != NULL);
 
-			for (i = 1; i < xref_len; i++)
+			for (i = 1; i < xref_len && !hasinfo && !hasroot; ++i)
 			{
 				pdf_xref_entry *entry = pdf_get_xref_entry(ctx, doc, i);
 				if (entry->type == 0 || entry->type == 'f')
@@ -1405,20 +1430,22 @@ pdf_init_document(fz_context *ctx, pdf_document *doc)
 
 				if (!hasroot)
 				{
-					obj = pdf_dict_get(ctx, dict, PDF_NAME_Type);
-					if (pdf_name_eq(ctx, obj, PDF_NAME_Catalog))
+					obj = pdf_dict_get(ctx, dict, PDF_NAME(Type));
+					if (pdf_name_eq(ctx, obj, PDF_NAME(Catalog)))
 					{
 						nobj = pdf_new_indirect(ctx, doc, i, 0);
-						pdf_dict_put_drop(ctx, pdf_trailer(ctx, doc), PDF_NAME_Root, nobj);
+						pdf_dict_put_drop(ctx, pdf_trailer(ctx, doc), PDF_NAME(Root), nobj);
+						hasroot = 1;
 					}
 				}
 
 				if (!hasinfo)
 				{
-					if (pdf_dict_get(ctx, dict, PDF_NAME_Creator) || pdf_dict_get(ctx, dict, PDF_NAME_Producer))
+					if (pdf_dict_get(ctx, dict, PDF_NAME(Creator)) || pdf_dict_get(ctx, dict, PDF_NAME(Producer)))
 					{
 						nobj = pdf_new_indirect(ctx, doc, i, 0);
-						pdf_dict_put_drop(ctx, pdf_trailer(ctx, doc), PDF_NAME_Info, nobj);
+						pdf_dict_put_drop(ctx, pdf_trailer(ctx, doc), PDF_NAME(Info), nobj);
+						hasinfo = 1;
 					}
 				}
 
@@ -1443,13 +1470,14 @@ pdf_init_document(fz_context *ctx, pdf_document *doc)
 	}
 	fz_catch(ctx)
 	{
-		fz_warn(ctx, "Ignoring Broken Optional Content");
+		fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+		fz_warn(ctx, "Ignoring broken Optional Content configuration");
 	}
 
 	fz_try(ctx)
 	{
 		const char *version_str;
-		obj = pdf_dict_getl(ctx, pdf_trailer(ctx, doc), PDF_NAME_Root, PDF_NAME_Version, NULL);
+		obj = pdf_dict_getl(ctx, pdf_trailer(ctx, doc), PDF_NAME(Root), PDF_NAME(Version), NULL);
 		version_str = pdf_to_name(ctx, obj);
 		if (*version_str)
 		{
@@ -1483,7 +1511,6 @@ pdf_drop_document_imp(fz_context *ctx, pdf_document *doc)
 	pdf_drop_xref_sections(ctx, doc);
 	fz_free(ctx, doc->xref_index);
 
-	pdf_drop_obj(ctx, doc->focus_obj);
 	fz_drop_stream(ctx, doc->file);
 	pdf_drop_crypt(ctx, doc->crypt);
 
@@ -1516,7 +1543,6 @@ pdf_drop_document_imp(fz_context *ctx, pdf_document *doc)
 	fz_free(ctx, doc->type3_fonts);
 
 	pdf_drop_ocg(ctx, doc);
-	pdf_drop_portfolio(ctx, doc);
 
 	pdf_empty_store(ctx, doc);
 
@@ -1536,6 +1562,12 @@ pdf_drop_document_imp(fz_context *ctx, pdf_document *doc)
 	fz_defer_reap_end(ctx);
 }
 
+/*
+	Closes and frees an opened PDF document.
+
+	The resource store in the context associated with pdf_document
+	is emptied.
+*/
 void
 pdf_drop_document(fz_context *ctx, pdf_document *doc)
 {
@@ -1591,8 +1623,8 @@ pdf_load_obj_stm(fz_context *ctx, pdf_document *doc, int num, pdf_lexbuf *buf, i
 	{
 		pdf_mark_obj(ctx, objstm);
 
-		count = pdf_to_int(ctx, pdf_dict_get(ctx, objstm, PDF_NAME_N));
-		first = pdf_to_int(ctx, pdf_dict_get(ctx, objstm, PDF_NAME_First));
+		count = pdf_dict_get_int(ctx, objstm, PDF_NAME(N));
+		first = pdf_dict_get_int(ctx, objstm, PDF_NAME(First));
 
 		if (count < 0 || count > PDF_MAX_OBJECT_NUMBER)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "number of objects in object stream out of range");
@@ -1818,24 +1850,29 @@ pdf_obj_read(fz_context *ctx, pdf_document *doc, int64_t *offset, int *nump, pdf
 static void
 pdf_load_hinted_page(fz_context *ctx, pdf_document *doc, int pagenum)
 {
+	pdf_obj *page = NULL;
+
 	if (!doc->hints_loaded || !doc->linear_page_refs)
 		return;
 
 	if (doc->linear_page_refs[pagenum])
 		return;
 
+	fz_var(page);
+
 	fz_try(ctx)
 	{
 		int num = doc->hint_page[pagenum].number;
-		pdf_obj *page = pdf_load_object(ctx, doc, num);
-		if (pdf_name_eq(ctx, PDF_NAME_Page, pdf_dict_get(ctx, page, PDF_NAME_Type)))
+		page = pdf_load_object(ctx, doc, num);
+		if (pdf_name_eq(ctx, PDF_NAME(Page), pdf_dict_get(ctx, page, PDF_NAME(Type))))
 		{
 			/* We have found the page object! */
 			DEBUGMESS((ctx, "LoadHintedPage pagenum=%d num=%d", pagenum, num));
 			doc->linear_page_refs[pagenum] = pdf_new_indirect(ctx, doc, num, 0);
 		}
-		pdf_drop_obj(ctx, page);
 	}
+	fz_always(ctx)
+		pdf_drop_obj(ctx, page);
 	fz_catch(ctx)
 	{
 		fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
@@ -1915,6 +1952,23 @@ read_hinted_object(fz_context *ctx, pdf_document *doc, int num)
 	return expected != 0;
 }
 
+pdf_obj *
+pdf_load_unencrypted_object(fz_context *ctx, pdf_document *doc, int num)
+{
+	pdf_xref_entry *x;
+
+	if (num <= 0 || num >= pdf_xref_len(ctx, doc))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "object out of range (%d 0 R); xref size %d", num, pdf_xref_len(ctx, doc));
+
+	x = pdf_get_xref_entry(ctx, doc, num);
+	if (x->type == 'n')
+	{
+		fz_seek(ctx, doc->file, x->ofs, SEEK_SET);
+		return pdf_parse_ind_obj(ctx, doc, doc->file, &doc->lexbuf.base, NULL, NULL, NULL, NULL);
+	}
+	return NULL;
+}
+
 pdf_xref_entry *
 pdf_cache_object(fz_context *ctx, pdf_document *doc, int num)
 {
@@ -1937,7 +1991,7 @@ object_updated:
 
 	if (x->type == 'f')
 	{
-		x->obj = pdf_new_null(ctx, doc);
+		x->obj = PDF_NULL;
 	}
 	else if (x->type == 'n')
 	{
@@ -1976,6 +2030,7 @@ object_updated:
 			}
 			fz_catch(ctx)
 			{
+				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
 				if (rnum == num)
 					fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse object (%d 0 R)", num);
 				else
@@ -2019,7 +2074,6 @@ pdf_obj *
 pdf_load_object(fz_context *ctx, pdf_document *doc, int num)
 {
 	pdf_xref_entry *entry = pdf_cache_object(ctx, doc, num);
-	assert(entry->obj != NULL);
 	return pdf_keep_obj(ctx, entry->obj);
 }
 
@@ -2079,6 +2133,9 @@ pdf_count_objects(fz_context *ctx, pdf_document *doc)
 	return pdf_xref_len(ctx, doc);
 }
 
+/*
+	Allocate a slot in the xref table and return a fresh unused object number.
+*/
 int
 pdf_create_object(fz_context *ctx, pdf_document *doc)
 {
@@ -2100,6 +2157,9 @@ pdf_create_object(fz_context *ctx, pdf_document *doc)
 	return num;
 }
 
+/*
+	Remove object from xref table, marking the slot as free.
+*/
 void
 pdf_delete_object(fz_context *ctx, pdf_document *doc, int num)
 {
@@ -2125,6 +2185,9 @@ pdf_delete_object(fz_context *ctx, pdf_document *doc, int num)
 	x->obj = NULL;
 }
 
+/*
+	Replace object in xref table with the passed in object.
+*/
 void
 pdf_update_object(fz_context *ctx, pdf_document *doc, int num, pdf_obj *newobj)
 {
@@ -2153,6 +2216,13 @@ pdf_update_object(fz_context *ctx, pdf_document *doc, int num, pdf_obj *newobj)
 	pdf_set_obj_parent(ctx, newobj, num);
 }
 
+/*
+	Replace stream contents for object in xref table with the passed in buffer.
+
+	The buffer contents must match the /Filter setting if 'compressed' is true.
+	If 'compressed' is false, the /Filter and /DecodeParms entries are deleted.
+	The /Length entry is updated.
+*/
 void
 pdf_update_stream(fz_context *ctx, pdf_document *doc, pdf_obj *obj, fz_buffer *newbuf, int compressed)
 {
@@ -2174,11 +2244,11 @@ pdf_update_stream(fz_context *ctx, pdf_document *doc, pdf_obj *obj, fz_buffer *n
 	fz_drop_buffer(ctx, x->stm_buf);
 	x->stm_buf = fz_keep_buffer(ctx, newbuf);
 
-	pdf_dict_put_int(ctx, obj, PDF_NAME_Length, (int)fz_buffer_storage(ctx, newbuf, NULL));
+	pdf_dict_put_int(ctx, obj, PDF_NAME(Length), (int)fz_buffer_storage(ctx, newbuf, NULL));
 	if (!compressed)
 	{
-		pdf_dict_del(ctx, obj, PDF_NAME_Filter);
-		pdf_dict_del(ctx, obj, PDF_NAME_DecodeParms);
+		pdf_dict_del(ctx, obj, PDF_NAME(Filter));
+		pdf_dict_del(ctx, obj, PDF_NAME(DecodeParms));
 	}
 }
 
@@ -2192,10 +2262,10 @@ pdf_lookup_metadata(fz_context *ctx, pdf_document *doc, const char *key, char *b
 	{
 		if (doc->crypt)
 			return (int)fz_snprintf(buf, size, "Standard V%d R%d %d-bit %s",
-					pdf_crypt_version(ctx, doc),
-					pdf_crypt_revision(ctx, doc),
-					pdf_crypt_length(ctx, doc),
-					pdf_crypt_method(ctx, doc));
+					pdf_crypt_version(ctx, doc->crypt),
+					pdf_crypt_revision(ctx, doc->crypt),
+					pdf_crypt_length(ctx, doc->crypt),
+					pdf_crypt_method(ctx, doc->crypt));
 		else
 			return (int)fz_strlcpy(buf, "None", size);
 	}
@@ -2203,10 +2273,10 @@ pdf_lookup_metadata(fz_context *ctx, pdf_document *doc, const char *key, char *b
 	if (strstr(key, "info:") == key)
 	{
 		pdf_obj *info;
-		char *s;
+		const char *s;
 		int n;
 
-		info = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Info);
+		info = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(Info));
 		if (!info)
 			return -1;
 
@@ -2214,9 +2284,8 @@ pdf_lookup_metadata(fz_context *ctx, pdf_document *doc, const char *key, char *b
 		if (!info)
 			return -1;
 
-		s = pdf_to_utf8(ctx, info);
+		s = pdf_to_text_string(ctx, info);
 		n = (int)fz_strlcpy(buf, s, size);
-		fz_free(ctx, s);
 		return n;
 	}
 
@@ -2255,6 +2324,15 @@ pdf_new_document(fz_context *ctx, fz_stream *file)
 	return doc;
 }
 
+/*
+	Opens a PDF document.
+
+	Same as pdf_open_document, but takes a stream instead of a
+	filename to locate the PDF document to open. Increments the
+	reference count of the stream. See fz_open_file,
+	fz_open_file_w or fz_open_fd for opening a stream, and
+	fz_drop_stream for closing an open stream.
+*/
 pdf_document *
 pdf_open_document_with_stream(fz_context *ctx, fz_stream *file)
 {
@@ -2265,12 +2343,30 @@ pdf_open_document_with_stream(fz_context *ctx, fz_stream *file)
 	}
 	fz_catch(ctx)
 	{
+		int caught = fz_caught(ctx);
 		fz_drop_document(ctx, &doc->super);
-		fz_rethrow(ctx);
+		fz_throw(ctx, caught, "Failed to open doc from stream");
 	}
 	return doc;
 }
 
+/*
+	Open a PDF document.
+
+	Open a PDF document by reading its cross reference table, so
+	MuPDF can locate PDF objects inside the file. Upon an broken
+	cross reference table or other parse errors MuPDF will restart
+	parsing the file from the beginning to try to rebuild a
+	(hopefully correct) cross reference table to allow further
+	processing of the file.
+
+	The returned pdf_document should be used when calling most
+	other PDF functions. Note that it wraps the context, so those
+	functions implicitly get access to the global state in
+	context.
+
+	filename: a path to a file as it would be given to open(2).
+*/
 pdf_document *
 pdf_open_document(fz_context *ctx, const char *filename)
 {
@@ -2326,13 +2422,13 @@ pdf_load_hints(fz_context *ctx, pdf_document *doc, int objnum)
 		if (dict == NULL || !pdf_is_dict(ctx, dict))
 			fz_throw(ctx, FZ_ERROR_GENERIC, "malformed hint object");
 
-		shared_hint_offset = pdf_to_int(ctx, pdf_dict_get(ctx, dict, PDF_NAME_S));
+		shared_hint_offset = pdf_dict_get_int(ctx, dict, PDF_NAME(S));
 
 		/* Malloc the structures (use realloc to cope with the fact we
 		 * may try this several times before enough data is loaded) */
-		doc->hint_page = fz_resize_array(ctx, doc->hint_page, doc->linear_page_count+1, sizeof(*doc->hint_page));
+		doc->hint_page = fz_realloc_array(ctx, doc->hint_page, doc->linear_page_count+1, pdf_hint_page);
 		memset(doc->hint_page, 0, sizeof(*doc->hint_page) * (doc->linear_page_count+1));
-		doc->hint_obj_offsets = fz_resize_array(ctx, doc->hint_obj_offsets, max_object_num, sizeof(*doc->hint_obj_offsets));
+		doc->hint_obj_offsets = fz_realloc_array(ctx, doc->hint_obj_offsets, max_object_num, int64_t);
 		memset(doc->hint_obj_offsets, 0, sizeof(*doc->hint_obj_offsets) * max_object_num);
 		doc->hint_obj_offsets_max = max_object_num;
 
@@ -2394,7 +2490,7 @@ pdf_load_hints(fz_context *ctx, pdf_document *doc, int objnum)
 			shared += num_shared_objs;
 		}
 		doc->hint_page[i].index = shared;
-		doc->hint_shared_ref = fz_resize_array(ctx, doc->hint_shared_ref, shared, sizeof(*doc->hint_shared_ref));
+		doc->hint_shared_ref = fz_realloc_array(ctx, doc->hint_shared_ref, shared, int);
 		memset(doc->hint_shared_ref, 0, sizeof(*doc->hint_shared_ref) * shared);
 		fz_sync_bits(ctx, stream);
 		/* Item 4: Shared references */
@@ -2428,7 +2524,7 @@ pdf_load_hints(fz_context *ctx, pdf_document *doc, int objnum)
 			}
 		}
 
-		doc->hint_shared = fz_resize_array(ctx, doc->hint_shared, shared_obj_count_total+1, sizeof(*doc->hint_shared));
+		doc->hint_shared = fz_realloc_array(ctx, doc->hint_shared, shared_obj_count_total+1, pdf_hint_shared);
 		memset(doc->hint_shared, 0, sizeof(*doc->hint_shared) * (shared_obj_count_total+1));
 
 		/* Item 1: Shared references */
@@ -2598,8 +2694,8 @@ pdf_obj *pdf_progressive_advance(fz_context *ctx, pdf_document *doc, int pagenum
 			pdf_obj *pages;
 			doc->linear_pos = doc->file_length;
 			pdf_load_xref(ctx, doc, buf);
-			catalog = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME_Root);
-			pages = pdf_dict_get(ctx, catalog, PDF_NAME_Pages);
+			catalog = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(Root));
+			pages = pdf_dict_get(ctx, catalog, PDF_NAME(Pages));
 
 			if (!pdf_is_dict(ctx, pages))
 				fz_throw(ctx, FZ_ERROR_GENERIC, "missing page tree");
@@ -2628,6 +2724,10 @@ pdf_obj *pdf_progressive_advance(fz_context *ctx, pdf_document *doc, int pagenum
 	return doc->linear_page_refs[pagenum];
 }
 
+/*
+		Down-cast generic fitz objects into pdf specific variants.
+		Returns NULL if the objects are not from a PDF document.
+*/
 pdf_document *pdf_document_from_fz_document(fz_context *ctx, fz_document *ptr)
 {
 	return (pdf_document *)((ptr && ptr->count_pages == (fz_document_count_pages_fn*)pdf_count_pages) ? ptr : NULL);
@@ -2638,11 +2738,10 @@ pdf_page *pdf_page_from_fz_page(fz_context *ctx, fz_page *ptr)
 	return (pdf_page *)((ptr && ptr->bound_page == (fz_page_bound_page_fn*)pdf_bound_page) ? ptr : NULL);
 }
 
-pdf_annot *pdf_annot_from_fz_annot(fz_context *ctx, fz_annot *ptr)
-{
-	return (pdf_annot *)((ptr && ptr->bound_annot == (fz_annot_bound_fn*)pdf_bound_annot) ? ptr : NULL);
-}
-
+/*
+	down-cast a fz_document to a pdf_document.
+	Returns NULL if underlying document is not PDF
+*/
 pdf_document *pdf_specifics(fz_context *ctx, fz_document *doc)
 {
 	return pdf_document_from_fz_document(ctx, doc);
@@ -2729,13 +2828,13 @@ pdf_document *pdf_create_document(fz_context *ctx)
 		pdf_get_populating_xref_entry(ctx, doc, 0);
 
 		trailer = pdf_new_dict(ctx, doc, 2);
-		pdf_dict_put_int(ctx, trailer, PDF_NAME_Size, 3);
-		pdf_dict_put_drop(ctx, trailer, PDF_NAME_Root, root = pdf_add_new_dict(ctx, doc, 2));
-		pdf_dict_put(ctx, root, PDF_NAME_Type, PDF_NAME_Catalog);
-		pdf_dict_put_drop(ctx, root, PDF_NAME_Pages, pages = pdf_add_new_dict(ctx, doc, 3));
-		pdf_dict_put(ctx, pages, PDF_NAME_Type, PDF_NAME_Pages);
-		pdf_dict_put_int(ctx, pages, PDF_NAME_Count, 0);
-		pdf_dict_put_array(ctx, pages, PDF_NAME_Kids, 1);
+		pdf_dict_put_int(ctx, trailer, PDF_NAME(Size), 3);
+		pdf_dict_put_drop(ctx, trailer, PDF_NAME(Root), root = pdf_add_new_dict(ctx, doc, 2));
+		pdf_dict_put(ctx, root, PDF_NAME(Type), PDF_NAME(Catalog));
+		pdf_dict_put_drop(ctx, root, PDF_NAME(Pages), pages = pdf_add_new_dict(ctx, doc, 3));
+		pdf_dict_put(ctx, pages, PDF_NAME(Type), PDF_NAME(Pages));
+		pdf_dict_put_int(ctx, pages, PDF_NAME(Count), 0);
+		pdf_dict_put_array(ctx, pages, PDF_NAME(Kids), 1);
 
 		/* Set the trailer of the final xref section. */
 		doc->xref_sections[0].trailer = trailer;
@@ -2789,7 +2888,7 @@ void pdf_mark_xref(fz_context *ctx, pdf_document *doc)
 				pdf_xref_entry *entry = &sub->table[e];
 				if (entry->obj)
 				{
-					entry->flags |= PDF_OBJ_FLAG_MARK;
+					entry->marked = 1;
 				}
 			}
 		}
@@ -2844,7 +2943,7 @@ void pdf_clear_xref_to_mark(fz_context *ctx, pdf_document *doc)
 				 * been updated */
 				if (entry->obj != NULL && entry->stm_buf == NULL)
 				{
-					if ((entry->flags & PDF_OBJ_FLAG_MARK) == 0 && pdf_obj_refs(ctx, entry->obj) == 1)
+					if (!entry->marked && pdf_obj_refs(ctx, entry->obj) == 1)
 					{
 						pdf_drop_obj(ctx, entry->obj);
 						entry->obj = NULL;
