@@ -29,7 +29,7 @@ svg_bound_page(fz_context *ctx, fz_page *page_)
 	svg_page *page = (svg_page*)page_;
 	svg_document *doc = page->doc;
 
-	svg_parse_document_bounds(ctx, doc, fz_xml_root(doc->xml));
+	svg_parse_document_bounds(ctx, doc, doc->root);
 
 	return fz_make_rect(0, 0, doc->width, doc->height);
 }
@@ -39,7 +39,7 @@ svg_run_page(fz_context *ctx, fz_page *page_, fz_device *dev, fz_matrix ctm, fz_
 {
 	svg_page *page = (svg_page*)page_;
 	svg_document *doc = page->doc;
-	svg_run_document(ctx, doc, fz_xml_root(doc->xml), dev, ctm);
+	svg_run_document(ctx, doc, doc->root, dev, ctm);
 }
 
 static void
@@ -80,7 +80,7 @@ svg_build_id_map(fz_context *ctx, svg_document *doc, fz_xml *root)
 }
 
 static fz_document *
-svg_open_document_with_buffer(fz_context *ctx, fz_buffer *buf)
+svg_open_document_with_xml(fz_context *ctx, fz_xml *xml, const char *base_uri, fz_archive *zip)
 {
 	svg_document *doc;
 
@@ -90,11 +90,43 @@ svg_open_document_with_buffer(fz_context *ctx, fz_buffer *buf)
 	doc->super.load_page = svg_load_page;
 
 	doc->idmap = NULL;
+	doc->xml = NULL;
+	doc->root = xml;
+	doc->zip = zip;
+
+	fz_try(ctx)
+	{
+		svg_build_id_map(ctx, doc, doc->root);
+	}
+	fz_catch(ctx)
+	{
+		fz_drop_document(ctx, &doc->super);
+		fz_rethrow(ctx);
+	}
+
+	return (fz_document*)doc;
+}
+
+static fz_document *
+svg_open_document_with_buffer(fz_context *ctx, fz_buffer *buf, const char *base_uri, fz_archive *zip)
+{
+	svg_document *doc;
+
+	doc = fz_new_derived_document(ctx, svg_document);
+	doc->super.drop_document = svg_drop_document;
+	doc->super.count_pages = svg_count_pages;
+	doc->super.load_page = svg_load_page;
+
+	doc->idmap = NULL;
+	if (base_uri)
+		fz_strlcpy(doc->base_uri, base_uri, sizeof doc->base_uri);
+	doc->zip = zip;
 
 	fz_try(ctx)
 	{
 		doc->xml = fz_parse_xml(ctx, buf, 0);
-		svg_build_id_map(ctx, doc, fz_xml_root(doc->xml));
+		doc->root = fz_xml_root(doc->xml);
+		svg_build_id_map(ctx, doc, doc->root);
 	}
 	fz_catch(ctx)
 	{
@@ -113,7 +145,7 @@ svg_open_document_with_stream(fz_context *ctx, fz_stream *file)
 
 	buf = fz_read_all(ctx, file, 0);
 	fz_try(ctx)
-		doc = svg_open_document_with_buffer(ctx, buf);
+		doc = svg_open_document_with_buffer(ctx, buf, NULL, NULL);
 	fz_always(ctx)
 		fz_drop_buffer(ctx, buf);
 	fz_catch(ctx)
@@ -126,12 +158,36 @@ svg_open_document_with_stream(fz_context *ctx, fz_stream *file)
 	Parse an SVG document into a display-list.
 */
 fz_display_list *
-fz_new_display_list_from_svg(fz_context *ctx, fz_buffer *buf, float *w, float *h)
+fz_new_display_list_from_svg(fz_context *ctx, fz_buffer *buf, const char *base_uri, fz_archive *zip, float *w, float *h)
 {
 	fz_document *doc;
 	fz_display_list *list = NULL;
 
-	doc = svg_open_document_with_buffer(ctx, buf);
+	doc = svg_open_document_with_buffer(ctx, buf, base_uri, zip);
+	fz_try(ctx)
+	{
+		list = fz_new_display_list_from_page_number(ctx, doc, 0);
+		*w = ((svg_document*)doc)->width;
+		*h = ((svg_document*)doc)->height;
+	}
+	fz_always(ctx)
+		fz_drop_document(ctx, doc);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return list;
+}
+
+/*
+	Parse an SVG document into a display-list.
+*/
+fz_display_list *
+fz_new_display_list_from_svg_xml(fz_context *ctx, fz_xml *xml, const char *base_uri, fz_archive *zip, float *w, float *h)
+{
+	fz_document *doc;
+	fz_display_list *list = NULL;
+
+	doc = svg_open_document_with_xml(ctx, xml, base_uri, zip);
 	fz_try(ctx)
 	{
 		list = fz_new_display_list_from_page_number(ctx, doc, 0);
@@ -150,13 +206,33 @@ fz_new_display_list_from_svg(fz_context *ctx, fz_buffer *buf, float *w, float *h
 	Create a scalable image from an SVG document.
 */
 fz_image *
-fz_new_image_from_svg(fz_context *ctx, fz_buffer *buf)
+fz_new_image_from_svg(fz_context *ctx, fz_buffer *buf, const char *base_uri, fz_archive *zip)
 {
 	fz_display_list *list;
 	fz_image *image = NULL;
 	float w, h;
 
-	list = fz_new_display_list_from_svg(ctx, buf, &w, &h);
+	list = fz_new_display_list_from_svg(ctx, buf, base_uri, zip, &w, &h);
+	fz_try(ctx)
+		image = fz_new_image_from_display_list(ctx, w, h, list);
+	fz_always(ctx)
+		fz_drop_display_list(ctx, list);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+	return image;
+}
+
+/*
+	Create a scalable image from an SVG document.
+*/
+fz_image *
+fz_new_image_from_svg_xml(fz_context *ctx, fz_xml *xml, const char *base_uri, fz_archive *zip)
+{
+	fz_display_list *list;
+	fz_image *image = NULL;
+	float w, h;
+
+	list = fz_new_display_list_from_svg_xml(ctx, xml, base_uri, zip, &w, &h);
 	fz_try(ctx)
 		image = fz_new_image_from_display_list(ctx, w, h, list);
 	fz_always(ctx)
