@@ -2,20 +2,22 @@
 #include "svg-imp.h"
 
 #include <string.h>
-#include <stdio.h> /* for sscanf */
 #include <math.h>
 
 /* default page size */
-#define DEF_WIDTH 12
+#define DEF_WIDTH 612
 #define DEF_HEIGHT 792
 #define DEF_FONTSIZE 12
 
+#define MAX_USE_DEPTH 100
+
 typedef struct svg_state_s svg_state;
 
 struct svg_state_s
 {
 	fz_matrix transform;
 	fz_stroke_state stroke;
+	int use_depth;
 
 	float viewport_w, viewport_h;
 	float viewbox_w, viewbox_h, viewbox_size;
@@ -39,13 +41,15 @@ static void svg_run_element(fz_context *ctx, fz_device *dev, svg_document *doc,
 static void svg_fill(fz_context *ctx, fz_device *dev, svg_document *doc, fz_path *path, svg_state *state)
 {
 	float opacity = state->opacity * state->fill_opacity;
-	fz_fill_path(ctx, dev, path, state->fill_rule, &state->transform, fz_device_rgb(ctx), state->fill_color, opacity, NULL);
+	if (path)
+		fz_fill_path(ctx, dev, path, state->fill_rule, state->transform, fz_device_rgb(ctx), state->fill_color, opacity, fz_default_color_params);
 }
 
 static void svg_stroke(fz_context *ctx, fz_device *dev, svg_document *doc, fz_path *path, svg_state *state)
 {
 	float opacity = state->opacity * state->stroke_opacity;
-	fz_stroke_path(ctx, dev, path, &state->stroke, &state->transform, fz_device_rgb(ctx), state->stroke_color, opacity, NULL);
+	if (path)
+		fz_stroke_path(ctx, dev, path, &state->stroke, state->transform, fz_device_rgb(ctx), state->stroke_color, opacity, fz_default_color_params);
 }
 
 static void svg_draw_path(fz_context *ctx, fz_device *dev, svg_document *doc, fz_path *path, svg_state *state)
@@ -65,13 +69,13 @@ static const float MAGIC_CIRCLE = 0.551915f;
 
 static void approx_circle(fz_context *ctx, fz_path *path, float cx, float cy, float rx, float ry)
 {
-	float rxs = rx * MAGIC_CIRCLE;
-	float rys = ry * MAGIC_CIRCLE;
+	float mx = rx * MAGIC_CIRCLE;
+	float my = ry * MAGIC_CIRCLE;
 	fz_moveto(ctx, path, cx, cy+ry);
-	fz_curveto(ctx, path, cx + rxs, cy + ry, cx + rx, cy + rys, cx + rx, cy);
-	fz_curveto(ctx, path, cx + rx, cy - rys, cx + rxs, cy - rys, cx, cy - ry);
-	fz_curveto(ctx, path, cx - rxs, cy - ry, cx - rx, cy - rys, cx - rx, cy);
-	fz_curveto(ctx, path, cx - rx, cy + rys, cx - rxs, cy + rys, cx, cy + ry);
+	fz_curveto(ctx, path, cx + mx, cy + ry, cx + rx, cy + my, cx + rx, cy);
+	fz_curveto(ctx, path, cx + rx, cy - my, cx + mx, cy - ry, cx, cy - ry);
+	fz_curveto(ctx, path, cx - mx, cy - ry, cx - rx, cy - my, cx - rx, cy);
+	fz_curveto(ctx, path, cx - rx, cy + my, cx - mx, cy + ry, cx, cy + ry);
 	fz_closepath(ctx, path);
 }
 
@@ -118,31 +122,37 @@ svg_run_rect(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, c
 		return;
 
 	path = fz_new_path(ctx);
-	if (rx == 0 || ry == 0)
+	fz_try(ctx)
 	{
-		fz_moveto(ctx, path, x, y);
-		fz_lineto(ctx, path, x + w, y);
-		fz_lineto(ctx, path, x + w, y + h);
-		fz_lineto(ctx, path, x, y + h);
-	}
-	else
-	{
-		float rxs = rx * MAGIC_CIRCLE;
-		float rys = rx * MAGIC_CIRCLE;
-		fz_moveto(ctx, path, x + w - rx, y);
-		fz_curveto(ctx, path, x + w - rxs, y, x + w, y + rys, x + w, y + ry);
-		fz_lineto(ctx, path, x + w, y + h - ry);
-		fz_curveto(ctx, path, x + w, y + h - rys, x + w - rxs, y + h, x + w - rx, y + h);
-		fz_lineto(ctx, path, x + rx, y + h);
-		fz_curveto(ctx, path, x + rxs, y + h, x, y + h - rys, x, y + h - rx);
-		fz_lineto(ctx, path, x, y + rx);
-		fz_curveto(ctx, path, x, y + rxs, x + rxs, y, x + rx, y);
-	}
-	fz_closepath(ctx, path);
+		if (rx == 0 || ry == 0)
+		{
+			fz_moveto(ctx, path, x, y);
+			fz_lineto(ctx, path, x + w, y);
+			fz_lineto(ctx, path, x + w, y + h);
+			fz_lineto(ctx, path, x, y + h);
+		}
+		else
+		{
+			float rxs = rx * MAGIC_CIRCLE;
+			float rys = rx * MAGIC_CIRCLE;
+			fz_moveto(ctx, path, x + w - rx, y);
+			fz_curveto(ctx, path, x + w - rxs, y, x + w, y + rys, x + w, y + ry);
+			fz_lineto(ctx, path, x + w, y + h - ry);
+			fz_curveto(ctx, path, x + w, y + h - rys, x + w - rxs, y + h, x + w - rx, y + h);
+			fz_lineto(ctx, path, x + rx, y + h);
+			fz_curveto(ctx, path, x + rxs, y + h, x, y + h - rys, x, y + h - rx);
+			fz_lineto(ctx, path, x, y + rx);
+			fz_curveto(ctx, path, x, y + rxs, x + rxs, y, x + rx, y);
+		}
+		fz_closepath(ctx, path);
 
-	svg_draw_path(ctx, dev, doc, path, &local_state);
+		svg_draw_path(ctx, dev, doc, path, &local_state);
+	}
+	fz_always(ctx)
+		fz_drop_path(ctx, path);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 
-	fz_drop_path(ctx, path);
 }
 
 static void
@@ -169,9 +179,15 @@ svg_run_circle(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node,
 		return;
 
 	path = fz_new_path(ctx);
-	approx_circle(ctx, path, cx, cy, r, r);
-	svg_draw_path(ctx, dev, doc, path, &local_state);
-	fz_drop_path(ctx, path);
+	fz_try(ctx)
+	{
+		approx_circle(ctx, path, cx, cy, r, r);
+		svg_draw_path(ctx, dev, doc, path, &local_state);
+	}
+	fz_always(ctx)
+		fz_drop_path(ctx, path);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static void
@@ -202,9 +218,15 @@ svg_run_ellipse(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node
 		return;
 
 	path = fz_new_path(ctx);
-	approx_circle(ctx, path, cx, cy, rx, ry);
-	svg_draw_path(ctx, dev, doc, path, &local_state);
-	fz_drop_path(ctx, path);
+	fz_try(ctx)
+	{
+		approx_circle(ctx, path, cx, cy, rx, ry);
+		svg_draw_path(ctx, dev, doc, path, &local_state);
+	}
+	fz_always(ctx)
+		fz_drop_path(ctx, path);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static void
@@ -232,10 +254,16 @@ svg_run_line(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, c
 	if (local_state.stroke_is_set)
 	{
 		fz_path *path = fz_new_path(ctx);
-		fz_moveto(ctx, path, x1, y1);
-		fz_lineto(ctx, path, x2, y2);
-		svg_stroke(ctx, dev, doc, path, &local_state);
-		fz_drop_path(ctx, path);
+		fz_try(ctx)
+		{
+			fz_moveto(ctx, path, x1, y1);
+			fz_lineto(ctx, path, x2, y2);
+			svg_stroke(ctx, dev, doc, path, &local_state);
+		}
+		fz_always(ctx)
+			fz_drop_path(ctx, path);
+		fz_catch(ctx)
+			fz_rethrow(ctx);
 	}
 }
 
@@ -257,32 +285,39 @@ svg_parse_polygon_imp(fz_context *ctx, svg_document *doc, fz_xml *node, int docl
 	nargs = 0;
 
 	path = fz_new_path(ctx);
-
-	while (*str)
+	fz_try(ctx)
 	{
-		while (svg_is_whitespace_or_comma(*str))
-			str ++;
-
-		if (svg_is_digit(*str))
+		while (*str)
 		{
-			str = svg_lex_number(&number, str);
-			args[nargs++] = number;
-		}
+			while (svg_is_whitespace_or_comma(*str))
+				str ++;
 
-		if (nargs == 2)
-		{
-			if (isfirst)
+			if (svg_is_digit(*str))
 			{
-				fz_moveto(ctx, path, args[0], args[1]);
-				isfirst = 0;
+				str = svg_lex_number(&number, str);
+				args[nargs++] = number;
 			}
-			else
+
+			if (nargs == 2)
 			{
-				fz_lineto(ctx, path, args[0], args[1]);
+				if (isfirst)
+				{
+					fz_moveto(ctx, path, args[0], args[1]);
+					isfirst = 0;
+				}
+				else
+				{
+					fz_lineto(ctx, path, args[0], args[1]);
+				}
+				nargs = 0;
 			}
-			nargs = 0;
 		}
 	}
+	fz_catch(ctx)
+	{
+		fz_drop_path(ctx, path);
+		fz_rethrow(ctx);
+	}
 
 	return path;
 }
@@ -297,8 +332,12 @@ svg_run_polyline(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *nod
 	if (local_state.stroke_is_set)
 	{
 		fz_path *path = svg_parse_polygon_imp(ctx, doc, node, 0);
-		svg_stroke(ctx, dev, doc, path, &local_state);
-		fz_drop_path(ctx, path);
+		fz_try(ctx)
+			svg_stroke(ctx, dev, doc, path, &local_state);
+		fz_always(ctx)
+			fz_drop_path(ctx, path);
+		fz_catch(ctx)
+			fz_rethrow(ctx);
 	}
 }
 
@@ -311,12 +350,16 @@ svg_run_polygon(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node
 	svg_parse_common(ctx, doc, node, &local_state);
 
 	path = svg_parse_polygon_imp(ctx, doc, node, 1);
-	svg_draw_path(ctx, dev, doc, path, &local_state);
-	fz_drop_path(ctx, path);
+	fz_try(ctx)
+		svg_draw_path(ctx, dev, doc, path, &local_state);
+	fz_always(ctx)
+		fz_drop_path(ctx, path);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static void
-svg_add_arc_segment(fz_context *ctx, fz_path *path, const fz_matrix *mtx, float th0, float th1, int iscw)
+svg_add_arc_segment(fz_context *ctx, fz_path *path, fz_matrix mtx, float th0, float th1, int iscw)
 {
 	float t, d;
 	fz_point p;
@@ -330,7 +373,7 @@ svg_add_arc_segment(fz_context *ctx, fz_path *path, const fz_matrix *mtx, float
 	{
 		for (t = th0 + d; t < th1 - d/2; t += d)
 		{
-			fz_transform_point_xy(&p, mtx, cosf(t), sinf(t));
+			p = fz_transform_point_xy(cosf(t), sinf(t), mtx);
 			fz_lineto(ctx, path, p.x, p.y);
 		}
 	}
@@ -339,7 +382,7 @@ svg_add_arc_segment(fz_context *ctx, fz_path *path, const fz_matrix *mtx, float
 		th0 += FZ_PI * 2;
 		for (t = th0 - d; t > th1 + d/2; t -= d)
 		{
-			fz_transform_point_xy(&p, mtx, cosf(t), sinf(t));
+			p = fz_transform_point_xy(cosf(t), sinf(t), mtx);
 			fz_lineto(ctx, path, p.x, p.y);
 		}
 	}
@@ -390,8 +433,8 @@ svg_add_arc(fz_context *ctx, fz_path *path,
 	else
 		sign = -1;
 
-	fz_rotate(&rotmat, rotation_angle);
-	fz_rotate(&revmat, -rotation_angle);
+	rotmat = fz_rotate(rotation_angle);
+	revmat = fz_rotate(-rotation_angle);
 
 	/* http://www.w3.org/TR/SVG11/implnote.html#ArcImplementationNotes */
 	/* Conversion from endpoint to center parameterization */
@@ -408,7 +451,7 @@ svg_add_arc(fz_context *ctx, fz_path *path,
 	/* F.6.5.1 */
 	pt.x = (x1 - x2) / 2;
 	pt.y = (y1 - y2) / 2;
-	fz_transform_vector(&pt, &revmat);
+	pt = fz_transform_vector(pt, revmat);
 	x1t = pt.x;
 	y1t = pt.y;
 
@@ -434,7 +477,7 @@ svg_add_arc(fz_context *ctx, fz_path *path,
 	/* F.6.5.3 */
 	pt.x = cxt;
 	pt.y = cyt;
-	fz_transform_vector(&pt, &rotmat);
+	pt = fz_transform_vector(pt, rotmat);
 	cx = pt.x + (x1 + x2) / 2;
 	cy = pt.y + (y1 + y2) / 2;
 
@@ -457,8 +500,8 @@ svg_add_arc(fz_context *ctx, fz_path *path,
 			dth -= ((FZ_PI / 180) * 360);
 	}
 
-	fz_pre_scale(fz_pre_rotate(fz_translate(&mtx, cx, cy), rotation_angle), rx, ry);
-	svg_add_arc_segment(ctx, path, &mtx, th1, th1 + dth, is_clockwise);
+	mtx = fz_pre_scale(fz_pre_rotate(fz_translate(cx, cy), rotation_angle), rx, ry);
+	svg_add_arc_segment(ctx, path, mtx, th1, th1 + dth, is_clockwise);
 
 	fz_lineto(ctx, path, point_x, point_y);
 }
@@ -466,7 +509,7 @@ svg_add_arc(fz_context *ctx, fz_path *path,
 static fz_path *
 svg_parse_path_data(fz_context *ctx, svg_document *doc, const char *str)
 {
-	fz_path *path = fz_new_path(ctx);
+	fz_path *path;
 
 	fz_point p;
 	float x1, y1, x2, y2;
@@ -484,6 +527,7 @@ svg_parse_path_data(fz_context *ctx, svg_document *doc, const char *str)
 	cmd = 0;
 	nargs = 0;
 
+	path = fz_new_path(ctx);
 	fz_try(ctx)
 	{
 		fz_moveto(ctx, path, 0.0f, 0.0f); /* for the case of opening 'm' */
@@ -785,8 +829,12 @@ svg_run_path(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, c
 	if (d_att)
 	{
 		fz_path *path = svg_parse_path_data(ctx, doc, d_att);
-		svg_draw_path(ctx, dev, doc, path, &local_state);
-		fz_drop_path(ctx, path);
+		fz_try(ctx)
+			svg_draw_path(ctx, dev, doc, path, &local_state);
+		fz_always(ctx)
+			fz_drop_path(ctx, path);
+		fz_catch(ctx)
+			fz_rethrow(ctx);
 	}
 }
 
@@ -794,46 +842,76 @@ svg_run_path(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, c
 void
 svg_parse_viewport(fz_context *ctx, svg_document *doc, fz_xml *node, svg_state *state)
 {
-	//fz_matrix *transform = &state->transform;
-
-	char *x_att = fz_xml_att(node, "x");
-	char *y_att = fz_xml_att(node, "y");
 	char *w_att = fz_xml_att(node, "width");
 	char *h_att = fz_xml_att(node, "height");
 
-	float x = 0;
-	float y = 0;
-	float w = state->viewport_w;
-	float h = state->viewport_h;
+	if (w_att)
+		state->viewport_w = svg_parse_length(w_att, state->viewbox_w, state->fontsize);
+	if (h_att)
+		state->viewport_h = svg_parse_length(h_att, state->viewbox_h, state->fontsize);
 
-	if (x_att) x = svg_parse_length(x_att, state->viewbox_w, state->fontsize);
-	if (y_att) y = svg_parse_length(y_att, state->viewbox_h, state->fontsize);
-	if (w_att) w = svg_parse_length(w_att, state->viewbox_w, state->fontsize);
-	if (h_att) h = svg_parse_length(h_att, state->viewbox_h, state->fontsize);
+}
 
-	/* TODO: new transform */
-	fz_warn(ctx, "push viewport: %g %g %g %g", x, y, w, h);
+static void
+svg_lex_viewbox(const char *s, float *x, float *y, float *w, float *h)
+{
+	while (svg_is_whitespace_or_comma(*s)) ++s;
+	if (svg_is_digit(*s)) s = svg_lex_number(x, s);
+	while (svg_is_whitespace_or_comma(*s)) ++s;
+	if (svg_is_digit(*s)) s = svg_lex_number(y, s);
+	while (svg_is_whitespace_or_comma(*s)) ++s;
+	if (svg_is_digit(*s)) s = svg_lex_number(w, s);
+	while (svg_is_whitespace_or_comma(*s)) ++s;
+	if (svg_is_digit(*s)) s = svg_lex_number(h, s);
+}
 
-	state->viewport_w = w;
-	state->viewport_h = h;
+static int
+svg_parse_preserve_aspect_ratio(const char *att, int *x, int *y)
+{
+	*x = *y = 1;
+	if (strstr(att, "none")) return 0;
+	if (strstr(att, "xMin")) *x = 0;
+	if (strstr(att, "xMid")) *x = 1;
+	if (strstr(att, "xMax")) *x = 2;
+	if (strstr(att, "YMin")) *y = 0;
+	if (strstr(att, "YMid")) *y = 1;
+	if (strstr(att, "YMax")) *y = 2;
+	return 1;
 }
 
 /* svg, symbol, image, foreignObject plus marker, pattern, view can use viewBox to set the transform */
 void
 svg_parse_viewbox(fz_context *ctx, svg_document *doc, fz_xml *node, svg_state *state)
 {
-	//fz_matrix *transform = &state->transform;
-	//float port_w = state->viewport_w;
-	//float port_h = state->viewport_h;
-	float min_x, min_y, box_w, box_h;
-
 	char *viewbox_att = fz_xml_att(node, "viewBox");
+	char *preserve_att = fz_xml_att(node, "preserveAspectRatio");
 	if (viewbox_att)
 	{
-		sscanf(viewbox_att, "%g %g %g %g", &min_x, &min_y, &box_w, &box_h);
+		/* scale and translate to fit [minx miny minx+w miny+h] to [0 0 viewport.w viewport.h] */
+		float min_x, min_y, box_w, box_h, sx, sy;
+		int align_x=1, align_y=1, preserve=1;
+		float pad_x=0, pad_y=0;
 
-		/* scale and translate to fit [x y w h] to [0 0 viewport.w viewport.h] */
-		fz_warn(ctx, "push viewbox: %g %g %g %g", min_x, min_y, box_w, box_h);
+		svg_lex_viewbox(viewbox_att, &min_x, &min_y, &box_w, &box_h);
+		sx = state->viewport_w / box_w;
+		sy = state->viewport_h / box_h;
+
+		if (preserve_att)
+			preserve = svg_parse_preserve_aspect_ratio(preserve_att, &align_x, &align_y);
+		if (preserve)
+		{
+			sx = sy = fz_min(sx, sy);
+			if (align_x == 1) pad_x = (box_w * sx - state->viewport_w) / 2;
+			if (align_x == 2) pad_x = (box_w * sx - state->viewport_w);
+			if (align_y == 1) pad_y = (box_h * sy - state->viewport_h) / 2;
+			if (align_y == 2) pad_y = (box_h * sy - state->viewport_h);
+			state->transform = fz_concat(fz_translate(-pad_x, -pad_y), state->transform);
+		}
+		state->transform = fz_concat(fz_scale(sx, sy), state->transform);
+		state->transform = fz_concat(fz_translate(-min_x, -min_y), state->transform);
+		state->viewbox_w = box_w;
+		state->viewbox_h = box_h;
+		state->viewbox_size = sqrtf(box_w*box_w + box_h*box_h) / sqrtf(2);
 	}
 }
 
@@ -842,13 +920,14 @@ void
 svg_parse_common(fz_context *ctx, svg_document *doc, fz_xml *node, svg_state *state)
 {
 	fz_stroke_state *stroke = &state->stroke;
-	fz_matrix *transform = &state->transform;
 
 	char *transform_att = fz_xml_att(node, "transform");
 
 	char *font_size_att = fz_xml_att(node, "font-size");
 	// TODO: all font stuff
 
+	char *style_att = fz_xml_att(node, "style");
+
 	// TODO: clip, clip-path, clip-rule
 
 	char *opacity_att = fz_xml_att(node, "opacity");
@@ -870,9 +949,17 @@ svg_parse_common(fz_context *ctx, svg_document *doc, fz_xml *node, svg_state *st
 	// TODO: overflow
 	// TODO: mask
 
+	/* Dirty hack scans of CSS style */
+	if (style_att)
+	{
+		svg_parse_color_from_style(ctx, doc, style_att,
+			&state->fill_is_set, state->fill_color,
+			&state->stroke_is_set, state->stroke_color);
+	}
+
 	if (transform_att)
 	{
-		svg_parse_transform(ctx, doc, transform_att, transform);
+		state->transform = svg_parse_transform(ctx, doc, transform_att, state->transform);
 	}
 
 	if (font_size_att)
@@ -987,6 +1074,19 @@ svg_run_svg(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, co
 	svg_state local_state = *inherit_state;
 	fz_xml *node;
 
+	char *w_att = fz_xml_att(root, "width");
+	char *h_att = fz_xml_att(root, "height");
+	char *viewbox_att = fz_xml_att(root, "viewBox");
+
+	/* get default viewport from viewBox if width and/or height is missing */
+	if (viewbox_att && (!w_att || !h_att))
+	{
+		float x, y;
+		svg_lex_viewbox(viewbox_att, &x, &y, &local_state.viewbox_w, &local_state.viewbox_h);
+		if (!w_att) local_state.viewport_w = local_state.viewbox_w;
+		if (!h_att) local_state.viewport_h = local_state.viewbox_h;
+	}
+
 	svg_parse_viewport(ctx, doc, root, &local_state);
 	svg_parse_viewbox(ctx, doc, root, &local_state);
 	svg_parse_common(ctx, doc, root, &local_state);
@@ -1033,11 +1133,17 @@ svg_run_use(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, co
 	float x = 0;
 	float y = 0;
 
+	if (++local_state.use_depth > MAX_USE_DEPTH)
+	{
+		fz_warn(ctx, "svg: too much recursion");
+		return;
+	}
+
 	svg_parse_common(ctx, doc, root, &local_state);
 	if (x_att) x = svg_parse_length(x_att, local_state.viewbox_w, local_state.fontsize);
 	if (y_att) y = svg_parse_length(y_att, local_state.viewbox_h, local_state.fontsize);
 
-	fz_pre_translate(&local_state.transform, x, y);
+	local_state.transform = fz_concat(fz_translate(x, y), local_state.transform);
 
 	if (xlink_href_att && xlink_href_att[0] == '#')
 	{
@@ -1055,55 +1161,147 @@ svg_run_use(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, co
 	fz_warn(ctx, "svg: cannot find linked symbol");
 }
 
+static void
+svg_run_image(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, const svg_state *inherit_state)
+{
+	svg_state local_state = *inherit_state;
+	float x=0, y=0, w=0, h=0;
+	const char *data;
+
+	static const char *jpeg_uri = "data:image/jpeg;base64,";
+	static const char *png_uri = "data:image/png;base64,";
+
+	char *href_att = fz_xml_att(root, "xlink:href");
+	char *x_att = fz_xml_att(root, "x");
+	char *y_att = fz_xml_att(root, "y");
+	char *w_att = fz_xml_att(root, "width");
+	char *h_att = fz_xml_att(root, "height");
+
+	svg_parse_common(ctx, doc, root, &local_state);
+	if (x_att) x = svg_parse_length(x_att, local_state.viewbox_w, local_state.fontsize);
+	if (y_att) y = svg_parse_length(y_att, local_state.viewbox_h, local_state.fontsize);
+	if (w_att) w = svg_parse_length(w_att, local_state.viewbox_w, local_state.fontsize);
+	if (h_att) h = svg_parse_length(h_att, local_state.viewbox_h, local_state.fontsize);
+
+	if (w <= 0 || h <= 0)
+		return;
+
+	if (!href_att)
+		return;
+
+	local_state.transform = fz_concat(fz_translate(x, y), local_state.transform);
+	local_state.transform = fz_concat(fz_scale(w, h), local_state.transform);
+
+	if (!strncmp(href_att, jpeg_uri, strlen(jpeg_uri)))
+		data = href_att + strlen(jpeg_uri);
+	else if (!strncmp(href_att, png_uri, strlen(png_uri)))
+		data = href_att + strlen(png_uri);
+	else
+		data = NULL;
+	if (data)
+	{
+		fz_image *img = NULL;
+		fz_buffer *buf;
+
+		fz_var(img);
+
+		buf = fz_new_buffer_from_base64(ctx, data, 0);
+		fz_try(ctx)
+		{
+			img = fz_new_image_from_buffer(ctx, buf);
+			fz_fill_image(ctx, dev, img, local_state.transform, 1, fz_default_color_params);
+		}
+		fz_always(ctx)
+		{
+			fz_drop_buffer(ctx, buf);
+			fz_drop_image(ctx, img);
+		}
+		fz_catch(ctx)
+			fz_warn(ctx, "svg: ignoring embedded image '%s'", href_att);
+	}
+	else if (doc->zip)
+	{
+		char path[2048];
+		fz_buffer *buf = NULL;
+		fz_image *img = NULL;
+
+		fz_var(buf);
+		fz_var(img);
+
+		fz_strlcpy(path, doc->base_uri, sizeof path);
+		fz_strlcat(path, "/", sizeof path);
+		fz_strlcat(path, href_att, sizeof path);
+		fz_urldecode(path);
+		fz_cleanname(path);
+
+		fz_try(ctx)
+		{
+			buf = fz_read_archive_entry(ctx, doc->zip, path);
+			img = fz_new_image_from_buffer(ctx, buf);
+			fz_fill_image(ctx, dev, img, local_state.transform, 1, fz_default_color_params);
+		}
+		fz_always(ctx)
+		{
+			fz_drop_buffer(ctx, buf);
+			fz_drop_image(ctx, img);
+		}
+		fz_catch(ctx)
+			fz_warn(ctx, "svg: ignoring external image '%s'", href_att);
+	}
+	else
+	{
+		fz_warn(ctx, "svg: ignoring external image '%s'", href_att);
+	}
+}
+
 static void
 svg_run_element(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, const svg_state *state)
 {
-	char *tag = fz_xml_tag(root);
-
-	if (!strcmp(tag, "svg"))
+	if (fz_xml_is_tag(root, "svg"))
 		svg_run_svg(ctx, dev, doc, root, state);
 
-	else if (!strcmp(tag, "g"))
+	else if (fz_xml_is_tag(root, "g"))
 		svg_run_g(ctx, dev, doc, root, state);
 
-	else if (!strcmp(tag, "title"))
+	else if (fz_xml_is_tag(root, "title"))
 		;
-	else if (!strcmp(tag, "desc"))
+	else if (fz_xml_is_tag(root, "desc"))
 		;
 
-	else if (!strcmp(tag, "defs"))
+	else if (fz_xml_is_tag(root, "defs"))
 		;
-	else if (!strcmp(tag, "symbol"))
+	else if (fz_xml_is_tag(root, "symbol"))
 		;
 
-	else if (!strcmp(tag, "use"))
+	else if (fz_xml_is_tag(root, "use"))
 		svg_run_use(ctx, dev, doc, root, state);
 
-	else if (!strcmp(tag, "path"))
+	else if (fz_xml_is_tag(root, "path"))
 		svg_run_path(ctx, dev, doc, root, state);
-	else if (!strcmp(tag, "rect"))
+	else if (fz_xml_is_tag(root, "rect"))
 		svg_run_rect(ctx, dev, doc, root, state);
-	else if (!strcmp(tag, "circle"))
+	else if (fz_xml_is_tag(root, "circle"))
 		svg_run_circle(ctx, dev, doc, root, state);
-	else if (!strcmp(tag, "ellipse"))
+	else if (fz_xml_is_tag(root, "ellipse"))
 		svg_run_ellipse(ctx, dev, doc, root, state);
-	else if (!strcmp(tag, "line"))
+	else if (fz_xml_is_tag(root, "line"))
 		svg_run_line(ctx, dev, doc, root, state);
-	else if (!strcmp(tag, "polyline"))
+	else if (fz_xml_is_tag(root, "polyline"))
 		svg_run_polyline(ctx, dev, doc, root, state);
-	else if (!strcmp(tag, "polygon"))
+	else if (fz_xml_is_tag(root, "polygon"))
 		svg_run_polygon(ctx, dev, doc, root, state);
 
+	else if (fz_xml_is_tag(root, "image"))
+		svg_run_image(ctx, dev, doc, root, state);
+
 #if 0
-	else if (!strcmp(tag, "image"))
-		svg_parse_image(ctx, doc, root);
-	else if (!strcmp(tag, "text"))
+	else if (fz_xml_is_tag(root, "text"))
 		svg_run_text(ctx, dev, doc, root);
-	else if (!strcmp(tag, "tspan"))
+	else if (fz_xml_is_tag(root, "tspan"))
 		svg_run_text_span(ctx, dev, doc, root);
-	else if (!strcmp(tag, "tref"))
+	else if (fz_xml_is_tag(root, "tref"))
 		svg_run_text_ref(ctx, dev, doc, root);
-	else if (!strcmp(tag, "textPath"))
+	else if (fz_xml_is_tag(root, "textPath"))
 		svg_run_text_path(ctx, dev, doc, root);
 #endif
 
@@ -1141,7 +1339,7 @@ svg_parse_document_bounds(fz_context *ctx, svg_document *doc, fz_xml *root)
 	if (w_att == NULL && h_att == NULL && viewbox_att != NULL)
 	{
 		float min_x, min_y, box_w, box_h;
-		sscanf(viewbox_att, "%g %g %g %g", &min_x, &min_y, &box_w, &box_h);
+		svg_lex_viewbox(viewbox_att, &min_x, &min_y, &box_w, &box_h);
 		doc->width = box_w;
 		doc->height = box_h;
 	}
@@ -1158,15 +1356,16 @@ svg_parse_document_bounds(fz_context *ctx, svg_document *doc, fz_xml *root)
 }
 
 void
-svg_run_document(fz_context *ctx, svg_document *doc, fz_xml *root, fz_device *dev, const fz_matrix *ctm)
+svg_run_document(fz_context *ctx, svg_document *doc, fz_xml *root, fz_device *dev, fz_matrix ctm)
 {
 	svg_state state;
 
 	svg_parse_document_bounds(ctx, doc, root);
 
 	/* Initial graphics state */
-	state.transform = *ctm;
+	state.transform = ctm;
 	state.stroke = fz_default_stroke_state;
+	state.use_depth = 0;
 
 	state.viewport_w = DEF_WIDTH;
 	state.viewport_h = DEF_HEIGHT;
