@@ -117,7 +117,7 @@ static void merge_range(const char *range)
 
 int pdfmerge_main(int argc, char **argv)
 {
-	pdf_write_options opts = { 0 };
+	pdf_write_options opts = pdf_default_write_options;
 	char *output = "out.pdf";
 	char *flags = "";
 	char *input;
