@@ -164,7 +164,7 @@ int pdfposter_main(int argc, char **argv)
 	char *outfile = "out.pdf";
 	char *password = "";
 	int c;
-	pdf_write_options opts = { 0 };
+	pdf_write_options opts = pdf_default_write_options;
 	pdf_document *doc;
 	fz_context *ctx;
 
